/*
 *  Set up the smearing functions that are
 *  required for the variational static calculation.
 *
 *  Currently this code produces a number of local
 *  smearing functions.
 *
 *
 *
 *
 */

#include<stdio.h>
#include<math.h>
#include<string.h>

/*** function declerations used in this code ****/
void titles (int nx, int ny, int nz, int, int );
void write_smear_wave(float *data, int nx, int dim, char filename[]);

void fft_smearing_func(float *array, int dim);

void create_local_oper(float *src, int dim, char filename[]);
void create_wall_oper(float *src, int dim, char filename[]);
void create_random_oper(float *src, int *seed, int dim, char filename[]);
void create_expon_oper(float *src, int nx, char filename[]);
void create_hydrogen(float *src, int quantum_n, float (*func)(float),
int nx, char filename[]);



float two_s_hydrogen_poly(float x) ;
float three_s_hydrogen_poly(float x) ;
float four_s_hydrogen_poly(float x) ;
float five_s_hydrogen_poly(float x) ;


void create_2S_polyexpon_oper(float *src, int nx, char filename[]);
void create_3S_polyexpon_oper(float *src, int nx, char filename[]);

void create_gaussian_oper(float *src, int nx, char filename[]) ; 
void create_milc_gaussian_oper(float *src, int nx, char filename[]) ;

void create_squueze_expon_oper(float *src, int nx, char filename[]) ;
void create_2S_polyexpon_oper_x_squeeze(float *src, int nx, char filename[]) ;
void create_3S_polyexpon_oper_x_squeeze(float *src, int nx, char filename[]) ;

/** types of smearing functions that cen be created ****/
enum smearing_func { LOCAL , WALL , RANDOM , EXP , H_2S, H_3S, H_4S, H_5S ,
		       POLYEXP_2S , POLYEXP_3S  , GAUSS , GAUSS_MILC , EXP_SQUEEZE , 
		       POLYEXP_2S_SQUEEZE , POLYEXP_3S_SQUEEZE } ;

int main()
{
  const int nx = 24 ;
  const int ny = nx ;
  const int nz = nx ;
  int nxyz = nx*ny*nz ;

/**     const int nosmear = 3 ;   ***/
  const int nosmear = 1 ; 

  float *wave_func ;
  int seed = -1020 ;
  int ismear ;
  char filename[80] ;
/**  int smear_type[] = { LOCAL, WALL , LOCAL, LOCAL };   ***/
/**     int smear_type[] = { RANDOM, EXP , RANDOM, WALL };   ***/ 
/**    int smear_type[] = { EXP , POLYEXP_2S  , POLYEXP_3S };  ****/
/**  int smear_type[] = { EXP , RANDOM };   ***/
/**  int smear_type[] = { POLYEXP_3S };    ****/
/** int smear_type[] = { GAUSS_MILC };  ***/
/**  int smear_type[] = { EXP_SQUEEZE , POLYEXP_2S_SQUEEZE , POLYEXP_3S_SQUEEZE } ;  ***/

/**  int smear_type[] = { GAUSS_MILC } ; **/
  int smear_type[] = { EXP } ;		       


  /********----------*********----------**************/

  /** titles ****/
  titles(nx, ny, nz, seed, nosmear);

  /** Reserve the required memory **********/
  if( (wave_func = calloc( (size_t) nxyz, sizeof(float) )  ) == NULL )
  {
    printf("ERROR: could not reserve  space for wave function");
    exit(1);
  }


  /** write to disk the local operator ****/
  create_local_oper(wave_func , nxyz, filename);  
  write_smear_wave(wave_func, nx,nxyz, filename);


  /*** loop over the smearing functions ****/
  for(ismear = 0 ; ismear < nosmear ; ++ismear )
  {

    /*** create the required smearing function *****/
    switch (smear_type[ismear] )
    {
    case LOCAL:
      create_local_oper(wave_func , nxyz, filename);
      break ;
    case WALL:
      create_wall_oper(wave_func , nxyz, filename);
      break ;
    case RANDOM:
      create_random_oper(wave_func , &seed, nxyz, filename);
      break ;
    case EXP:
      create_expon_oper(wave_func ,  nx, filename);
      break ;
    case H_2S:
      create_hydrogen(wave_func ,  2,&two_s_hydrogen_poly ,nx, filename);
      break ;
    case H_3S:
      create_hydrogen(wave_func ,  3,&three_s_hydrogen_poly ,nx, filename);
      break ;
    case H_4S:
      create_hydrogen(wave_func , 4,&four_s_hydrogen_poly ,nx, filename);
      break ;
    case H_5S:
      create_hydrogen(wave_func ,  5,&five_s_hydrogen_poly ,nx, filename);
      break ;
    case POLYEXP_2S:
      create_2S_polyexpon_oper(wave_func, nx, filename) ;
      break ;
    case POLYEXP_3S:
      create_3S_polyexpon_oper(wave_func, nx, filename) ;
      break ;
    case  GAUSS :
      create_gaussian_oper( wave_func  , nx, filename) ;
      break ;
    case  GAUSS_MILC :
      create_milc_gaussian_oper(wave_func  , nx, filename) ;
      break ;
    case EXP_SQUEEZE :
      create_squueze_expon_oper(wave_func  , nx, filename) ;
      break ;
    case POLYEXP_2S_SQUEEZE : 
      create_2S_polyexpon_oper_x_squeeze(wave_func  , nx, filename) ;
      break ;
    case POLYEXP_3S_SQUEEZE : 
      create_3S_polyexpon_oper_x_squeeze(wave_func  , nx, filename) ;
      break ;
    default :
      printf("ERROR: Bad smearing function flag %d\n",ismear);
      exit(10);
  }

    write_smear_wave(wave_func, nx,nxyz, filename);
    printf("\n\n\n");

  } /*** end the loop over the smearing functions ***/


  /** free up the required memory ****/
  free(wave_func);


  return 0 ;
}














