/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/*
 * $Log: Histogram.c,v $
 * Revision 1.28  2000/08/21 00:34:27  paradyn
 * Added additional colors for graphs
 *
 * Revision 1.27  1999/08/09 05:38:19  csserra
 * - added support for (mips-sgi-irix6.4) native compiler build
 * - eliminated misc. compiler warnings
 *
 * Revision 1.26  1997/06/16 21:02:47  tamches
 * some free()'s changed to XtFree()'s, as appropriate.
 *
 * Revision 1.25  1996/08/16 19:03:36  tamches
 * updated copyright notice for release 1.1
 *
 * Revision 1.24  1996/01/19 21:01:31  newhall
 * changed parameters to HistSetCurveData
 *
 * Revision 1.23  1995/12/11 02:27:07  newhall
 * bug fix to HistHideCurve so that y-axis is redrawn when last curve is hidden
 *
 * Revision 1.22  1995/10/20  20:56:12  newhall
 * fix to SmoothData so that is no longer trying to smooth NaN values
 *
 * Revision 1.21  1995/10/19  02:42:09  newhall
 * added support for changing between black-and-white and color displays.
 * added defaultGC resource to Histogram Widget--if true, start in color mode,
 * if false, start in BW mode. added HistColorChange function.  purify fixes.
 *
 * Revision 1.20  1995/10/13  22:36:23  newhall
 * filled rectangles and  (smoothed) label
 *
 * Revision 1.19  1995/09/27  16:21:55  newhall
 * added endpoints to curves to designate begining/end of valid curve values
 * fix so that zero valued curve data shows up
 *
 * Histogram.c - Histogram Widget.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <memory.h>
#include <ctype.h>
#include <assert.h>

#include <X11/X.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/FormP.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/Scrollbar.h> 
#include <X11/Xaw/Simple.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Toggle.h>


#include <X11/cursorfont.h>
#include "HistogramP.h"

int fprintf(FILE *stream, const char *format, ... );
int sscanf(const char *s, const char *format, ... );

#define ASS_EQ(s1,s2)
#define ASS_NOTEQ(s1,s2) 
#define ASS_LE(s1,s2) 
#define ASS_T(s1) 

#define MAX_WIDTH		900
#define MAX_HEIGHT		900
#define DEF_NO_BORDER		0

#define DEF_GRAPH_BORDER_WIDTH	1 /* for graph window */
#define DEF_GRAPH_WIDTH		600 /* for graph window */
#define DEF_GRAPH_HEIGHT	200 /* for graph window */
#define DEF_MAX_TIME		50  
#define MIN_GRAPH_WIDTH		30
#define MIN_GRAPH_HEIGHT	20

#define DEF_START_TIME		0.0
#define DEF_TIME_AXIS_HEIGHT	30
#define DEF_NUM_MARKS		10
#define DEF_PERCENT_CUT		0.5

#define DEF_Y_MAX		5.0
#define MAX_Y_MIN		100.0
#define DEF_NUM_Y_MARKS 	10
#define DEF_Y_INC		0.2
#define MAX_Y_INC 		3.0
#define T_MARGIN_Y		9

#define INIT_BOTTOM_HEIGHT	40
#define BITMAP_LINE_WIDTH	50


#define NUM_NEIGHBOR 		8  /* for smoothing curves */
#define DEFAULT_COLOR_MODE 	TRUE  /* color GC false */

#define DEF_NUM_CURVE 		5  /* def # curves */
#define DEF_NUM_GROUP		5 	
#define MAX_GROUP		20
#define MAX_CURVE		20
#define DEF_INC_CURVE		3 /*increament to maxCurves when it's not enough*/
#define DEF_INC_GROUP		3 

#define R_MARGIN		15 /* right margin from graph->boundary*/
#define T_MARGIN		40 /* top margin */
#define SPACE			4 /* space between labels */
#define MARK_WIDTH		5 /* marker len in y-axis */
#define EXPWIDTH		5

#define DEF_FONT_NAME 		"6x13"
#define DEF_BW_LINE             "fg:solid;fg:1,1;fg:1,3;fg:2,2;fg:3,1"
#define DEF_COLOR_LINE          \
"blue:solid;yellow:solid;red:solid;turquoise1:solid;DarkViolet:solid;magenta:solid;medium sea green:solid;orange2:solid;brown:solid;green:solid;maroon:solid;pink:solid"
#define NUM_COMP		2

/* values of timeUnit field */
#define TIME_UNIT_SEC		0
#define TIME_UNIT_MIN		1	/* really min:sec */
#define TIME_UNIT_HOUR		2	/* really hour:min */

/* Label for time Units axis */
#define STRING_SEC		"Seconds"
#define STRING_MIN_SEC		"Min:sec"
#define STRING_HR_MIN		"Hour:Min"

/* Cut over for next units */
#define MAX_SEC			200.0
#define MAX_MIN			(60.0*60.0)

#define EPSILON			0.0001

#define WIDTH(w, cw)  (w->hist.cw->core.width)
#define HEIGHT(w, cw) (w->hist.cw->core.height)
#define BORDER_W(w, cw) (w->hist.cw->core.border_width)

/*y position of time value string on time Axis */
#define Y_POS_TIME_LAB(hw) (MARK_WIDTH + SPACE + hw->hist.numberStruct.ascent)

/*y pos of time unit string (seconds or min_sec) */
#define Y_POS_TIME_UNIT(hw) (Y_POS_TIME_LAB(hw) + hw->hist.numberHeight + 2*SPACE)

#define Y_INTERVAL_W(y)   (y.unitWidth * y.interval)

#define LABEL_HEIGHT(hw) (hw->hist.fontHeight )

#define CHANGE_YAXES_SIZE(hw)					\
    if (hw->hist.yAxesWidth < hw->hist.timeMarginWidth) {  	\
       	XtMakeResizeRequest(hw->hist.yAxesW, hw->hist.timeMarginWidth, HEIGHT(hw,yAxesW), NULL, NULL);							\
     }								\
    else {							\
        XtMakeResizeRequest(hw->hist.yAxesW, hw->hist.yAxesWidth, HEIGHT(hw,yAxesW), NULL, NULL);							\
     }

#define MIN(a,b) (((a)<(b))?(a):(b))

#define MAX(x, y) (((x) > (y)) ? (x) : (y))

/* redisplay time axis window, when possible, save redisplay the time
 * unit string */
#define QUICK_DISPLAY_TIME_AXIS(hw, unit, event) 	\
    XClearWindow(XtDisplay(hw->hist.timeAxisW), XtWindow(hw->hist.timeAxisW)); \
    RedisplayTimeAxis(hw->hist.timeAxisW, False, NULL);  
#if	defined(SOLARIS)
#endif	/* defined(SOLARIS) */

#define FIRSTPT 	0

#define offset(field) XtOffsetOf(HistogramRec, hist.field)

static float y_inc = DEF_Y_INC, 
	     y_min = EPSILON;

static XtResource resources[] = {
      {
	XtNgraphHeight, XtCGraphHeight, XtRDimension, sizeof(Dimension),
	offset(gHeight), XtRImmediate, (XtPointer)DEF_GRAPH_HEIGHT},
      {
	XtNgraphWidth, XtCGraphWidth, XtRDimension, sizeof(Dimension),
	offset(gWidth), XtRImmediate, (XtPointer)DEF_GRAPH_WIDTH},
      {
	XtNmaxGroup, XtCMaxGroup, XtRInt, sizeof(int),
	offset(maxGroup), XtRImmediate, (XtPointer)DEF_NUM_GROUP },
      {
	XtNmaxCurve, XtCMaxCurve, XtRInt, sizeof(int),
	offset(maxCurve), XtRImmediate, (XtPointer)DEF_NUM_CURVE },
      {
	XtNgraphForeground, XtCGraphForeground, XtRPixel, sizeof(Pixel),
	offset(gForeground), XtRString,XtDefaultForeground},
      { 
	XtNgraphBackground, XtCGraphBackground, XtRPixel, sizeof(Pixel),
	offset(gBackground), XtRString,XtDefaultBackground},
      { 
	XtNfont, XtCString, XtRString, sizeof(char *),
	offset(font), XtRString, (caddr_t)DEF_FONT_NAME},
      { 
	XtNnumXMarks, XtCNumXMarks, XtRInt, sizeof(int),
	offset(numXMarks), XtRImmediate, (XtPointer) DEF_NUM_MARKS},
      { 
	XtNyInc, XtCYInc, XtRFloat, sizeof(float),
	offset(yInc), XtRImmediate, (XtPointer)&y_inc},
      { 
	XtNyMin, XtCYMin, XtRFloat, sizeof(float),
	offset(yMin), XtRImmediate, (XtPointer) &y_min},
      { 
	XtNsmoothNpts, XtCSmoothNpts, XtRInt, sizeof(int),
	offset(smoothNpts), XtRImmediate, (XtPointer) NUM_NEIGHBOR},
      { 
	XtNdefaultGC, XtCDefaultGC, XtRInt, sizeof(Boolean),
	offset(defaultGC), XtRImmediate, (XtPointer) DEFAULT_COLOR_MODE},
      { 
	XtNcolorLineStyles, XtCColorLineStyles, XtRString, sizeof(char*),
	offset(colorLineStyles), XtRString, (caddr_t) DEF_COLOR_LINE },
      { 
	XtNbwLineStyles, XtCBwLineStyles, XtRString, sizeof(char*),
	offset(bwLineStyles), XtRString, (caddr_t) DEF_BW_LINE },
};


/* clean this up to be part of histogram struct */
static int fontHeight;
static GC  invisible_gc;

/* Declaration of methods appearing in class record */

#ifndef USES_NATIVE_CC
static void InitializeHist();
static void RealizeHist();
static XtGeometryResult GeometryManagerHist();
static void ResizeHist();  /* resize */
static void ConstraintInitializeHist();
static void DestroyHist(); /* destroy */
static int CreateLineStyles();
static int CreateBWLineStyles();
#else
static void InitializeHist(Widget, Widget, ArgList, Cardinal *);
static void RealizeHist(Widget, XtValueMask *, XSetWindowAttributes *);
static XtGeometryResult GeometryManagerHist();
static void ResizeHist(Widget);  /* resize */
static void ConstraintInitializeHist(Widget, Widget, ArgList, Cardinal *);
static void DestroyHist(Widget); /* destroy */
static int CreateLineStyles();
static int CreateBWLineStyles();
#endif /* USES_NATIVE_CC */

/* private functions in Histogram.c */

/* exposure of graph window */
static void RedisplayGraph(Widget, XtPointer, XExposeEvent *);
static void RedisplayYAxes();
static void RedisplayTimeAxis();
static void RedisplayPAN();
static void RedisplayZOOM();

static void ReLayout();
static void RedisplayLine();
static void Get_GCs();


static void ChangeGraphVisible();
static XExposeEvent *GetExposeBox();

/* our callbacks for horizontal scrollbar and zoombar */
static void ZoomScrollFunc(), ZoomHistogram();
static void HScrollFunc(), HJumpFunc();
static void ScrollHistogram();

static void ShiftGraph();

/* private functions called by exposure or convenience functions */

static int  NextLineStyleGCInd();
static void FreeALineStyle();
static void AddNewGroup();
static void AddYAxis();
static void AddBottomLabel();
static void LabelToggle();
static void DrawYAxis();
static void DrawCurve();
static void CalculateXPoints();
static void RefigureUnitWidth();
static void RefigureTimeUnitWidth();
static void ProcessYMaxChange();
static void SmoothData();


HistogramClassRec histogramClassRec = {
  { 
  /* core class fields */
    /* superclass 		*/	(WidgetClass) &formClassRec,
    /* class_name		*/	"Histogram",
    /* widget_size		*/	sizeof(HistogramRec),
    /* class_initialize         */      NULL,
    /* class_part_initialize    */      NULL,
    /* class_inited             */      FALSE,
    /* initialize               */      InitializeHist,
    /* initialize_hook          */      NULL,
    /* realize                  */      RealizeHist,
    /* actions                  */      NULL,
    /* num_actions              */      0,
    /* resources                */      resources,
    /* num_resources            */      XtNumber(resources),
    /* xrm_class                */      NULLQUARK,
    /* compress_motion          */      TRUE,
    /* compress_exposure        */      XtExposeCompressSeries,
    /* compress_enterleave      */      TRUE,
    /* visible_interest         */      TRUE,
    /* destroy                  */      DestroyHist,
    /* resize                   */      ResizeHist,
    /* expose                   */      XtInheritExpose,
    /* set_values               */      NULL,
    /* set_values_hook          */      NULL,
    /* set_values_almost        */      XtInheritSetValuesAlmost,
    /* get_values_hook          */      NULL,
    /* accept_focus             */      NULL,
    /* version                  */      XtVersion,
    /* callback_private         */      NULL,
    /* tm_table                 */      NULL,
    /* query_geometry           */      (XtGeometryHandler)XtInheritQueryGeometry,
    /* display_accelerator	*/      XtInheritDisplayAccelerator,
    /* extension        	*/      NULL,
  }, {
    /* composite part 	*/
    /* geometry_manager         */   	(XtGeometryHandler)GeometryManagerHist, 
    /* change_managed     	*/      XtInheritChangeManaged,
    /* insert_child       	*/      XtInheritInsertChild,
    /* delete_child       	*/      XtInheritDeleteChild,
    /* extension            	*/      NULL,
  }, {
    /* contraints class fields */
    /* subresources		*/ 	NULL,
    /* subresource_count	*/      0,
    /* constraint_size		*/	sizeof(HistogramConstraintsRec),
    /* initialize		*/	ConstraintInitializeHist,
    /* destroy			*/	NULL,
    /* set_values		*/	NULL,
    /* extension		*/	NULL,
  },{
    /* form class fields */
    /* layout			*/	XtInheritLayout,
  }, {
    /* dummy 			*/	0,
    }
};

WidgetClass histogramWidgetClass = (WidgetClass) &histogramClassRec;

#ifndef USES_NATIVE_CC
inline 
#endif
float getShown(Widget w)
{
    float ret;

    XtVaGetValues(w, XtNshown, (char *) &ret, NULL);
    return(ret);
}

void computeLabelSpace(int space, double val, int *number, double *inc, char *format)
/* space:  the space available for labels */
/* val:    the largest value to be shown on the histogram */
/* number: the number of labels on the histogram (not including the 0.0 label) */
/* inc:    the increment to get to the max from 0 */
{
    register double spaceNeeded;   /* spaceNeeded * increment == val */
    register double increment;
    register int decimal;  /* number of digits after the decimal point of inc */
    register int integral;
    int precision;

    if (val <= 0.00000001) {
	spaceNeeded = 0.99;
	increment = 1.0;
    } else if (val < 0.0) {
	spaceNeeded = -val;
	increment = -1.0;
    } else {
	spaceNeeded = val;
	increment = 1.0;
    }
    decimal = 0;
    if (spaceNeeded < 1.0) {
	integral = 1;
	while (spaceNeeded < 0.1) {
	    spaceNeeded *= 10.0;
	    increment /= 10.0;
	    decimal++;
	}
    } else {
	integral = 0;
	while (spaceNeeded >= 1.0) {
	    spaceNeeded /= 10.0;
	    increment *= 10.0;
	    decimal--;
	    integral++;
	}
    }
    precision = 0;
    while (1 + (int)(3.0 * spaceNeeded) < space) {
	if (1 + (int)(6.0 * spaceNeeded) < space)
	    if (1 + (int)(15.0 * spaceNeeded) < space) {
		increment /= 10.0;
		spaceNeeded *= 10.0;
	    } else {
		increment /= 5.0;
		spaceNeeded *= 5.0;
	    }
	else {
	    increment /= 2.0;
	    spaceNeeded *= 2.0;
	}
	precision++;
	decimal++;
    }
    (*inc) = increment;
    (*number) = 1 + (int)spaceNeeded;
    if (val == (*number - 1) * *inc)
	(*number)--;
    if (integral + (decimal > 0 ? decimal + 1 : 0) > precision + EXPWIDTH) {
	(void)sprintf(format, "%%%d.%de", 0, precision - 1);
    } else if (decimal > 0) {
	(void)sprintf(format, "%%%d.%df", 0, decimal);
    } else {
	(void)sprintf(format, "%%%d.0f", 0);
    }
}

void ComputeGroupYLabel(HistogramWidget hw, Group *group)
{
  int space;

  /* compute labels and increments */
  fontHeight = (hw->hist.fontInfo->max_bounds).ascent +
      (hw->hist.fontInfo->max_bounds).descent;
  space = (int)((float)HEIGHT(hw,graphW)/ (float)fontHeight);
  computeLabelSpace(space, group->y.max, &group->y.count, 
      &group->y.interval, group->y.format);
  group->y.tickSpace = (float)HEIGHT(hw,graphW) / (float) (group->y.count);
  group->y.unitWidth = group->y.tickSpace/group->y.interval;

#ifdef notdef
  printf("height = %d\n", HEIGHT(hw,graphW));
  printf("unitWidth = %f\n", group->y.unitWidth);
  printf("ymax = %f\n", group->y.max);
  printf("count = %d\n", group->y.count);
  printf("increment = %f\n", group->y.interval);
#endif

}

  
void ComputeGroupWidth(HistogramWidget hw, Group *group)
{
  int len;
  char buff[30];
  int numWidth;

  ComputeGroupYLabel(hw, group);

  sprintf(buff, group->y.format, group->y.max);
  len = strlen(buff);
  numWidth = XTextWidth(hw->hist.fontInfo, buff, len);
  group->needWidth = (group->titleWidth > numWidth + SPACE) ? group->titleWidth
      + MARK_WIDTH + SPACE + 1: numWidth + MARK_WIDTH + 2*SPACE + 1;
}


/**************************************************************
    private histogram class methods 
***************************************************************/

/* ARGSUSED */
static void InitializeHist(Widget request_, Widget new_, 
			   ArgList dummy0, Cardinal *dummy1)
{
    HistogramWidget request = (HistogramWidget)request_;
    HistogramWidget new = (HistogramWidget)new_;
    int i;
    Arg args[20];
    Cursor down_cursor;
    int dummy;
    XCharStruct numberStruct, charStruct;
    char buff[20];
    int temp;


    new->form.default_spacing = 0;
    new->hist.timeUnit = TIME_UNIT_SEC;
    new->hist.startTime = DEF_START_TIME;

    if (request->hist.numXMarks > 0) {
	new->hist.numXMarks = request->hist.numXMarks;
    } else {
	new->hist.numXMarks = DEF_NUM_MARKS;
    }

    new->hist.numCurve = new->hist.numShownCurve = 0;
    if (new->hist.maxCurve < 1 || new->hist.maxCurve > MAX_CURVE)  
    	new->hist.maxCurve = DEF_NUM_CURVE;
    if (new->hist.maxGroup < 1 || new->hist.maxGroup > MAX_GROUP)
	new->hist.maxGroup = DEF_NUM_GROUP;
    if (new->hist.yInc < 0.0 || new->hist.yInc > MAX_Y_INC){
	new->hist.yInc = DEF_Y_INC;
   }

    /* TEMP 
    new->hist.yInc = DEF_Y_INC;
    */

    if (new->hist.smoothNpts < 5 || new->hist.smoothNpts > 100) {
	new->hist.smoothNpts = NUM_NEIGHBOR;
    }
    new->hist.numGroup = 0;
    new->hist.currCurveID = 0;
    new->hist.timeWidth = 0.0;
    new->hist.maxSample = 0;
    new->hist.startTime = DEF_START_TIME;

    new->hist.curves = (InternalCurveRec **) XtCalloc(new->hist.maxCurve,sizeof(InternalCurveRec *));
    for (i=0; i< new->hist.maxCurve; i++)
    	new->hist.curves[i] = NULL;

    new->hist.groups = (Group **) XtCalloc(new->hist.maxGroup, sizeof(Group *));
    for (i=0; i< new->hist.maxGroup; i++)
	new->hist.groups[i] = NULL;

    new->hist.fontInfo = XLoadQueryFont(XtDisplay(new), (new->hist).font);
    new->hist.fontID = new->hist.fontInfo->fid; 
    new->hist.fontHeight = new->hist.fontInfo->descent + new->hist.fontInfo->ascent;
    XTextExtents(new->hist.fontInfo, "0", 1, &dummy, &dummy, &dummy, &(numberStruct)); 

    XTextExtents(new->hist.fontInfo, "S", 1, &dummy, &dummy, &dummy, &(charStruct)); 

    memcpy(&(new->hist.numberStruct),(char*)&numberStruct,sizeof(XCharStruct)); 
    memcpy(&(new->hist.charStruct), (char*)&charStruct, sizeof(XCharStruct)); 
    new->hist.numberHeight = numberStruct.descent + numberStruct.ascent;
    new->hist.numberWidth = numberStruct.width;
    new->hist.charHeight = charStruct.descent + charStruct.ascent;
    new->hist.charWidth = charStruct.width;

    sprintf(buff, "%s", "900.00");
    temp = XTextWidth(new->hist.fontInfo, buff, strlen(buff));
    new->hist.timeMarginWidth = (int) ((temp + 1)/2);

    new->hist.minYLabelSpace = 1.5 * new->hist.numberHeight;
    new->hist.minXLabelSpace = (float)temp;

    new->hist.yAxesWidth = 0;

    new->hist.clearGraph = False;
    new->hist.clearY = False;
    new->hist.clearX = False;

    new->hist.foreground = BlackPixel(XtDisplay(new), 
	DefaultScreen(XtDisplay(new)));
    new->hist.patternGC = NULL;
    new->hist.bw_patternGC = NULL;
    new->hist.currLineStyle = 0;

    CreateLineStyles(new);
    CreateBWLineStyles(new);

/***************************************************
HistogramWidget includes the following widgets
"X" represents nothing 

--------------------------------------------------
|			emptyW			 |			
--------------------------------------------------
|	|				|X | | z |
|vertical|		data		|X | | o |
|       |				|X | | o |
|_______|_______________________________|X_|Z| m |
| X  |		horizontal		   |O|	 |
|_X__|_____________________________________|O| b |
|		|			   |M| a |
|bottomLabelW	|			   | | r |
|_______________|__________________________|_| W |
|		PAN			     |   |
|____________________________________________|___|
|		scrollBarW			 |
|________________________________________________|
****************************************************/


    /********** Y TITLES WIDGET */
    i = 0;
    XtSetArg(args[i], XtNx, 0); i++;
    XtSetArg(args[i], XtNy, 0); i++;
    XtSetArg(args[i], XtNwidth, new->hist.timeMarginWidth); i++;
    XtSetArg(args[i], XtNborderWidth, DEF_NO_BORDER);i++;
    XtSetArg(args[i], XtNresizable, True);i++;
    XtSetArg(args[i], XtNright, XtChainLeft); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainTop); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;

    new->hist.yAxesW = XtCreateManagedWidget("yAxesW", coreWidgetClass,
        (Widget) new, args, i);
    
    XtAddEventHandler(new->hist.yAxesW, ExposureMask, FALSE, 
	RedisplayYAxes, (XtPointer)True);

    /* get around with form widget, this part is for top empty space 
     * above the graph window */
    i = 0; 
    XtSetArg(args[i], XtNheight, T_MARGIN); i++;
    XtSetArg(args[i], XtNwidth,10); i++;
    XtSetArg(args[i], XtNborderWidth, 0); i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainTop); i++;
    XtSetArg(args[i], XtNbottom, XtChainTop); i++;
    XtSetArg(args[i], XtNfromHoriz, new->hist.yAxesW); i++;
    XtSetArg(args[i], XtNhorizDistance, 0); i++;
    new->hist.emptyW= XtCreateManagedWidget("emptyW", coreWidgetClass,
        (Widget) new, args, i);

   

    /********** GRAPH WIDGET */
    /*!!!!!!!!!!!!!!!have to set foregound and background later */
    i = 0;
    XtSetArg(args[i], XtNheight, ((new->hist.gHeight > 0) ? 
			new->hist.gHeight : DEF_GRAPH_HEIGHT));i++; 
    XtSetArg(args[i], XtNwidth, ((new->hist.gWidth > 0) ? 
			new->hist.gWidth : DEF_GRAPH_WIDTH)); i++; 
    XtSetArg(args[i], XtNborderWidth, DEF_GRAPH_BORDER_WIDTH); i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainTop); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;
    XtSetArg(args[i], XtNfromHoriz, new->hist.yAxesW); i++;
    XtSetArg(args[i], XtNhorizDistance, 0); i++;
    XtSetArg(args[i], XtNfromVert, new->hist.emptyW); i++;
    XtSetArg(args[i], XtNvertDistance, 0); i++;

    new->hist.graphW = XtCreateManagedWidget("data", simpleWidgetClass,
        (Widget) new, args, i);
    
    XtAddEventHandler(new->hist.graphW, ExposureMask, FALSE, 
	(XtEventHandler)RedisplayGraph, 
	(XtPointer)True);
    XtAddEventHandler(new->hist.graphW, VisibilityChangeMask, FALSE, 
	ChangeGraphVisible, new);
    down_cursor = XCreateFontCursor(XtDisplay(new->hist.graphW), 
	XC_sb_down_arrow);

    XtVaSetValues(new->hist.graphW, XtNcursor, down_cursor, NULL);

    /*********************** remaining of YAXES WIDGET ******/

    XtVaSetValues(new->hist.yAxesW, XtNheight, new->hist.graphW->core.height+ T_MARGIN+ new->hist.numberHeight/2 + 2, NULL);


    /********* TIME AXIS WIDGET */
    i = 0;
    XtSetArg(args[i], XtNx, new->hist.yAxesW->core.width - new->hist.timeMarginWidth);
    XtSetArg(args[i], XtNwidth, new->hist.graphW->core.width + 2 * new->hist.timeMarginWidth); i++;
    XtSetArg(args[i], XtNheight, MARK_WIDTH + SPACE + new->hist.numberHeight + 2*SPACE + new->hist.charHeight + SPACE);i++;
    XtSetArg(args[i], XtNborderWidth, DEF_NO_BORDER);i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainBottom); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;
    XtSetArg(args[i], XtNfromVert, new->hist.graphW); i++;
    XtSetArg(args[i], XtNvertDistance, 0); i++;
    new->hist.timeAxisW = XtCreateManagedWidget("emptyW2", coreWidgetClass, 
	(Widget) new, args, i);
    
    XtAddEventHandler(new->hist.timeAxisW, ExposureMask, FALSE, 
	RedisplayTimeAxis, (XtPointer) True);
   

    /********* BOTTOM LABEL WIDGET */
    i = 0;
    /* make sure it has nonzero height and width */
    XtSetArg(args[i], XtNheight, LABEL_HEIGHT(new)); i++;
    XtSetArg(args[i], XtNresizable, True); i++;
    XtSetArg(args[i], XtNwidth, 10); i++;
    XtSetArg(args[i], XtNborderWidth, DEF_NO_BORDER);i++;
    XtSetArg(args[i], XtNdefaultDistance, 0); i++;
    XtSetArg(args[i], XtNresizable, True);i++;
    XtSetArg(args[i], XtNright, XtChainLeft); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainBottom); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;
    XtSetArg(args[i], XtNfromVert, new->hist.timeAxisW); i++;
    XtSetArg(args[i], XtNvertDistance, 0); i++;

    new->hist.bottomLabelW = XtCreateManagedWidget("bottomLabel", 
	formWidgetClass, (Widget) new, args, i);

    /********** string PAN Widget */
    i = 0;
    XtSetArg(args[i], XtNborderWidth, DEF_NO_BORDER);i++;
    XtSetArg(args[i], XtNheight, new->hist.charHeight +1); i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainBottom); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;
    XtSetArg(args[i], XtNfromVert, new->hist.bottomLabelW); i++;
    XtSetArg(args[i], XtNvertDistance, 0); i++;
    new->hist.PANW = XtCreateManagedWidget("PANW", coreWidgetClass,
        (Widget) new, args, i);
    
    XtAddEventHandler(new->hist.PANW, ExposureMask, FALSE, 
	RedisplayPAN, (XtPointer)True);

    /********** string ZOOM Widget */
    i = 0;
    XtSetArg(args[i], XtNwidth, new->hist.charWidth+ 1); i++;
    XtSetArg(args[i], XtNborderWidth, DEF_NO_BORDER);i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainRight); i++;
    XtSetArg(args[i], XtNtop, XtChainTop); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;
    XtSetArg(args[i], XtNfromVert, new->hist.emptyW); i++;
    XtSetArg(args[i], XtNvertDistance, 0); i++;
    XtSetArg(args[i], XtNfromHoriz, new->hist.PANW); i++;
    XtSetArg(args[i], XtNhorizDistance, 0); i++;
    new->hist.ZOOMW = XtCreateManagedWidget("ZOOMW", coreWidgetClass,
        (Widget) new, args, i);
    
    XtAddEventHandler(new->hist.ZOOMW, ExposureMask, FALSE, 
	RedisplayZOOM, (XtPointer)True);

    /********** Horiz. SCROLL BAR WIDGET */
    i = 0;
    XtSetArg(args[i], XtNorientation, XtorientHorizontal); i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNtop, XtChainBottom); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;

    XtSetArg(args[i], XtNfromVert, new->hist.PANW); i++;

    XtSetArg(args[i], XtNvertDistance, 0); i++;
    new->hist.scrollW = XtCreateManagedWidget("horizontal", 
	scrollbarWidgetClass, (Widget) new, args, i);
    
    XtAddCallback(new->hist.scrollW, XtNscrollProc, HScrollFunc, new);
    XtAddCallback(new->hist.scrollW, XtNjumpProc, HJumpFunc, new);
    XawScrollbarSetThumb(new->hist.scrollW, 0.0, 1.0);
   

    /********** ZOOM BAR WIDGET */
    i = 0;
    XtSetArg(args[i], XtNorientation, XtorientVertical); i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainRight); i++;
    XtSetArg(args[i], XtNtop, XtChainTop); i++;
    XtSetArg(args[i], XtNbottom, XtChainBottom); i++;
    XtSetArg(args[i], XtNfromHoriz, new->hist.ZOOMW); i++;
    XtSetArg(args[i], XtNhorizDistance, 0); i++;

    new->hist.zoomBarW = XtCreateManagedWidget("zoom", scrollbarWidgetClass, 
	(Widget) new, args, i);
    
    XtAddCallback(new->hist.zoomBarW, XtNscrollProc, ZoomScrollFunc, new);
    XtAddCallback(new->hist.zoomBarW, XtNjumpProc, ZoomHistogram, new);
    XawScrollbarSetThumb(new->hist.zoomBarW, 0.8, 0.2);

/*
 * Set necessary width and height of widgets 
 */
    new->core.border_width = DEF_NO_BORDER;
    new->hist.oldWidth = new->core.width = WIDTH(new,graphW) + 2 * BORDER_W(new,graphW) + WIDTH(new,yAxesW) + new->hist.timeMarginWidth+ WIDTH(new,zoomBarW) + 2 * BORDER_W(new,zoomBarW) + WIDTH(new, ZOOMW); 

    new->hist.oldHeight = new->core.height = HEIGHT(new,graphW) + 2 * BORDER_W(new,graphW) + HEIGHT(new,emptyW) + HEIGHT(new,timeAxisW) + HEIGHT(new,bottomLabelW) + HEIGHT(new,scrollW) + 2* BORDER_W(new,scrollW) + HEIGHT(new,PANW);

   /*????????? should be this one 
   XtVaSetValues(new->hist.scrollW, XtNwidth, new->core.width - 2 * BORDER_W(new,scrollW), NULL);
   */
   XtVaSetValues(new->hist.scrollW, XtNwidth, new->core.width - 2 * BORDER_W(new,scrollW) - 1, NULL);
   XtVaSetValues(new->hist.PANW, XtNwidth, WIDTH(new,yAxesW) + WIDTH(new,graphW) + BORDER_W(new,graphW) + new->hist.timeMarginWidth, NULL);

   XtVaSetValues(new->hist.zoomBarW, XtNheight, new->core.height - HEIGHT(new,scrollW) - 2 * BORDER_W(new,scrollW) - 2 * BORDER_W(new,zoomBarW), NULL); 
   XtVaSetValues(new->hist.ZOOMW, XtNheight, HEIGHT(new, zoomBarW) - HEIGHT(new, emptyW), NULL);
  
   new->hist.zoom= 1.0;
   new->hist.start = 0.0;
   new->hist.focus = 0.5;
   new->hist.currXLeft = 0;

   Get_GCs(new);
}





/*******???????????????????wiered, have to move timewidget to
get the correct position of it */

#define superclass 	(&formClassRec)

static void 
RealizeHist(Widget hw_, XtValueMask *valueMask, 
	    XSetWindowAttributes *attributes)
{
   HistogramWidget hw = (HistogramWidget)hw_;
   XtMoveWidget(hw->hist.timeAxisW, hw->hist.graphW->core.x - 
       hw->hist.timeMarginWidth, hw->hist.timeAxisW->core.y);
   (*superclass->core_class.realize) ((Widget) hw, valueMask, attributes);

   return;
}

static void ResizeHist(Widget hw_)
{
    HistogramWidget hw = (HistogramWidget)hw_;
    int diff_w, diff_h;
    int w = hw->core.width, h = hw->core.height;

    if (!hw->hist.graphW) return;
    if ((hw->hist.graphW->core.width < MIN_GRAPH_WIDTH) || 
        (hw->hist.graphW->core.height < MIN_GRAPH_HEIGHT)) {
	fprintf(stderr, "      no resize, too smal w=%d; h=%d\n", 
	    hw->core.width, hw->core.height);
	hw->hist.oldWidth = hw->core.width;
	hw->hist.oldHeight = hw->core.height;
	XClearWindow(XtDisplay(hw), XtWindow(hw));
	XDrawString(XtDisplay(hw), XtWindow(hw), hw->hist.histFgGC, 3, 
	    hw->hist.charHeight, "too small", strlen("too small"));
	return;
     }
     /* determine if really change size */
     if (w == hw->hist.oldWidth && h == hw->hist.oldHeight) {
       	return;
     }
     /* difference between old and new width and height */
     diff_w = hw->core.width - hw->hist.oldWidth, 
     diff_h = hw->core.height - hw->hist.oldHeight;
     /* new width and height of graphW */
     w = WIDTH(hw,graphW)+ diff_w,
     h = HEIGHT(hw,graphW) + diff_h;
     RefigureUnitWidth(hw, w, h);

     if (diff_w > 0 && diff_h > 0) { 
	hw->hist.clearGraph = hw->hist.clearY = hw->hist.clearX = True;
   	(*superclass->core_class.resize) ((Widget) hw);
    } else if (diff_w > 0 && diff_h == 0) {
   	hw->hist.clearGraph = hw->hist.clearX = True;
   	(*superclass->core_class.resize) ((Widget) hw);
    } else if (diff_w == 0 && diff_h > 0) {
   	hw->hist.clearGraph = hw->hist.clearY = True;
   	(*superclass->core_class.resize) ((Widget) hw);
    } else if (diff_w < 0 && diff_h < 0) {
	/* to replace StructureNotify event for getting smaller window, ugly */
    	hw->hist.clearGraph = hw->hist.clearY = hw->hist.clearX = True;
        (*superclass->core_class.resize) ((Widget) hw);
	RedisplayYAxes(hw->hist.yAxesW, (XtPointer)False, NULL);
	RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
	RedisplayTimeAxis(hw->hist.timeAxisW, (XtPointer)False, NULL);
	RedisplayPAN(hw->hist.PANW, (XtPointer)False, NULL);
	RedisplayZOOM(hw->hist.ZOOMW, (XtPointer)False, NULL);
    } else if (diff_w < 0) {
   	hw->hist.clearGraph = hw->hist.clearX = hw->hist.clearY = True;
        (*superclass->core_class.resize) ((Widget) hw);
	RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
	RedisplayTimeAxis(hw->hist.timeAxisW, (XtPointer)False, NULL);
	RedisplayPAN(hw->hist.PANW, (XtPointer)False, NULL);
    } else if (diff_h < 0) {
   	hw->hist.clearGraph = hw->hist.clearY = hw->hist.clearX = True;
        (*superclass->core_class.resize) ((Widget) hw);
	RedisplayYAxes(hw->hist.yAxesW, (XtPointer)False, NULL);
	RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
	RedisplayZOOM(hw->hist.ZOOMW, (XtPointer)False, NULL);
   }
   return;
}


static void
ConstraintInitializeHist(Widget request, Widget new,
			 ArgList dummy0, Cardinal *dummy1)
{

   (*superclass->constraint_class.initialize) (request, new, NULL, 0);
   return;
}




/*
 * Parse the line style resource.  This resource is really a list of
 *   line styles.
 *
 */
static int CreateLineStyles(hw)
HistogramWidget hw;
{
    Display*display = XtDisplay(hw);
    char *end;
    int count;
    char *curr;
    char *orig;
    char *style;
    XColor color;
    XColor exact;
    char *colorName;
    int comp1, comp2;				/* line comonenents */
    char message[80];
    char buf[NUM_COMP+1];
    XGCValues 	gcv, gcvr;
    int numCells;

    numCells = XDisplayCells(display, DefaultScreen(display));
    if (numCells > 2) {
            orig = curr = XtMalloc(strlen(hw->hist.colorLineStyles)+1);
            strcpy(curr, hw->hist.colorLineStyles);

    } else {
            orig = curr = XtMalloc(strlen(hw->hist.bwLineStyles)+1);
            strcpy(curr, hw->hist.bwLineStyles);
    }

    /* count the number of line styles */
    hw->hist.numPatterns = 1;
    while (*curr) {
	if (*(curr++) == ';') hw->hist.numPatterns++;
    }

    /* Create GCs and an usage array for them */
    hw->hist.patternGC = (GC *) XtCalloc(hw->hist.numPatterns, sizeof(GC));
    hw->hist.patternUsed = (int*) XtCalloc(hw->hist.numPatterns, sizeof(int));
    memset((char*)hw->hist.patternUsed, '\0', hw->hist.numPatterns*sizeof(int));


    curr = orig;
    count = 0;
    while (*curr) {
	/* parse a line style */

	/* first is the color to use or fg for the foreground color */
	colorName = curr;
	while (*curr && (*curr != ':')) curr++;
	if (*curr != ':') {
	    XtWarning("line style resource is not valid");
	    hw->hist.numPatterns = count;
	    return(0);
	}
	*curr++ = '\0';

	/* now comes the style component */
	if (!isdigit(*curr)) {
	    end = (char *) strchr(curr, ';');
	    style = curr;
	} else {
	    /* two numbers should follow */
	    sscanf(curr, "%d,%d", &comp1, &comp2);
	    style = NULL;
	}
	end = (char *) strchr(curr, ';');
	if (!end) {
	    end = &(curr[strlen(curr)]);
	    curr = end;
	} else {
	    *end = '\0';
	    curr = ++end;
	}

	/* now get ready to create the GC */
	gcv.font = gcvr.font = hw->hist.fontInfo->fid;
	gcv.line_width = gcvr.line_width = 1;

	if (!strcmp(colorName, "fg")) {
	    /* use standard colors */
	    gcv.foreground = gcvr.background = hw->hist.foreground;
	    gcv.background = gcvr.foreground = hw->core.background_pixel;
	} else {
	    gcv.background = gcvr.foreground = hw->core.background_pixel;
	    if (XAllocNamedColor(display,
		DefaultColormapOfScreen(XtScreen(hw)),
		colorName, &color, &exact) == 0) {
		    sprintf(message, "Can not allocate color %s", colorName);
		    XtWarning(message);
		gcv.foreground = gcvr.background = hw->hist.foreground;
	    } else {
		gcv.foreground = gcvr.background = color.pixel;
	    }
	}

	if (!style) {
	    buf[0] = comp1;
	    buf[1] = comp2;
	    gcv.line_style = gcvr.line_style = LineOnOffDash;
	    hw->hist.patternGC[count] = 
		XCreateGC(display, RootWindow(display, DefaultScreen(display)),
		  GCLineWidth|GCForeground|GCBackground|GCLineStyle, &gcv);
	    
	    XSetDashes(display, hw->hist.patternGC[count], 0, buf, NUM_COMP);
	} 
	else {
	      hw->hist.patternGC[count] = 
		XCreateGC(display, RootWindow(display, DefaultScreen(display)),
			GCLineWidth | GCForeground | GCBackground, &gcv);
	}
	count++;
    }
    if (count != hw->hist.numPatterns) {
	XtWarning("Syntax error in line description");
	hw->hist.numPatterns = count;
    }
    XtFree(orig);

    return(0);
}

/*
 * Parse the bw line style resource.  This resource is really a list of
 *   line styles.
 *
 */
static int CreateBWLineStyles(hw)
HistogramWidget hw;
{
    Display*display = XtDisplay(hw);
    char *end;
    int count;
    char *curr;
    char *orig;
    char *style;
    XColor color;
    XColor exact;
    char *colorName;
    int comp1, comp2;				/* line comonenents */
    char message[80];
    char buf[NUM_COMP+1];
    XGCValues 	gcv, gcvr;

    orig = curr = XtMalloc(strlen(hw->hist.bwLineStyles)+1);
    strcpy(curr, hw->hist.bwLineStyles);

    /* count the number of line styles */
    hw->hist.bw_numPatterns = 1;
    while (*curr) {
	if (*(curr++) == ';') hw->hist.bw_numPatterns++;
    }

    /* Create GCs and an usage array for them */
    hw->hist.bw_patternGC = (GC *) XtCalloc(hw->hist.bw_numPatterns, sizeof(GC));
    hw->hist.bw_patternUsed = (int*) XtCalloc(hw->hist.bw_numPatterns, sizeof(int));
    memset((char*)hw->hist.bw_patternUsed, '\0', hw->hist.bw_numPatterns*sizeof(int));


    curr = orig;
    count = 0;
    while (*curr) {
	/* parse a line style */

	/* first is the color to use or fg for the foreground color */
	colorName = curr;
	while (*curr && (*curr != ':')) curr++;
	if (*curr != ':') {
	    XtWarning("line style resource is not valid");
	    hw->hist.bw_numPatterns = count;
	    return(0);
	}
	*curr++ = '\0';

	/* now comes the style component */
	if (!isdigit(*curr)) {
	    end = (char *) strchr(curr, ';');
	    style = curr;
	} else {
	    /* two numbers should follow */
	    sscanf(curr, "%d,%d", &comp1, &comp2);
	    style = NULL;
	}
	end = (char *) strchr(curr, ';');
	if (!end) {
	    end = &(curr[strlen(curr)]);
	    curr = end;
	} else {
	    *end = '\0';
	    curr = ++end;
	}

	/* now get ready to create the GC */
	gcv.font = gcvr.font = hw->hist.fontInfo->fid;
	gcv.line_width = gcvr.line_width = 1;

	if (!strcmp(colorName, "fg")) {
	    /* use standard colors */
	    gcv.foreground = gcvr.background = hw->hist.foreground;
	    gcv.background = gcvr.foreground = hw->core.background_pixel;
	} else {
	    gcv.background = gcvr.foreground = hw->core.background_pixel;
	    if (XAllocNamedColor(display,
		DefaultColormapOfScreen(XtScreen(hw)),
		colorName, &color, &exact) == 0) {
		    sprintf(message, "Can not allocate color %s", colorName);
		    XtWarning(message);
		gcv.foreground = gcvr.background = hw->hist.foreground;
	    } else {
		gcv.foreground = gcvr.background = color.pixel;
	    }
	}

	if (!style) {
	    buf[0] = comp1;
	    buf[1] = comp2;
	    gcv.line_style = gcvr.line_style = LineOnOffDash;
	    hw->hist.bw_patternGC[count] = 
		XCreateGC(display, RootWindow(display, DefaultScreen(display)),
		  GCLineWidth|GCForeground|GCBackground|GCLineStyle, &gcv);
	    
	    XSetDashes(display, hw->hist.bw_patternGC[count], 0, buf, NUM_COMP);
	} 
	else {
	      hw->hist.bw_patternGC[count] = 
		XCreateGC(display, RootWindow(display, DefaultScreen(display)),
			GCLineWidth | GCForeground | GCBackground, &gcv);
	}
	count++;
    }
    if (count != hw->hist.bw_numPatterns) {
	XtWarning("Syntax error in line description");
	hw->hist.bw_numPatterns = count;
    }
    XtFree(orig);

    return(0);
}

static void Get_GCs(new)
HistogramWidget new;
{
    XGCValues gcv;
    XtGCMask  mask; 

    gcv.font = new->hist.fontID;
    gcv.line_style = LineSolid;
    gcv.foreground = new->hist.foreground;
    gcv.background = new->core.background_pixel;
    mask = GCFont | GCForeground | GCBackground;
    new->hist.histFgGC = XtGetGC((Widget) new, mask, &gcv);

    gcv.background = new->hist.foreground;
    gcv.foreground = new->core.background_pixel;
    mask = GCFont | GCForeground | GCBackground;
    new->hist.histBgGC = XtGetGC((Widget) new, mask, &gcv);
    invisible_gc = new->hist.histBgGC;

}




/* ARGSUSED */
static XtGeometryResult 
GeometryManagerHist(w, request, reply)
Widget w;
XtWidgetGeometry *request, *reply;
{
  HistogramConstraints hcons =(HistogramConstraints) (w->core.constraints);
  HistogramWidget hw = (HistogramWidget) XtParent(w);
  XtWidgetGeometry allowed;

  if (!hcons->form.allow_resize)
	return XtGeometryNo;
  allowed.height = request->height;
  allowed.width = request->width;
  if (allowed.height == w->core.height && allowed.width == w->core.width)
	return XtGeometryNo;
  
  w->core.width = allowed.width;
  w->core.height = allowed.height;
  if (!hw->hist.layoutPending) { 
  	ReLayout(w);
  }

  return XtGeometryYes;
}



static void ReLayout(childw)
Widget childw;
{
  HistogramWidget hw = (HistogramWidget) XtParent(childw);
  int diff_w, diff_h;
  Position graph_x, graph_y, time_x, time_y, bott_x, bott_y;
  Dimension graph_w, graph_h, time_w, time_h, yAxes_w, yAxes_h,
	    bott_w, bott_h;
  Widget gw = (hw->hist).graphW;
  Widget yw = (hw->hist).yAxesW;
  Widget tw = (hw->hist).timeAxisW;
  Widget bw = (hw->hist).bottomLabelW;
  HistogramConstraints hconsT, hconsG, hconsY, hconsB;

  hconsB = (HistogramConstraints) (hw->hist.bottomLabelW->core.constraints);
  hconsY = (HistogramConstraints) (hw->hist.yAxesW->core.constraints);
  hconsG = (HistogramConstraints) (hw->hist.graphW->core.constraints);
  hconsT = (HistogramConstraints) (hw->hist.timeAxisW->core.constraints);

  /* calculate diff width of yAxes and diff height of bottomLabel 
   * note, new size is in core->... but the old size in form virtual...
   * bottomLabel only request height change, yAxes only width change 
   */
  diff_h = HEIGHT(hw,bottomLabelW) - hconsB->form.virtual_height;
  diff_w = WIDTH(hw, yAxesW) - hconsY->form.virtual_width;

  if (diff_h == 0 && diff_w == 0) return;

  graph_h = hconsG->form.virtual_height = HEIGHT(hw,graphW) - diff_h;
  graph_w = hconsG->form.virtual_width = WIDTH(hw,graphW) - diff_w;
  time_h = hconsT->form.virtual_height = HEIGHT(hw,timeAxisW);
  time_w = hconsT->form.virtual_width = WIDTH(hw,timeAxisW) - diff_w;
  yAxes_w = hconsY->form.virtual_width = WIDTH(hw,yAxesW);
  yAxes_h = hconsY->form.virtual_height = HEIGHT(hw,yAxesW) - diff_h;
  bott_w = hconsB->form.virtual_width = WIDTH(hw,bottomLabelW);
  bott_h = hconsB->form.virtual_height = HEIGHT(hw,bottomLabelW);

  if (graph_w < MIN_GRAPH_WIDTH || graph_h < MIN_GRAPH_HEIGHT) {
      XtError("too many curves to display");
  }
  graph_x = gw->core.x + diff_w;
  graph_y = gw->core.y;
  time_x = graph_x - hw->hist.timeMarginWidth;
  time_y = tw->core.y - diff_h;
  bott_x = bw->core.x;
  bott_y = bw->core.y - diff_h;

  RefigureUnitWidth(hw, graph_w, graph_h);

#ifdef DEBUGHIST1
  fprintf(stderr,"diff_w = %d diff_h = %d\n",diff_w,diff_h);
  fprintf(stderr,"graph_x = %d graph_y = %d\n",graph_x,graph_y);
  fprintf(stderr,"time_x = %d time_y = %d\n",time_x,time_y);
  fprintf(stderr,"bott_x = %d bott_y = %d\n",bott_x,bott_y);
#endif

  if (diff_w < 0) { hw->hist.clearGraph = hw->hist.clearX = True; }
  else 	hw->hist.clearY = True;
  if (diff_h < 0) {
	hw->hist.clearY = hw->hist.clearGraph = True;
  }
  if (diff_w == 0) {
	XtConfigureWidget(bw,bott_x,bott_y,bott_w,bott_h,bw->core.border_width);
	XtResizeWidget(yw,yAxes_w, yAxes_h, yw->core.border_width);
	XtResizeWidget(gw,graph_w, graph_h, gw->core.border_width);
	XtMoveWidget(tw,time_x,time_y);
  } else if (diff_h == 0) {
	XtResizeWidget(yw,yAxes_w, yAxes_h, yw->core.border_width);
	XtConfigureWidget(gw,graph_x,graph_y,graph_w, graph_h, 
	    gw->core.border_width);
	hw->hist.clearX = True;
	XtConfigureWidget(tw,time_x,time_y,time_w,time_h,tw->core.border_width);
  } else {
	XtConfigureWidget(bw,bott_x,bott_y,bott_w,bott_h,bw->core.border_width);
	XtResizeWidget(yw,yAxes_w, yAxes_h, yw->core.border_width);
	XtConfigureWidget(gw,graph_x,graph_y,graph_w, graph_h, 
	    gw->core.border_width);
	XtConfigureWidget(tw,time_x,time_y,time_w,time_h,tw->core.border_width);
  }
  XFlush(XtDisplay(yw));

  /* to replace struct notify event */
  if (diff_w < 0) {
	hw->hist.clearY = True;
	RedisplayYAxes(yw, (XtPointer)False, NULL);
  } else if (diff_h > 0 && diff_w > 0) {
	hw->hist.clearX = hw->hist.clearY = hw->hist.clearGraph = True;
	RedisplayYAxes(yw, (XtPointer)False, NULL);
	RedisplayGraph(gw, (XtPointer)False, NULL);
	RedisplayTimeAxis(tw, (XtPointer)False, NULL);
  } else if (diff_h > 0) {
	hw->hist.clearGraph = hw->hist.clearY = True;
	RedisplayYAxes(yw, (XtPointer)False, NULL);
	RedisplayGraph(gw, (XtPointer)False, NULL);
  } else if (diff_w > 0) {
	hw->hist.clearGraph = hw->hist.clearX = True;
	RedisplayGraph(gw, (XtPointer)False, NULL);
	RedisplayTimeAxis(tw, (XtPointer)False, NULL);
  }
  else if (diff_h < 0){
	hw->hist.clearX = hw->hist.clearY = hw->hist.clearGraph = True;
	RedisplayYAxes(yw, (XtPointer)False, NULL);
        RedisplayGraph(gw, (XtPointer)False, NULL);
	RedisplayTimeAxis(tw, (XtPointer)False, NULL);
  }

}



static void RefigureUnitWidth(hw, w, h)
HistogramWidget hw;
int w; 	/* new width of graphW */
int h; 	/* new height of graphW */
{
   int i;

   RefigureTimeUnitWidth(hw, w);
   for (i=0; i< hw->hist.numGroup; i++) { 
	if (hw->hist.groups[i]->numShown)
	   ComputeGroupWidth(hw, hw->hist.groups[i]);
   }
}


static void RefigureTimeUnitWidth(hw, w)
HistogramWidget hw;
int w; /* new graph width */
{

   int multiple = 0, i;

   hw->hist.time.unitWidth = ((float)w/hw->hist.zoom)/(float)hw->hist.maxSample;
   if (hw->hist.zoom != 1.0) {
       if (hw->hist.zoom <= 0.1) multiple = 2;
       else if (hw->hist.zoom <= 0.5) multiple = 1;
   }
	
   hw->hist.time.interval = (hw->hist.time.max - hw->hist.startTime)/
       DEF_NUM_MARKS;
   if (hw->hist.time.interval > 1.0) 
       hw->hist.time.interval = (float)(int)((double)hw->hist.time.interval);

   for (i=1; i<= multiple; i++) {
   	hw->hist.time.interval /= 2.0;
   }

   return;
}




static void DestroyHist(Widget hw_)
{
   HistogramWidget hw = (HistogramWidget)hw_;
   int i;

   XtDestroyWidget(hw->hist.scrollW);
   XtDestroyWidget(hw->hist.zoomBarW);
   XtDestroyWidget(hw->hist.graphW);
   XtDestroyWidget(hw->hist.yAxesW);
   XtDestroyWidget(hw->hist.timeAxisW);
   XtDestroyWidget(hw->hist.bottomLabelW);
   XtDestroyWidget(hw->hist.ZOOMW);
   XtDestroyWidget(hw->hist.PANW);

   XFreeFont(XtDisplay(hw), hw->hist.fontInfo);

   XtReleaseGC((Widget)hw, hw->hist.histFgGC);
   XtReleaseGC((Widget)hw, hw->hist.histBgGC);

   if (hw->hist.patternGC) {
      for (i=0; i < hw->hist.numPatterns; i++) {
    	XFreeGC(XtDisplay(hw), hw->hist.patternGC[i]);
      }
      XtFree((void*)hw->hist.patternGC);
   }

   for (i=0; i< hw->hist.numGroup; i++)
	XtFree((void *) hw->hist.groups[i]);

   for (i=0; i< hw->hist.numCurve; i++)
      if(hw->hist.curves[i]){
	  XtFree((void *)hw->hist.curves[i]->data);
	  XtFree((void *)hw->hist.curves[i]->points);
	  XtFree((void *)hw->hist.curves[i]->name);
	  XtFree((void *) hw->hist.curves[i]);
      }

   XtFree((void *) hw->hist.groups);
   XtFree((void *) hw->hist.curves);

   return;
}



/********************************************************
 * following two exposure event handlers for two strings  
 * "ZOOM" and "PAN" redisplay 
 *******************************************************/
/* ARGSUSED */
static void 
RedisplayPAN(Pw, expose, event)
Widget Pw;
Boolean expose;
XExposeEvent *event;
{
   HistogramWidget hw = (HistogramWidget) XtParent(Pw);
   Display *disp = XtDisplay(Pw);
   Window win = XtWindow(Pw);
   int x, y;
   char buff[20];

   if (!XtIsRealized(Pw)) return;

   if (expose) event = GetExposeBox(event);
   XClearWindow(disp, win);

   x = WIDTH(hw,PANW)/2;
   y = hw->hist.charStruct.ascent;

   sprintf(buff, "%s", "PAN");
   XDrawString(disp, win, hw->hist.histFgGC, x, y, buff, strlen(buff));
}




/* ARGSUSED */
static void 
RedisplayZOOM(Zw, c_data, event)
Widget Zw;
XtPointer c_data;
XExposeEvent *event;
{
   HistogramWidget hw = (HistogramWidget) XtParent(Zw);
   Display *disp = XtDisplay(Zw);
   Window win = XtWindow(Zw);
   int x, y;

   if (!XtIsRealized(Zw)) return;

   XClearWindow(disp, win);

   x = 1;
   y = HEIGHT(hw,ZOOMW)/2 - 4*hw->hist.charHeight;

   XDrawString(disp, win, hw->hist.histFgGC, x, y, "Z", 1);
   XDrawString(disp, win, hw->hist.histFgGC, x, y+=hw->hist.charHeight+SPACE, "O", 1);
   XDrawString(disp, win, hw->hist.histFgGC, x, y+=hw->hist.charHeight+SPACE, "O", 1);
   XDrawString(disp, win, hw->hist.histFgGC, x, y+=hw->hist.charHeight+SPACE, "M", 1);
}





/**************************************************************
    exposure events and callback functions for child widget 
***************************************************************/

/* ARGSUSED */
static void RedisplayGraph(Widget gw, XtPointer expose_, XExposeEvent *event)
/* gw: graph widget */
{
     Boolean expose = (Boolean)expose_;
     HistogramWidget hw = (HistogramWidget) XtParent(gw);
     InternalCurveRec *curve;
     int i;
     int start = 0;
     int first_point;
     XPoint *points;
	
    if (!XtIsRealized(gw) || !hw->core.visible ) return;
    if (hw->hist.timeWidth == 0.0 || hw->hist.maxSample == 0) {
       /* time unit width is not set */
       return;
    }
    if (expose) event = GetExposeBox(event);

    hw->hist.oldWidth = hw->core.width;
    hw->hist.oldHeight = hw->core.height;

    if (event) {
     	start = MAX(0, event->x - hw->hist.time.unitWidth);
    }

    if (hw->hist.clearGraph) { 
	XClearWindow(XtDisplay(gw), XtWindow(gw));
	for (i=0; i< hw->hist.numCurve; i++) {
	    curve = hw->hist.curves[i];
	    if (curve && curve->data)
	       CalculateXPoints(hw, curve, 0, curve->npoints);
        }
	hw->hist.clearGraph = False;
	start = 0;
    }
    first_point = (int)((hw->hist.currXLeft + start)/hw->hist.time.unitWidth);
    /*
    if ((first_point > 1) && (hw->hist.currXLeft == 0)) first_point -= 2;
    */
    if ((first_point > 1)) first_point -= 2;
    /* make sure we get the last point also displayed */


    for (i = 0; i < hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
	if (curve && curve->npoints && curve->visible) {
    	    points = curve->points;
    	    DrawCurve(hw,gw, curve, first_point, curve->npoints);
	}
    }
    return;
}

/* ARGSUSED */
static void Redisplay2Graph(Widget gw, XtPointer expose_, XExposeEvent *event)
/* gw: graph widget */
{
     Boolean expose = (Boolean)expose_;
     HistogramWidget hw = (HistogramWidget) XtParent(gw);
     InternalCurveRec *curve;
     int i;
     int start = 0;
     int end; 
     int first_point;
     XPoint *points;
	
    if (!XtIsRealized(gw) || !hw->core.visible ) return;
    if (hw->hist.timeWidth == 0.0 || hw->hist.maxSample == 0) {
       /* time unit width is not set */
       return;
    }
    if (expose) event = GetExposeBox(event);

    hw->hist.oldWidth = hw->core.width;
    hw->hist.oldHeight = hw->core.height;

    if (event) {
     	start = MAX(0, event->x - hw->hist.time.unitWidth);
    }

    if (hw->hist.clearGraph) { 
	XClearWindow(XtDisplay(gw), XtWindow(gw));
	for (i=0; i< hw->hist.numCurve; i++) {
	    curve = hw->hist.curves[i];
	    if (curve && curve->data)
	       CalculateXPoints(hw, curve, 0, curve->npoints);
        }
	hw->hist.clearGraph = False;
	start = 0;
    }
    first_point = (int)((hw->hist.currXLeft + start)/hw->hist.time.unitWidth);
    if (first_point > 1) first_point -= 2; 
    /* make sure we get the last point also displayed */


    for (i = 0; i < hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
	if (curve && curve->npoints && curve->visible) {
    	    points = curve->points;
	    end =  curve->npoints;
	    if(event){
	       if(event->x == 0){
		  /* add 2 for overlap since subtracted 2 from first_point */
                  end = (int)((hw->hist.currXLeft + event->width)/
			hw->hist.time.unitWidth) +2;  
		  if (end > curve->npoints)
		      end =  curve->npoints;
	       }
	    }
	    if (curve && curve->data)
	       CalculateXPoints(hw, curve, 0, curve->npoints);
#ifdef DEBUGHIST
            fprintf(stderr,"fp = %d end = %d event.width = %d event.x = %d\n",
		    first_point,end,event->width,event->x);
#endif
    	    DrawCurve(hw,gw, curve, first_point, end);
	}
    }
    return;
}




/* assume curve->points and data are all filled up */
static void CalculateXPoints(hw, curve, first_point, npoints)
HistogramWidget hw;
InternalCurveRec *curve;
int first_point; /* the first pt to be calculated */
int npoints;	 /* how many points to be calculated */
{
   int i;
   int gID = curve->groupIndex; /* group index this curve belong to */
   XPoint *points = curve->points; 
   float  *data;
   float time_w;
   float y_w;
   int g_h;
   int left;
   float x;
   int last_point;


   if (curve->smoothData) {
	data = curve->smoothData;
   } else {
	data = curve->data;
   }

   time_w = hw->hist.time.unitWidth;
   y_w = hw->hist.groups[gID]->y.unitWidth;
   g_h = hw->hist.graphW->core.height; 
   left = hw->hist.currXLeft;

   x = (float)first_point * time_w - left;
   last_point = first_point + npoints;

   for (i = first_point; i < last_point; i++, x += time_w) {
	points[i].x = (short)((float)((x < 0)? x - 1.0 : x));
	if(isnan(data[i]))
	    points[i].y = 0;
        else
	    points[i].y = (short)((float)(g_h - data[i] * y_w));  
          
        /* make sure small values (and zero values) show up on graph */
        if ((!isnan(data[i])) && (points[i].y == g_h)) {
	    points[i].y = (short)(g_h - 1);
	}

   }
   return;
}

/*
 * Draws a 2x2 pixel rectangle about the point indicated by x,y
 * if (x,y) is on a boundry of the histogram window then it shifts the 
 * rectangle so that it show up in the window 
 */
static void DrawRectangle(hw, display, drawable, gc, x, y, point)
HistogramWidget hw;
Display *display;
Drawable drawable;
GC gc;
short x;
short y;
int point;
{
    int g_h = hw->hist.graphW->core.height;
    if(y == g_h) {
        y -= 2;
    }
    if(y == (g_h - 1)){
	y -=1;
    }
    /*
    XDrawRectangle(display, drawable,gc,x,y,2,2);
    */
    XFillRectangle(display, drawable,gc,x-1,y-1,4,4);
}

/*
 * Draw a curve, given graphW widget, gc to use
 * last_point : last point in this curve
 * first_point : staring from it to look for first pt to draw
 * start, end : two end points of exposed area
 */
static void DrawCurve(hw, gw, curve, first_point, last_point) 
HistogramWidget hw;
Widget gw;
InternalCurveRec *curve;
int last_point;
int first_point;
{
   Display *disp = XtDisplay(gw);
   Window  win = XtWindow(gw);
   GC gc = curve->gc;
   XPoint *points = curve->points;
   int temp;


   int last_valid = -1; 
   /* find and plot all valid segments */ 
   while(first_point < last_point){

      if(first_point != 0){
           /* if this is the first valid point, then draw endpoint */
          if((isnan(curve->data[first_point-1])) 
              && (!isnan(curve->data[first_point]))){
              DrawRectangle(hw, disp, win, gc,curve->points[first_point].x, 
            	            curve->points[first_point].y,first_point);
          }
          /* if this is the first nan point then draw and endpoint */
          else if((isnan(curve->data[first_point])) 
    	    && (!isnan(curve->data[first_point-1]))){ 
              DrawRectangle(hw, disp, win, gc,curve->points[first_point].x, 
	                    curve->points[first_point].y,first_point);
          }
      }
      if(!isnan(curve->data[first_point])){
          last_valid = first_point; 
      }

      /* find the first valid point */
      while((isnan(curve->data[first_point])) && (first_point < last_point)){
         first_point++;
      }

      /* if previous values were NaN then draw end point */
      if((first_point < last_point) && (first_point > last_valid)){
           DrawRectangle(hw, disp, win, gc,curve->points[first_point].x, 
		      curve->points[first_point].y,first_point);
           last_valid = first_point; 
      }

      for(temp = first_point; 
	 (!isnan(curve->data[temp])) && (temp < last_point); 
	 temp++) ;

      
      ASS_T(!isnan(curve->data[first_point]));
      ASS_LE(temp,last_point);
      ASS_LE(last_point,curve->npoints);

      if (first_point <= temp) {
          XDrawLines(disp, win, gc, &(points[first_point]),
		     temp - first_point, CoordModeOrigin);
      }
      /* if this is the end of a valid segment draw endpoint */
      if((temp < last_point) && (temp > 0)){
           DrawRectangle(hw, disp, win, gc,curve->points[temp -1].x, 
		      curve->points[temp - 1].y,first_point);
           last_valid = first_point; 
      }

      first_point = temp + 1;
      if(!isnan(curve->data[temp])){
          last_valid = temp; 
      }
      else {
          last_valid = temp-1;
      }
   }

    return;
}

#ifdef DEBUGHIST
static void Draw2Curve(gw, curve, gc, hw, first_point, last_point) 
Widget gw;
InternalCurveRec *curve;
GC gc;
HistogramWidget hw;
int last_point;
int first_point;
{
   Display *disp = XtDisplay(gw);
   Window  win = XtWindow(gw);
   XPoint *points = curve->points;
   int temp;


   /* find and plot all valid segments */ 
   while(first_point < last_point){

      /* find the first valid point */
      while((isnan(curve->data[first_point])) && (first_point < last_point)){
         first_point++;
      }
      
      for(temp = first_point; 
	 (!isnan(curve->data[temp])) && (temp < last_point); 
	 temp++) ;

      ASS_T(!isnan(curve->data[first_point]));
      ASS_LE(temp,last_point);
      ASS_LE(last_point,curve->npoints);

        if (first_point <= temp) {
	   XDrawLines(disp, win, gc, &(points[first_point]),
		   temp - first_point, CoordModeOrigin);
        }
      first_point = temp + 1;
   }

    return;
}
#endif


/* ARGSUSED */
static void RedisplayYAxes(yw, expose, event)
Widget yw;
Boolean expose;
XExposeEvent *event;
{
   HistogramWidget hw = (HistogramWidget) XtParent(yw);
   HistogramConstraints ycon = (HistogramConstraints) yw->core.constraints;
   int   i;
   Boolean y_line = False;
   Group *group;
   Position x = yw->core.width, 
	    y = T_MARGIN - hw->hist.numberHeight - hw->hist.charHeight;
   int start_x = -2, 
       /* when graphW gets smaller by XtResize() in Relayout()
	* it is not only set smaller,but I also call this redisplay 
	* function to replace the structurenotify event
	* sometimes, the redisplay call gets service first, then
	* it will get wrong size, here I thus call its virtual width
	* which is set to be correct in Relayout()
	*/
       end_x = ycon->form.virtual_width + 2,
       start_y = T_MARGIN - 2, 
       end_y = T_MARGIN + ycon->form.virtual_height + 2;

    if (!XtIsRealized(yw) || !hw->core.visible)
	return;
    if (hw->hist.timeWidth == 0.0 || hw->hist.maxSample == 0) {
       /* time unit width is not set */
       return;
    }
   if (expose) event = GetExposeBox(event);

   if (hw->hist.clearY) {
   	XClearWindow(XtDisplay(yw), XtWindow(yw));
   } else if (event) {
	start_x = event->x - 2;
	end_x = event->x + event->width + 2;
	if (event->y > T_MARGIN)  
		start_y = MAX(start_y, event->y - 2 - hw->hist.numberHeight);
	end_y = MIN(end_y, event->y + event->height + 2 + hw->hist.numberHeight);
   }
   for (i= 0; i< hw->hist.numGroup; i++) {
	group = hw->hist.groups[i];
	if (group->numShown && x >= start_x && x - group->needWidth <= end_x) {
	    DrawYAxis(hw, x, y, start_y, end_y,hw->hist.groups[i], y_line);
	    y_line = True;
	    x -= hw->hist.groups[i]->needWidth;
	}
    }
    hw->hist.clearY = False;
}


/* draw a block of y axis in widget yw with GC gc */
static void DrawYAxis(hw, x, y, start_y, end_y, group, y_line)
HistogramWidget hw;
Position x, y;  /* right-top x, y of markers for this block of axis */
int start_y, end_y;  /* exposure range of y pixel */
Group *group;   /*  ptr to the group info for this axis */
Boolean y_line;
{
     Widget yw = hw->hist.yAxesW;
     Display *disp = XtDisplay(yw);
     Window win = XtWindow(yw);
     GC gc = hw->hist.histFgGC;
     int x2 = x ; /*x-right-end of markers*/
     int x1 = x2 + 1 - MARK_WIDTH; /*x-left-end of marker */
     int y0 = T_MARGIN + HEIGHT(hw, graphW); /* pos of y-orig */
#ifdef notdef
     char buff[20];
     int len;
     float value;
     Dimension lab;  
     XFontStruct *fontInfo = hw->hist.fontInfo;
     float j = (float)y0; 
     float interval_width =  Y_INTERVAL_W(group->y);
#endif

     float val;
     int labelX;
     char buffer[255];
     int i, l, originY, currY;

    ComputeGroupYLabel(hw, group);
    originY = T_MARGIN + HEIGHT(hw, graphW);
    for (i=0; i <= group->y.count; i++) {
	 currY = originY - (int) ((double) (i * group->y.tickSpace));
	 XDrawLine(disp, win, gc, x1, currY, x2, currY);
	 val = group->y.interval * i;
	 (void)sprintf(buffer, group->y.format, val);
	 l = strlen(buffer);
	 labelX = x1 - XTextWidth(hw->hist.fontInfo, buffer, l) - SPACE;
	 /* !!!! here */
	 currY = currY - fontHeight/2 + hw->hist.fontInfo->max_bounds.ascent;
	 XDrawString(disp, win, gc, labelX, currY, buffer, l);
    }

    /* draw y title */
    XDrawString(disp, win, gc, x1 - group->titleWidth - SPACE, y, 
	 group->title, strlen(group->title));

    /* draw y axis line */
    if (y_line)
	XDrawLine(disp, win, gc, x + 1, T_MARGIN, x+1, y0);
    return;
}


/* ARGSUSED */
static void RedisplayTimeAxis(tw, expose, event)
Widget tw;
Boolean expose;
XExposeEvent *event;
{
   Display *disp = XtDisplay(tw);
   Window win = XtWindow(tw);
   HistogramWidget hw = (HistogramWidget) XtParent(tw);
   int start_x = hw->hist.timeMarginWidth - 4, 
       end_x = hw->hist.timeMarginWidth + WIDTH(hw,graphW) + 4, 
       end_y = HEIGHT(hw, timeAxisW);

   Label *time = &(hw->hist.time);

   char buff[20];
   int len, temp;

   float xL_value; /* left-most value */
   float xL_shown; /* left value shown */
   float value; 
   float multiple_f;
   int   pix_i; 
   float xL_pix_f, pix;
   int   multiple; 
   int y_lab = Y_POS_TIME_LAB(hw); /*y position of time value string*/
   float interval_width;
   int min, hour;
   int x, y;

   if (!XtIsRealized(tw) || !hw->core.visible) return;
    if (hw->hist.timeWidth == 0.0 || hw->hist.maxSample == 0) {
       /* time unit width is not set */
       return;
    }
   if (expose) event = GetExposeBox(event);

   if (hw->hist.time.max <= MAX_SEC) {
       hw->hist.timeUnit = TIME_UNIT_SEC;
   } else if (hw->hist.time.max <= MAX_MIN) {
       hw->hist.timeUnit = TIME_UNIT_MIN;
   } else {
       hw->hist.timeUnit = TIME_UNIT_HOUR;
   }
   if (hw->hist.clearX) {
   	XClearWindow(XtDisplay(hw->hist.timeAxisW), 
	    XtWindow(hw->hist.timeAxisW));
	hw->hist.clearX = False;
   } else if (event) {
	start_x = event->x - 4;
	end_x = MIN(end_x, event->x + event->width + 4);
	end_y = event->y + event->height ;
   }

   xL_value =  hw->hist.startTime + (float)hw->hist.currXLeft * hw->hist.timeWidth / time->unitWidth;
   multiple_f =  xL_value/time->interval;
   multiple = ceil((double)multiple_f);

   xL_shown = multiple * time->interval;
   interval_width = time->unitWidth * time->interval / hw->hist.timeWidth;
   xL_pix_f = (xL_shown - xL_value) * time->unitWidth / hw->hist.timeWidth + (float)hw->hist.timeMarginWidth;

   if (hw->hist.timeWidth <= 0) return;

   for (value = xL_shown, pix = xL_pix_f; pix < start_x;
		value += time->interval, pix += interval_width); 

    for ( ; pix <= end_x; value += time->interval, pix += interval_width) {
	pix_i = (int)((double)pix); 
	XDrawLine(disp, win, hw->hist.histFgGC, pix_i, 0, pix_i,MARK_WIDTH);
	    if (hw->hist.timeUnit == TIME_UNIT_SEC) {
		sprintf(buff, "%.0f", value);
	    } else if (hw->hist.timeUnit == TIME_UNIT_MIN) {
		min = (int)(value/60.0);
		sprintf(buff, "%d:%02d", min, (int)(value - (float)min * 60.0));
	    } else {
		hour = (int)(value/(3600.0));
		min = (int)((value - hour*3600.0)/60.0);
		sprintf(buff, "%d:%02d", hour, min);
	    }
	    temp = XTextWidth(hw->hist.fontInfo, buff, (len = strlen(buff)));
	    XDrawString(disp, win, hw->hist.histFgGC, 
		pix_i - (int)((temp+1)/2), (Position)y_lab, buff, len);
    }

    memset(buff, '\0', sizeof(buff));
    switch (hw->hist.timeUnit) {
	case TIME_UNIT_SEC:
	    strcpy(buff, STRING_SEC);
	    break;
	case TIME_UNIT_MIN:
	    strcpy(buff, STRING_MIN_SEC);
	    break;
	case TIME_UNIT_HOUR:
	    strcpy(buff, STRING_HR_MIN);
	    break;
	default:
	    XtError("Invalid time Unit");
	    exit(-1);
    }
    len = strlen(buff);

    y = Y_POS_TIME_UNIT(hw);
    x = WIDTH(hw, graphW)/2;
    if ( y - hw->hist.charStruct.ascent < end_y &&
       	       (x <= end_x && x + len >= start_x) ) { 
   	XDrawString(disp, win, hw->hist.histFgGC, x, y, buff, len);
    }
    return;
}





/*
 * Get all the pending expose events for the named window and create a 
 *   single expose event.
 * This is similar to exposure compression in a widget.
 *
 */
static XExposeEvent *GetExposeBox(event)
XExposeEvent *event;
{
    XEvent new;
    XExposeEvent *ex;
    short lx, ly, hx, hy;
    XWindowAttributes att;

    if (!event) return(NULL);
    lx = event->x;
    ly = event->y;
    hx = event->x + event->width;
    hy = event->y + event->height;

    XFlush(event->display);

XGetWindowAttributes(event->display, event->window, &att);
    
    while (XCheckWindowEvent(event->display, event->window, 
	ExposureMask, &new) == True) {
	ex = (XExposeEvent *) &new;
	lx = MIN(lx, ex->x);
	ly = MIN(ly, ex->y);

/*!!!!!!!!!!?????????*/
if (ex->width > att.width) {
/*!!!!!!!!!!!! a potential error, get box gets width greater than current width,
 * I just ignore this error and set its expose width to be curr width

fprintf(stderr, "GetExposeBox() error; get width = %d; change to width = %d\n", ex->width, att.width);
*/
ex->width = att.width;
}
	hx = MAX(hx, ex->x + ex->width);
	if (hx < lx) abort();
	hy = MAX(hy, ex->y + ex->height);
	memset((char*)&new, '\0', sizeof(new));
    }
    event->x = lx;
    event->y = ly;
    if (hx < lx) abort();
    event->width = hx - lx;
    event->height = hy - ly;
    return(event);
}





/* ARGSUSED */
static void 
ChangeGraphVisible(gw, client_data, event) 
Widget gw;
XtPointer client_data;
XVisibilityEvent *event;
{
    HistogramWidget hw = (HistogramWidget) client_data;

    hw->hist.visibility = event->state;
}



/******************************************************
  our callbacks for horizontal scrollbar and zoombar 
 ******************************************************/

static void ZoomHistogram(Widget zw, 
			  XtPointer client_data, 
			  XtPointer call_data)
{
   HistogramWidget hw = (HistogramWidget) client_data;
   float percent = *(float*) call_data;
#ifdef notdef
   XExposeEvent event;
   int unit;
#endif

    float sbArea;
    float new_zoom, diff;

    /* SB doesn't prevent scrolling off the end */
    sbArea = getShown(zw);
    if ((sbArea + percent) > 1.0) {
	percent = 1.0 - sbArea;
	XawScrollbarSetThumb(zw, percent, sbArea);
    }
    new_zoom = percent + sbArea;
    if ((diff = new_zoom - hw->hist.zoom) > EPSILON || diff < -EPSILON ) {
	hw->hist.zoom = new_zoom;
	hw->hist.start = hw->hist.focus - (hw->hist.zoom / 2.0);
	if ((hw->hist.zoom + hw->hist.start) > 1.0)
	    hw->hist.start = 1.0 - hw->hist.zoom; 
	hw->hist.start = (0.0 > hw->hist.start) ? 0.0 : hw->hist.start;
	hw->hist.currXLeft = (hw->hist.start * ((double)hw->hist.graphW->core.width / hw->hist.zoom));
	XawScrollbarSetThumb(hw->hist.scrollW, hw->hist.start, hw->hist.zoom);
	RefigureTimeUnitWidth(hw, hw->hist.graphW->core.width);

	/* pseudo exposure to redisplay timeAxisW & graphW */
	QUICK_DISPLAY_TIME_AXIS(hw,unit,event);
	hw->hist.clearGraph= True;
	RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
    }
return;
}



static void ZoomScrollFunc(zw, client_data, call_data)
Widget zw;
XtPointer client_data;
int call_data;
{
   HistogramWidget hw = (HistogramWidget) client_data;
   int position = (int) call_data;
   float sbArea, change;
   float curr_position, new_position, diff;

    sbArea = getShown(zw);
    curr_position = hw->hist.zoom - sbArea;
    change = (((float) position) / (float)zw->core.height) * sbArea;
    new_position = curr_position + change;

    /* SB doesn't prevent scrolling off the end */
    if ((new_position + sbArea) > 1.0) {
	new_position = 1.0 - sbArea;
        XawScrollbarSetThumb(zw, new_position, sbArea);
    }

    if (new_position < 0.0) new_position = 0.0;
    if ((diff = new_position - curr_position) > EPSILON ||
	 diff < -EPSILON ) {
        XawScrollbarSetThumb(zw, new_position, sbArea);
	ZoomHistogram(zw, client_data, (XtPointer) &new_position);
    }
   return; 
} 



static void  HJumpFunc(sw, client_data, call_data)
Widget sw;
XtPointer client_data, call_data;
{
   HistogramWidget hw = (HistogramWidget) client_data;
   float percent = *(float*) call_data;
   int old_left;
   float sbArea;
   float diff;

   sbArea = getShown(sw);
   /* SB doesn't prevent scrolling off the end */
   if ((sbArea + percent ) > 1.0) {
 	percent = 1.0 - sbArea;
	XawScrollbarSetThumb(sw, percent, sbArea);
   }

   if ((diff = hw->hist.start - percent) > EPSILON || diff < -EPSILON) {
	hw->hist.start = percent;
	old_left = hw->hist.currXLeft;
	hw->hist.currXLeft = (int) (hw->hist.start * ((float)hw->hist.graphW->core.width / hw->hist.zoom));
	ScrollHistogram(hw, old_left);
   }
   return;
}


static void HScrollFunc(sw, client_data, call_data)
Widget sw;
XtPointer client_data;
int call_data;
{
   HistogramWidget hw = (HistogramWidget) client_data;
   int position =  call_data;
   int old_left;
   float sbArea;
   float change;
   float cur_position;
   float new_position, diff;

   sbArea = getShown(sw);
   cur_position = hw->hist.start;
   change = (((float) position) / (float)hw->hist.graphW->core.width) * sbArea;
   new_position = cur_position + change;
   if ((new_position + sbArea) > 1.0) {
       new_position = 1.0 - sbArea;
       XawScrollbarSetThumb(sw, new_position, sbArea);
   }
   if (new_position < 0.0) new_position = 0.0;
   if ((diff= new_position - cur_position) > EPSILON || diff < -EPSILON) {
       XawScrollbarSetThumb(sw, new_position, sbArea);
	hw->hist.start = new_position;
	old_left = hw->hist.currXLeft;
	hw->hist.currXLeft = hw->hist.start * (float)hw->hist.graphW->core.width / 
	     hw->hist.zoom;
	ScrollHistogram(hw, old_left);
   }
    return ;
}


static void ScrollHistogram(hw, old_left)
HistogramWidget hw;
int old_left;
{
    int i, j;
    int moved, overlap;
    Widget gw = hw->hist.graphW;
    InternalCurveRec *curve;

#ifdef notdef
    XExposeEvent event;
    int unit;
#endif

    /* time axis is affected */
    QUICK_DISPLAY_TIME_AXIS(hw,unit,event);

    moved = hw->hist.currXLeft - old_left;
    overlap = gw->core.width - fabs((double)moved);

    /* shift x points according to moved */ 
    for (i=0; i< hw->hist.numCurve; i++) { 
	curve = hw->hist.curves[i];
	if(curve){
	  for (j=0; j <= curve->npoints; j++) {
	     curve->points[j].x -= moved;
          }
	}
    }

    /*
     * If we are not un-obscured, then we can't just copy the bits.
     *
     */
    if ((overlap < 1)|| (hw->hist.visibility != VisibilityUnobscured)) {
	/* nothing left to copy */
	XClearWindow(XtDisplay(gw), XtWindow(gw));
	RedisplayGraph(gw, (XtPointer)False, NULL);
    } else {
	ShiftGraph(hw, moved, overlap);
    }
    hw->hist.focus = hw->hist.start + (hw->hist.zoom / 2.0);
   return;
}



static void ShiftGraph(hw, moved, overlap)
HistogramWidget hw;
int moved;
int overlap;
{
    XExposeEvent event;
    Widget gw = hw->hist.graphW;
    Display *disp = XtDisplay(gw);
    Window win = XtWindow(gw);

    if (moved > 0) {
	XCopyArea(disp, win, win, hw->hist.histFgGC, moved, 0, overlap, 
	    gw->core.height, 0, 0);
	XClearArea(disp, win, overlap, 0, moved, gw->core.height, False);
	event.x = overlap;
	event.width = moved; 
	RedisplayGraph(gw, (XtPointer)False, &event);
	XFlush(disp);

    } else {
	XCopyArea(disp, win, win, hw->hist.histFgGC, 0, 0, overlap,
	    gw->core.height, -moved, 0);
	XClearArea(disp, win, 0, 0, -moved, gw->core.height, False);
	XFlush(disp);
	event.x = 0;
	event.width = -moved; 
#ifdef DEBUGHIST
        fprintf(stderr,"event.width = %d event.x = %d\n",
		event.width,event.x);
#endif
	Redisplay2Graph(gw, (XtPointer)False, &event);
    }
}


/*************************************************************
	  Histogram Convience functions 
 *************************************************************/

void HistAddNewCurve(w, curve_info_mask, curve_rec, shown)
Widget w; 
unsigned long curve_info_mask;
CurveRec *curve_rec;
Boolean shown;
{
  int 	i, j, groupIndex, curveID = -1, len;
  Boolean newGroup = False;
  XGCValues gcv;
  CurveInfo *cInfo = &(curve_rec->curveInfo);
  InternalCurveRec *curve;
  HistogramWidget hw = (HistogramWidget) w;
  int found = 0;
  Widget prev;
  int oldmax;


   ASS_LE(hw->hist.numShownCurve, hw->hist.numCurve);
   ASS_LE(hw->hist.numCurve, hw->hist.maxCurve);

  if (hw->hist.timeWidth <= 0) {
	char message[80];
	sprintf(message, "rtIPS error: Invalid bucket width %f", 
	    hw->hist.timeWidth);
	XtError(message);
  }

  /* if there is a vacant slot in curves array use it for new curve */
  for(i=0;i< hw->hist.maxCurve; i++){
    if(!hw->hist.curves[i]){
      curveID = i;
      found = 1;
      break;
    }
  }
  hw->hist.layoutPending = True;
  if (!found ) {
       hw->hist.curves = (InternalCurveRec **)XtRealloc((void *)hw->hist.curves,
	     (hw->hist.maxCurve + DEF_INC_CURVE) * sizeof(InternalCurveRec *));
       oldmax =  hw->hist.maxCurve; 
       hw->hist.maxCurve += DEF_INC_CURVE;
       for (j=oldmax; j< hw->hist.maxCurve; j++)
		hw->hist.curves[j] = NULL;

  }
	
  for (i=0; i < hw->hist.numGroup; i++) {
	if (hw->hist.groups[i]->groupID == curve_rec->groupID) 
		break;
  }
  if ((hw->hist.numGroup == hw->hist.maxGroup) && (i == hw->hist.numGroup)) { 
	hw->hist.maxGroup += DEF_INC_GROUP;
	hw->hist.groups = (Group **) XtRealloc((void *) hw->hist.groups, 
	    hw->hist.maxGroup * sizeof(Group *));
  }
  if (i < hw->hist.numGroup) { 
  	hw->hist.groups[i]->numMember++;
  	if(shown && ++(hw->hist.groups[i]->numShown) == 1)
		AddYAxis(w, i);
  } else  {
  	/* a new group */
	newGroup = True;
  	hw->hist.groups[i] = (Group *) XtCalloc(1, sizeof(Group));
	AddNewGroup(w, i, curve_info_mask, curve_rec);
	hw->hist.groups[i]->numShown = (shown) ? 1 : 0;
  	(hw->hist.numGroup)++;
  }
  if(!found || (hw->hist.numCurve <= curveID)){
    if (shown) hw->hist.numShownCurve++;
    groupIndex = i;
    if(!found)
      curveID = hw->hist.numCurve;
    hw->hist.numCurve++;
  } else {
    groupIndex = i;
    if (shown) hw->hist.numShownCurve++;
  }

  hw->hist.curves[curveID] = (InternalCurveRec *) 
      XtMalloc(sizeof(InternalCurveRec));
  curve = hw->hist.curves[curveID];
  curve->data = (float *) XtCalloc(hw->hist.maxSample + 1, sizeof(float));
  curve->points = (XPoint *) XtCalloc(hw->hist.maxSample+1, sizeof(XPoint));
  curve->smoothData = NULL;
  curve->smooth = False;
  curve->npoints = 0;
  curve->curveID = curveID;
  curve->groupIndex = groupIndex;
  if (shown) curve->visible = True;
  else 	curve->visible = False;
  curve->select = False;
  len = strlen(curve_rec->name) + 1;
  curve->name = (char *) XtMalloc((unsigned)len);
  strcpy(curve->name, curve_rec->name);
  curve->nameWidth = XTextWidth(hw->hist.fontInfo,curve_rec->name,len);


  if (curve_info_mask & LINEPROPERTY) {
  	gcv.foreground = cInfo->foreground;
	gcv.line_style = cInfo->lineStyle;
	gcv.line_width = cInfo->lineWidth;
  	curve->gc = XCreateGC(XtDisplay(hw), XtWindow(hw->hist.graphW), 
	    GCForeground | GCLineStyle | GCLineWidth , &gcv);
  } else if (shown) {
	curve->color_inPattern = NextLineStyleGCInd(hw,True);
	curve->color_gc = hw->hist.patternGC[curve->color_inPattern];
	curve->bw_inPattern = NextLineStyleGCInd(hw,False);
	curve->bw_gc = hw->hist.bw_patternGC[curve->bw_inPattern];
	if(hw->hist.defaultGC){
	    curve->gc = curve->color_gc;
	}
	else {
	    curve->gc = curve->bw_gc;
	}
  } else {
	curve->color_inPattern = -1;
	curve->bw_inPattern = -1;
	curve->gc = NULL;
  }

  if (shown && newGroup) {
	AddYAxis(w, groupIndex);
  }

  hw->hist.layoutPending = False;
  if (shown) AddBottomLabel(w, curveID);

  prev = NULL;
   for (i=0; i < hw->hist.numCurve; i++) {
       if (hw->hist.curves[i]) {
           XtVaSetValues(hw->hist.curves[i]->lineW, XtNfromVert, prev, NULL);
           XtVaSetValues(hw->hist.curves[i]->labelW, XtNfromVert, prev, NULL);
	   prev = hw->hist.curves[i]->lineW;
       }
  }

  curve_rec->curveID = curveID;

   ASS_LE(hw->hist.numShownCurve, hw->hist.numCurve);
   ASS_LE(hw->hist.numCurve, hw->hist.maxCurve);
   ASS_LE(curveID, hw->hist.numCurve);

  return;
}



static void AddNewGroup(hw, ind, mask, curve_rec)
HistogramWidget hw;
int ind;
unsigned long mask;
CurveRec *curve_rec;
{
  CurveInfo *cInfo = &(curve_rec->curveInfo);
  Group *group = hw->hist.groups[ind];
  XFontStruct *fontInfo = hw->hist.fontInfo;
  int len;

  group->numMember = 1;
  group->groupID = curve_rec->groupID;
  if (mask & YTITLE) { 
  	group->title = (char *) XtMalloc((unsigned) strlen(cInfo->yTitle) + 1);
  	strcpy(group->title, cInfo->yTitle);
  } else { 
  	group->title = (char *) XtMalloc((unsigned) strlen(curve_rec->name) + 1);
  	strcpy(group->title, curve_rec->name);
  }
  len = strlen(group->title);
  group->titleWidth = XTextWidth(fontInfo, group->title, len);

  group->y.max = EPSILON;
  group->y.interval = group->y.max/DEF_NUM_Y_MARKS;
  if (mask & YINC){
	group->yInc = cInfo->yInc;
  }
  else {
	group->yInc = hw->hist.yInc;
  }

  ComputeGroupWidth(hw, group);

  return;
}

static int NextLineStyleGCInd(hw,color) 
HistogramWidget hw;
Boolean color;
{
   int i;
   int best = 0;

   if(color){ /* use color patterns */
       for (i=0; i< hw->hist.numPatterns; i++) {
	    if (hw->hist.patternUsed[best] > hw->hist.patternUsed[i]) {
	        best = i;
	    }
       }
       (hw->hist.patternUsed[best])++;
   }
   else {
       for (i=0; i< hw->hist.bw_numPatterns; i++) {
	    if (hw->hist.bw_patternUsed[best] > hw->hist.bw_patternUsed[i]) {
	        best = i;
	    }
       }
       (hw->hist.bw_patternUsed[best])++;
   }
   return(best);
}


static void FreeALineStyle(hw, cID)
HistogramWidget hw;
int cID;	/* curve ID */
{
   (hw->hist.patternUsed[hw->hist.curves[cID]->color_inPattern])--;
   (hw->hist.bw_patternUsed[hw->hist.curves[cID]->bw_inPattern])--;
   hw->hist.curves[cID]->color_inPattern = -1;
   hw->hist.curves[cID]->bw_inPattern = -1;
   return;
}

static void AddYAxis(w, gIndex)
HistogramWidget w;
int gIndex;
{
    /* calculate needed width for yAxes widget */
    w->hist.yAxesWidth += w->hist.groups[gIndex]->needWidth;

    if (w->hist.yAxesWidth > (int)WIDTH(w,yAxesW)) { 
      	XtMakeResizeRequest(w->hist.yAxesW, w->hist.yAxesWidth, 
	    w->hist.yAxesW->core.height, NULL, NULL);
    } else {
	DrawYAxis(w, WIDTH(w,yAxesW) - w->hist.yAxesWidth - 1, 
	    T_MARGIN - w->hist.numberHeight - w->hist.charHeight, T_MARGIN - 2, 
	    HEIGHT(w, yAxesW), w->hist.groups[gIndex], 
	    (Boolean)(w->hist.numShownCurve-1)); 
    }
    return;
}


static void AddBottomLabel(hw, curveID)
HistogramWidget hw;
int curveID;
{
    Arg arg[20];
    Widget prev;
    int i;
    InternalCurveRec *curve = hw->hist.curves[curveID]; 
    int label_h = LABEL_HEIGHT(hw);

    if(!curve){
      return;
    }
    prev = NULL;
    for (i=curveID-1; i >= 0; i--) {
	if (hw->hist.curves[i]) { 
	    prev = hw->hist.curves[i]->lineW;
	    break;
	}
    }

    /* create line left to the label */
    i = 0;
    XtSetArg(arg[i], XtNwidth, BITMAP_LINE_WIDTH);i++;
    XtSetArg(arg[i], XtNheight, label_h);i++;
    XtSetArg(arg[i], XtNborderWidth, 0); i++;
    XtSetArg(arg[i], XtNfromVert, prev); i++;

    curve->lineW = XtCreateManagedWidget(curve->name, coreWidgetClass, 
	hw->hist.bottomLabelW, arg, i);
    RedisplayLine(curve->lineW, (XtPointer)curve, NULL);
    XtAddEventHandler(curve->lineW, ExposureMask, FALSE, RedisplayLine, 
	(XtPointer)curve);

    /* create toggle widget for curve label */
    i = 0;
    XtSetArg(arg[i], XtNresizable, True);i++;
    XtSetArg(arg[i], XtNheight, label_h);i++;
    XtSetArg(arg[i], XtNfont, hw->hist.fontInfo); i++;
    XtSetArg(arg[i], XtNforeground, hw->hist.foreground); i++;
    XtSetArg(arg[i], XtNbackground, hw->core.background_pixel); i++;
    XtSetArg(arg[i], XtNlabel, curve->name); i++;
    XtSetArg(arg[i], XtNborderWidth, 0); i++;
    XtSetArg(arg[i], XtNjustify, XtJustifyLeft); i++;
    XtSetArg(arg[i], XtNradioData, curve->name); i++;
    XtSetArg(arg[i], XtNstate, False); i++;
    XtSetArg(arg[i], XtNresize, True); i++;
    XtSetArg(arg[i], XtNfromHoriz, curve->lineW); i++;
    XtSetArg(arg[i], XtNfromVert, prev); i++;

    curve->labelW = XtCreateManagedWidget(curve->name, toggleWidgetClass, 
	hw->hist.bottomLabelW, arg, i);

    XtAddCallback(curve->labelW, XtNcallback, LabelToggle, (XtPointer) curve);

#ifdef notdef
    prev = NULL;
    for (i=0; i < hw->hist.numCurve; i++) {
       if (hw->hist.curves[i] 
	   && hw->hist.curves[i]->visible 
	   && (hw->hist.curves[i]->lineW)) {
           XtVaSetValues(hw->hist.curves[i]->lineW,XtNfromVert,prev,NULL);
           XtVaSetValues(hw->hist.curves[i]->labelW,XtNfromVert,prev,NULL);
	   prev = hw->hist.curves[i]->lineW;
       }
    }
#endif
    return;
}


/* ARGSUSED */
static void LabelToggle(lw, client_data, call_data)
Widget lw;
XtPointer client_data, call_data;
{
   InternalCurveRec *curve = (InternalCurveRec *) client_data;

   curve->select = !(curve->select);
   return;

}


/* ARGSUSED */
static void RedisplayLine(lw, c_data, event)
Widget lw; /* curve line widget */
XtPointer c_data;
XExposeEvent *event;
{
     InternalCurveRec *curve = (InternalCurveRec *) c_data;
     Display *disp = XtDisplay(lw);
     Window win = XtWindow(lw);
     int label_h = lw->core.height;

     if (!XtIsRealized(lw)) return;

     if (curve->bw_inPattern < 0) {
	XtWarning("curve->gc undefined");
	return;
     }
     if(curve->visible){
        XDrawLine(disp, win, curve->gc, 
          	0, label_h/2, BITMAP_LINE_WIDTH, label_h/2);
     }
     else {
        XDrawLine(disp, win, invisible_gc, 
          	0, label_h/2, BITMAP_LINE_WIDTH, label_h/2);
     }
      
}

static void HistShiftLeft(hw, shiftCount)
HistogramWidget hw;
int shiftCount;
{
    int i, j;
    int moved;
    int overlap;
    float timePass; /* time discarded */
    InternalCurveRec *curve;
    Widget gw = hw->hist.graphW;
#ifdef notdef
    int timeUnit;
    float *dataTemp;
    XPoint *pointTemp;
    XExposeEvent event;
    int old_npoints, new_npoints, min_npoints = hw->hist.maxSample;
#endif

    if (shiftCount > hw->hist.maxSample) {
	for (i=0; i< hw->hist.numCurve; i++) {
	    if(hw->hist.curves[i]) {
	        hw->hist.curves[i]->npoints = 0;
	        memset(hw->hist.curves[i]->data, '\0', 
		       sizeof(float) * hw->hist.maxSample);
	        memset(hw->hist.curves[i]->points, '\0',
		       sizeof(float)*hw->hist.maxSample);
	    }
        } 
        XClearWindow(XtDisplay(hw->hist.graphW),XtWindow(hw->hist.graphW));
    } else {
	timePass = (float) (shiftCount) * hw->hist.timeWidth;

	for (i=0; i < hw->hist.numCurve; i++) {
	    curve = hw->hist.curves[i];
	    if(curve){
	      /* cut curve->npoints by *shiftCount */
	      curve->npoints -= shiftCount;
	      if (curve->npoints  < 0) curve->npoints = 0;

	      memcpy((char*)(curve->data),(char*)&(curve->data[shiftCount])
		     ,(hw->hist.maxSample - shiftCount)* sizeof(float));

	      memset(&(curve->data[hw->hist.maxSample - shiftCount]), '\0',
		     sizeof(float) * shiftCount);
	      if (curve->smooth) {
	         memcpy((char*)(curve->smoothData), 
		     (char*)&(curve->smoothData[shiftCount]), 
		     (hw->hist.maxSample - shiftCount)* sizeof(float));
	         memset(&(curve->smoothData[hw->hist.maxSample-shiftCount])
			,'\0', sizeof(float) * shiftCount);
	      }
	      if (!curve->visible) continue;

	      /* shift left the points content */
	      moved = curve->points[shiftCount].x - curve->points[0].x;
	      memcpy((char*)(curve->points), 
		  (char*)&(curve->points[shiftCount]), 
		  (hw->hist.maxSample - shiftCount)* sizeof(XPoint));
	      memset(&(curve->points[hw->hist.maxSample - shiftCount]), '\0',
		sizeof(float) * shiftCount);
	      for (j = 0; j < curve->npoints; j++) {
		if (curve->visible) curve->points[j].x -= moved;
	      }
            }
	}
	/* moved = shiftCount * hw->hist.time.unitWidth; */
	overlap = gw->core.width - moved;
	if ((hw->hist.visibility != VisibilityUnobscured) || (overlap < 1)) {
	    XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
	    RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
	} else {
	    ShiftGraph(hw, moved, overlap);
	}
   }
   QUICK_DISPLAY_TIME_AXIS(hw, timeUnit, event);

   return;
}

void HistSampleInfo(w, timeWidth, maxSample, firstSampleTime, salvage)
Widget w;
double timeWidth;
int maxSample;
double firstSampleTime;
Boolean salvage;  /* retain and move old data when firstSampleTime changes */
{
   int i;
   int shiftCount;
#ifdef notdef
   int unit;
   XExposeEvent event;
#endif
   InternalCurveRec *curve;
   Boolean changeTimeBase = True;
   HistogramWidget hw = (HistogramWidget) w;

    if (timeWidth <= 0.0)
	XtError("time width can't be <= 0");
    if (maxSample <= 0)
	XtError("max samples can't be <= 0");

    if ((hw->hist.timeWidth == timeWidth) && 
	(hw->hist.maxSample == maxSample) &&
	(hw->hist.startTime == firstSampleTime))
	return;

    if ((hw->hist.timeWidth == timeWidth) &&
	(hw->hist.startTime == firstSampleTime)) {
	changeTimeBase = False;
    }

    hw->hist.timeWidth = timeWidth;
    hw->hist.maxSample = maxSample;

    /* handle starting time */
    if (hw->hist.startTime != firstSampleTime) { 
	if (firstSampleTime < hw->hist.startTime) {
	    XtError("invalid start time");
	    return;
	}
	shiftCount = (firstSampleTime - hw->hist.startTime)/ hw->hist.timeWidth;
	hw->hist.startTime = firstSampleTime;
    }
    hw->hist.time.max = hw->hist.startTime + timeWidth * (double)maxSample; 
  
    RefigureTimeUnitWidth(hw, WIDTH(hw,graphW));

    /* invalidate all curves shown in the graph window 
     * if timeWidth not change, redisplay all curves ourselves
     * otherwise, just wait for user to set curve data again
     */
    if (hw->hist.numShownCurve) {
	if (!changeTimeBase) {
	   hw->hist.clearGraph = True;
	   RedisplayGraph(hw->hist.graphW, False,NULL);
  	} else if (salvage) {
	   HistShiftLeft(hw, shiftCount);
	} else {
	   for (i=0; i< hw->hist.numCurve; i++) {
		curve = hw->hist.curves[i];
		if(curve){
		   curve->npoints = 0;
		   memset((char*)curve->data, '\0',
		       (hw->hist.maxSample) * sizeof(float));
	   	   memset((char*)curve->points, '\0',
		       (hw->hist.maxSample) *sizeof(XPoint));
		   if (curve->smooth) {
		      memset((char*)curve->smoothData, '\0',
		          hw->hist.maxSample+1*sizeof(float));
		   }
		}
	   }
	   XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
	}
	QUICK_DISPLAY_TIME_AXIS(hw,unit,event);
    }

    return;
}



/* process a new added segment belonging to thisCurve 
 * the new segment could be the whole curve 
 */
void ProcessNewSegment(w, thisCurve, first_point, change)
Widget w;
InternalCurveRec *thisCurve;
Boolean change; /* yMax is changed or not due to this new segment */
int first_point; /* the first position in thisCurve for new segment*/
{
   int i;
   InternalCurveRec *curve;
   int gID = thisCurve->groupIndex;
   HistogramWidget hw = (HistogramWidget) w;

    if (change) {
       for(i=0; i< hw->hist.numCurve; i++) {
	 curve = hw->hist.curves[i];
	 if(curve){
	   if (hw->hist.curves[i]->groupIndex == gID) {
	       CalculateXPoints(hw, curve, 0, curve->npoints);
	   }
         }
        }
	XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
	RedisplayGraph(hw->hist.graphW, (XtPointer) False, NULL);
    } else { 
	/* yMax is unchanged, only draw this new segment */
	CalculateXPoints(hw, thisCurve, 0, thisCurve->npoints);
	/* refresh connecting point too */
	if (first_point > 0) {
	    first_point -= 1;
	}
	DrawCurve(hw,(Widget)hw->hist.graphW, thisCurve, first_point,
	    thisCurve->npoints);
    }
}



void HistSetCurveData(w, curveID, firstSample, numSamples, sample)
Widget w;
int curveID;
int firstSample;
int numSamples;
const float *sample;
{
   HistogramWidget hw = (HistogramWidget) w;
   int groupIndex;
   Group *group; 
   InternalCurveRec *curve;
   Boolean change = False;
   int i;

   if(!hw->hist.curves[curveID]){
     return;
   }

   groupIndex = (hw->hist.curves[curveID])->groupIndex;
   group = hw->hist.groups[groupIndex];
   curve = hw->hist.curves[curveID];

   if (firstSample+numSamples > (hw->hist).maxSample) { 
	XtError("exceeding specified max sample number");
   }

   memset((char*)&curve->points[firstSample],
	    '\0', numSamples*sizeof(XPoint));
   memset((char*)&curve->data[firstSample],
	    '\0', numSamples*sizeof(float));
   for (i=0; i < numSamples; i++) {
       curve->data[firstSample + i] = sample[i]; 
   }
   if (firstSample + numSamples > curve->npoints) {
       curve->npoints = firstSample + numSamples;
   }

   if (curve->smooth)  {
      memset((char*)curve->smoothData, '\0', 
	  (hw->hist.maxSample + 1)*sizeof(float));
      SmoothData(hw, curve, 0, curve->npoints, hw->hist.smoothNpts); 
   }
   if (curve->visible) {
	if (curve->smooth) {
   	   ProcessYMaxChange(hw, group, curve->npoints, curve->smoothData, 
	       &change, True);
	} else {
   	   ProcessYMaxChange(hw, group, numSamples, 
	       &(curve->data[firstSample]), &change, True);
	}
	ProcessNewSegment(hw, curve, firstSample, change);
   }
   return;
}


void HistRestoreCurve(w, curveList, numCurve)
Widget w;
int *curveList;
int numCurve;
{
   InternalCurveRec *curve;
   int i;
   int cID;
   Widget prev;
   int needWidth = 0;
   Boolean change = False;
   Group *group ;
   int tmp_groupID;
   HistogramWidget hw = (HistogramWidget) w;

   if (numCurve < 1) return;

   for (i=0; i< numCurve; i++) {
	cID = curveList[i];
	curve = hw->hist.curves[cID];
        if(curve){  /* if this curve has not been deleted */
	    /*
	    curve->inPattern = NextLineStyleGCInd(hw);
	    curve->gc = hw->hist.patternGC[curve->inPattern];
	    */

	    curve->color_inPattern = NextLineStyleGCInd(hw,True);
	    curve->color_gc = hw->hist.patternGC[curve->color_inPattern];
	    curve->bw_inPattern = NextLineStyleGCInd(hw,False);
	    curve->bw_gc = hw->hist.bw_patternGC[curve->bw_inPattern];
	    if(hw->hist.defaultGC){
	        curve->gc = curve->color_gc;
	    }
	    else {
	        curve->gc = curve->bw_gc;
	    }

	    tmp_groupID = curve->groupIndex;
	    group = hw->hist.groups[tmp_groupID];
	     
   	    curve->select = False;
	    if (curve->smooth) {
   	       ProcessYMaxChange(hw, group, curve->npoints, curve->smoothData, 
	         &change, False);
	    } else {
   	       ProcessYMaxChange(hw, group, curve->npoints, curve->data, 
	         &change, False);
	    }
   	    if (++(group->numShown) == 1) {
	       needWidth += group->needWidth;
  	    }
        }
    }
   hw->hist.numShownCurve += numCurve;
   for (i=0; i< numCurve; i++) {
	if(hw->hist.curves[curveList[i]])
	  hw->hist.curves[curveList[i]]->visible = True;
   }
   hw->hist.yAxesWidth += needWidth;
   if (hw->hist.yAxesWidth > (int)WIDTH(hw,yAxesW)) { 
       hw->hist.layoutPending = True;
       for (i=0; i < numCurve; i++) {
	   AddBottomLabel(hw, curveList[i]);
       }

       /* increas yAxesW width, it'll also make graphW redrawed */
   	hw->hist.layoutPending = False;
	CHANGE_YAXES_SIZE(hw);
   } else {
   	hw->hist.layoutPending = True;
       /* make bottom lables appear at the bottom again */
       for (i=0; i < numCurve - 1; i++) {
	   AddBottomLabel(hw, curveList[i]);
       }
   	hw->hist.layoutPending = False;
	AddBottomLabel(hw, curveList[numCurve-1]);
   }

    prev = NULL;
    for (i=0; i < hw->hist.numCurve; i++) {
       if (hw->hist.curves[i] && hw->hist.curves[i]->visible) {
           XtVaSetValues(hw->hist.curves[i]->lineW, XtNfromVert, prev, NULL);
           XtVaSetValues(hw->hist.curves[i]->labelW, XtNfromVert, prev, NULL);
	   prev = hw->hist.curves[i]->lineW;
       }
    }
    XFlush(XtDisplay(hw->hist.yAxesW));
}



/* 
 * process the ymax change of this group according to the content
 * in "sample", which has "numSample" data,
 * return "change" to indicate if y max in this group has been changed
 * "redraw" is to tell us if we have to perform redrawing y axes when
 * y max changes. NOTE : if "redraw" is False, we only update the needWidth
 * of this group, but we don't update the total width needed by all y axes
 */
static void 
ProcessYMaxChange(w, group, numSample, sample, change, redraw)
Widget w;
Group *group;
int numSample; 
float *sample;
Boolean *change;
Boolean redraw;
{
   double ymax;
   int i;
   double oldNeedWidth;
   double diff_w;
   HistogramWidget hw = (HistogramWidget) w;


   ymax = group->y.max;
   for(i=0; i < numSample; i++) {
	if(!isnan(sample[i]))
	    if (sample[i] > ymax) ymax = sample[i];
   }
   /*
   if ((ymax <= group->y.max) && (ymax != 0.0)) 
   */
   if ((ymax <= group->y.max)){ 
	*change = False;
	return;
   } else {
	*change = True;

   	group->y.max = ymax + ymax * group->yInc;

	oldNeedWidth = group->needWidth;
	ComputeGroupWidth(hw,group);

	if (redraw) {
	  /* pseudo expose */
	  if ((diff_w = group->needWidth - oldNeedWidth) != 0.0) { 
	     hw->hist.yAxesWidth += diff_w;
	     CHANGE_YAXES_SIZE(hw);
	  } else {
	      /* if ymax change but the width does not change, then 
	       * we still need to redisplay y axis 
	       * should only display the one changed, but 
	       * I display all, because I don't know where the axis is
	       * on the screen */
	     hw->hist.clearY = True;
	     RedisplayYAxes(hw->hist.yAxesW, (XtPointer)False, NULL);
	  }
       }
   }
   return;

}


/* 
 * HistHideCurve : delete a curve, but not its label from graphW 
 * read bottomLabelW to get curves to be deleted 
 */
void HistHideCurve(w) 
Widget w;
{
   int i;
   HistogramWidget hw = (HistogramWidget) w;


   /* find selected curves */
   for (i=0; i< hw->hist.numCurve; i++) {
     if (hw->hist.curves[i] &&    /* this is a valid curve */
	 hw->hist.curves[i]->select &&  
	 hw->hist.curves[i]->visible){ /* curve not already hidden */
        ASS_T(hw->hist.curves[i]->lineW);
        hw->hist.curves[i]->visible = False;
	RedisplayLine(hw->hist.curves[i]->lineW,
		      (XtPointer)hw->hist.curves[i], NULL);
      }
   }

   /* unhighlight the labels */
   for (i=0; i< hw->hist.numCurve; i++) {
        if (hw->hist.curves[i] && hw->hist.curves[i]->select) {
	    XtVaSetValues(hw->hist.curves[i]->labelW, XtNstate, False, NULL);
	    hw->hist.curves[i]->select = False;
	}

   }

/*
   XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
   RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
*/

   hw->hist.clearY = True;
   HistResetYmax((Widget) hw);
   XFlush(XtDisplay(hw));
   return;
}

/*
 * HistColorChange: if color is True change the GC to color GC's
 * otherwise change the GC's to BW GC's
 */
void HistColorChange(w,color)
Widget w;
Boolean color;
{
    int i;
    HistogramWidget hw = (HistogramWidget) w;
    InternalCurveRec *curve;

    if((color && !hw->hist.defaultGC) || (!color && hw->hist.defaultGC)){
        hw->hist.defaultGC = color;
        for (i=0; i< hw->hist.numCurve; i++) {
	    curve = hw->hist.curves[i];
            if (curve){   /* this is a valid curve */
	        if(hw->hist.defaultGC){
	            curve->gc = curve->color_gc;
	        }
	        else {
		    curve->gc = curve->bw_gc;
	        }
	    }
        }
    }
    for (i=0; i< hw->hist.numCurve; i++) {
        curve = hw->hist.curves[i];
	 if((curve) && (curve->visible)){ /* curve not already shown */
            ASS_T(curve->lineW);
	    curve->visible = False;
	    RedisplayLine(curve->lineW, (XtPointer)curve, NULL);
	    curve->visible = True;
	    RedisplayLine(curve->lineW, (XtPointer)curve, NULL);
         }
    }
    HistResetYmax((Widget) hw);
    return;
}


/* 
 * HistShowCurve : redisplay a previously hidden curve
 * read bottomLabelW to get curves to be redisplayed 
 */
void HistShowCurve(w) 
Widget w;
{
   int i;
   HistogramWidget hw = (HistogramWidget) w;

   /* find selected curves */
   for (i=0; i< hw->hist.numCurve; i++) {
     if (hw->hist.curves[i] &&    /* this is a valid curve */
	 hw->hist.curves[i]->select &&  
	 !hw->hist.curves[i]->visible){ /* curve not already shown */
        ASS_T(hw->hist.curves[i]->lineW);
        hw->hist.curves[i]->visible = True;
	RedisplayLine(hw->hist.curves[i]->lineW,
		      (XtPointer)hw->hist.curves[i], NULL);
      }
   }

   /* unhighlight the labels */
   for (i=0; i< hw->hist.numCurve; i++) {
        if (hw->hist.curves[i] && hw->hist.curves[i]->select) {
	    XtVaSetValues(hw->hist.curves[i]->labelW, XtNstate, False, NULL);
	    hw->hist.curves[i]->select = False;
	}
   }

/*
   XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
   RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
*/
   HistResetYmax((Widget) hw);
   XFlush(XtDisplay(hw));
   return;
}









/* HistDeleteCurve : delete a curve from graphW
 * if *numDelete > 0, then read the curve IDs from the array **delete
 * otherwise read our bottomLabelW and see what are the curves be deleted 
 */
void HistDeleteCurve(w, delete, numDelete) 
Widget w;
int **delete;
int *numDelete;
{
   int i;
   int gID;
   int cID;
   Widget prev;
   int needWidth = 0;
   static int *temp_delete; 
   HistogramWidget hw = (HistogramWidget) w;

   ASS_LE(hw->hist.numShownCurve, hw->hist.numCurve);
   ASS_LE(hw->hist.numCurve, hw->hist.maxCurve);

   if (temp_delete == NULL || sizeof(temp_delete) < sizeof(int)*hw->hist.numCurve) {
   	temp_delete = (int*) XtRealloc((void *) temp_delete, 
	    hw->hist.numCurve * sizeof(int));
   }
   ASS_T(temp_delete);

   if (*numDelete <= 0) {
   	memset(temp_delete, '\0', sizeof(temp_delete));
   	*numDelete = 0;

   	for (i=0; i< hw->hist.numCurve; i++) {
	   if (hw->hist.curves[i]   /* a valid curve */
	       && hw->hist.curves[i]->select){  /* curve is selected */
	      temp_delete[(*numDelete)++] = i;
	      FreeALineStyle(hw, i);
	      gID = hw->hist.curves[i]->groupIndex;
   	      if(--(hw->hist.groups[gID]->numShown) == 0) { 
		   needWidth += hw->hist.groups[gID]->needWidth;
	      }
 	   }
        }
   	*delete = temp_delete;
   	hw->hist.numShownCurve -= *numDelete;
   } else {
	temp_delete = *delete;
   	hw->hist.numShownCurve -= *numDelete;
	for (i= 0; i< *numDelete; i++) {
	      cID = temp_delete[i];
	      FreeALineStyle(hw, cID);
	      gID = hw->hist.curves[cID]->groupIndex;
   	      if(--(hw->hist.groups[gID]->numShown) == 0) { 
		   needWidth += hw->hist.groups[gID]->needWidth;
	      }
	}
   }

   if (*numDelete <= 0) return;

   if (needWidth) {
   	hw->hist.layoutPending = True;
	for (i=0; i< *numDelete; i++) {
	    if(hw->hist.curves[temp_delete[i]]){
	        XtDestroyWidget(hw->hist.curves[temp_delete[i]]->lineW);
	        XtDestroyWidget(hw->hist.curves[temp_delete[i]]->labelW);
		XtFree((void*)hw->hist.curves[temp_delete[i]]->data);
		XtFree((void*)hw->hist.curves[temp_delete[i]]->name);
		XtFree((void*)hw->hist.curves[temp_delete[i]]->points);
		XtFree((void *) hw->hist.curves[temp_delete[i]]->smoothData);
		XtFree((void*)hw->hist.curves[temp_delete[i]]);
		hw->hist.curves[temp_delete[i]] = 0;
            }
	}
       	hw->hist.yAxesWidth -= needWidth;
   	hw->hist.layoutPending = False;
       	CHANGE_YAXES_SIZE(hw);
   } else {
   	hw->hist.layoutPending = True;
	for (i=0; i< *numDelete - 1; i++) {
	    if(hw->hist.curves[temp_delete[i]]){
	        XtDestroyWidget(hw->hist.curves[temp_delete[i]]->lineW);
	        XtDestroyWidget(hw->hist.curves[temp_delete[i]]->labelW);
		XtFree((void*)hw->hist.curves[temp_delete[i]]->data);
		XtFree((void*)hw->hist.curves[temp_delete[i]]->name);
		XtFree((void*)hw->hist.curves[temp_delete[i]]->points);
		XtFree((void *) hw->hist.curves[temp_delete[i]]->smoothData);
		XtFree((void*)hw->hist.curves[temp_delete[i]]);
		hw->hist.curves[temp_delete[i]] = 0;
	    }
	}
   	hw->hist.layoutPending = False;
	if(hw->hist.curves[temp_delete[*numDelete-1]]){
	  XtDestroyWidget(hw->hist.curves[temp_delete[*numDelete-1]]->lineW);
	  XtDestroyWidget(hw->hist.curves[temp_delete[*numDelete-1]]->labelW);
	  XtFree((void*)hw->hist.curves[temp_delete[*numDelete-1]]->data);
	  XtFree((void*)hw->hist.curves[temp_delete[*numDelete-1]]->name);
	  XtFree((void*)hw->hist.curves[temp_delete[*numDelete-1]]->points);
  	  XtFree((void *)hw->hist.curves[temp_delete[*numDelete-1]]->smoothData);
	  XtFree((void*)hw->hist.curves[temp_delete[*numDelete-1]]);
	  hw->hist.curves[temp_delete[*numDelete-1]] = 0;
	}
   }

   prev = NULL;
   for (i=0; i < hw->hist.numCurve; i++) {
       if (hw->hist.curves[i]) {
           XtVaSetValues(hw->hist.curves[i]->lineW, XtNfromVert, prev, NULL);
           XtVaSetValues(hw->hist.curves[i]->labelW, XtNfromVert, prev, NULL);
	   prev = hw->hist.curves[i]->lineW;
       }
   }

   CHANGE_YAXES_SIZE(hw);
   HistResetYmax((Widget) hw);
   XFlush(XtDisplay(hw));

  
   ASS_LE(hw->hist.numShownCurve, hw->hist.numCurve);
   ASS_LE(hw->hist.numCurve, hw->hist.maxCurve);
   return;
}

void HistClose(hw)
Widget hw;
{
  
  DestroyHist(hw);
  return;
}


void reComputeYmax(w)
HistogramWidget w;
{
    int i;
    Group *group;
    Boolean change;
    InternalCurveRec *curve;
    HistogramWidget hw = (HistogramWidget) w;
   
    for (i=0; i< hw->hist.numGroup; i++) { 
	group = hw->hist.groups[i]; 
	group->y.max = 0.0;
    }

    for (i=0; i < hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
	if (curve && curve->visible) {
	    group = hw->hist.groups[curve->groupIndex];
	    if (curve->smooth) {
		ProcessYMaxChange(hw, group, curve->npoints, curve->smoothData, 
		&change, True);
	    } else {
		ProcessYMaxChange(hw, group, curve->npoints, curve->data, 
		&change, True);
	    }
	}
    }
    hw->hist.clearGraph = True;
}

void HistSmoothCurve(w, curve_list, numCurve)
Widget w;
int **curve_list;
int *numCurve;
{
   int i;
   InternalCurveRec *curve;
   static int *temp_smooth; 
   HistogramWidget hw = (HistogramWidget) w;
   char *name;
   char *new_name;

   *numCurve = 0;
   if (temp_smooth == NULL || sizeof(temp_smooth) < sizeof(int)*hw->hist.numCurve)
   	temp_smooth = (int* )XtRealloc((void *) temp_smooth, 
	    hw->hist.numCurve * sizeof(int));

   memset(temp_smooth, '\0', sizeof(temp_smooth));

   for (i=0; i< hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
        if (curve && curve->select && curve->visible && !curve->smooth) {
	   curve->smooth = True;
	   temp_smooth[(*numCurve)++] = i;
   	   curve->smoothData = (float*) XtCalloc(hw->hist.maxSample + 1, sizeof(float));
	   SmoothData(hw, curve, 0, curve->npoints + 1, hw->hist.smoothNpts);
   	   CalculateXPoints(hw, curve, 0, curve->npoints);
           /* add smooothed to labels of smooothed curves */
           name = malloc(sizeof(char)*(strlen(curve->name)+1));
	   strcpy(name,curve->name);
           new_name = malloc(sizeof(char)*(strlen(name) + 14));
	   if(new_name){
               strcpy(new_name, name);
	       strcat(new_name, " (smoothed)");
               XtVaSetValues(curve->labelW, XtNlabel,new_name,NULL); 
	       free(new_name);
	   }
	   free(name);
	}
   }
   *curve_list = temp_smooth; /* return list of curve id */


   if (*numCurve) {
        reComputeYmax(hw);
   	XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
   	RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
   }

   /* unhighlight the labels */
   for (i=0; i< hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
        if (curve && curve->select) {
	    XtVaSetValues(curve->labelW, XtNstate, False, NULL);
	    curve->select = False;
	}
   }
}


void HistUnsmoothCurve(w, curveList, numCurve)
Widget w;
int **curveList;
int *numCurve;
{
   int i;
   InternalCurveRec *curve;
   static int *temp_unsmooth; 
   HistogramWidget hw = (HistogramWidget) w;
   char *name;

   *numCurve = 0;
   if (temp_unsmooth == NULL || sizeof(temp_unsmooth) < sizeof(int)*hw->hist.numCurve) {
   	temp_unsmooth = (int* )XtRealloc((void *) temp_unsmooth, 
	    hw->hist.numCurve * sizeof(int));
   }
   memset(temp_unsmooth, '\0', sizeof(temp_unsmooth));

   for (i=0; i< hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
        if (curve && curve->select && curve->visible && curve->smooth) {
	   curve->smooth = False;
	   temp_unsmooth[(*numCurve)++] = i;
	   XtFree((void *) curve->smoothData);
	   curve->smoothData = NULL;
   	   CalculateXPoints(hw, curve, 0, curve->npoints);
           /* add smooothed to labels of smooothed curves */
           name = malloc(sizeof(char)*(strlen(curve->name)+1));
	   if(name){
	       strcpy(name,curve->name);
               XtVaSetValues(curve->labelW, XtNlabel,name,NULL); 
	       free(name);
	   }
	}
   } 
   *curveList = temp_unsmooth; /* return list of curve id */

   if (*numCurve) {
        reComputeYmax(hw);
   	XClearWindow(XtDisplay(hw->hist.graphW), XtWindow(hw->hist.graphW));
   	RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
   }

   /* unhighlight the labels */
   for (i=0; i< hw->hist.numCurve; i++) {
	curve = hw->hist.curves[i];
        if (curve && curve->select) {
	    XtVaSetValues(curve->labelW, XtNstate, False, NULL);
	    curve->select = False;
	}
   }
}




/* ARGSUSED */
static void SmoothData(w, curve, first, last, forward)
Widget w;
InternalCurveRec *curve;
int first, last;
int forward;
{
#ifdef n_def
   float sum;
   int i, j, k = first;
   int npts = forward+1;

   if (first <= forward) {
   	for (i = first; i <= forward; i++) {
	   sum = 0.0;
	   for (j= 0; j<= i; j++) {
	       sum += curve->data[j];
	   }
	   curve->smoothData[i] = sum/(i+1); 
      	}
	k = i;
   }
   for ( ; k <= last; k++) {
	sum = 0.0;
	for (j= k - forward; j<= k; j++) {
	    sum += curve->data[j];
	}
	curve->smoothData[k] = sum/npts; 
   }
#endif

    float sum;
    int i, j, k = first;
    int npts = forward+1;

    if (first <= forward) {
	/* find next valid range of data */
	while( (first <= forward) && (isnan(curve->data[first]))){
            curve->smoothData[first] = curve->data[first];
	    first++;
	}
	i = first;
	while((i <= forward) && (!isnan(curve->data[i]))){
            sum = 0.0;
	    for(j = first; j <= i; j++){
                sum += curve->data[j];
	    }
	    curve->smoothData[i] = sum/(i+1);
	    i++;
	}
	first = i;
	k = i;
    }
    while(k <= last){
	/* find next valid range of data */
	j = k - forward;
        while( (j <= k) && (isnan(curve->data[j]))){
            curve->smoothData[j] = curve->data[j];
	    j++;
        }
	sum = 0.0;
        while( (j <= k) && (!isnan(curve->data[j]))){
            sum += curve->data[j];
	    j++;
	}
	curve->smoothData[k] = sum/npts;
	k++;
   }
}


void HistResetYmax(Widget w)
{
   HistogramWidget hw = (HistogramWidget) w;

    reComputeYmax(hw);
    RedisplayYAxes(hw->hist.yAxesW, (XtPointer)False, NULL);
    RedisplayGraph(hw->hist.graphW, (XtPointer)False, NULL);
}

