/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

/*
 * $Log: HistogramP.h,v $
 * Revision 1.7  1996/08/16 19:03:39  tamches
 * updated copyright notice for release 1.1
 *
 * Revision 1.6  1995/10/19 02:42:15  newhall
 * added support for changing between black-and-white and color displays.
 * added defaultGC resource to Histogram Widget--if true, start in color mode,
 * if false, start in BW mode. added HistColorChange function.  purify fixes.
 *
 * Revision 1.5  1995/04/01  22:04:01  newhall
 * added test for Nan values when computing points
 *
 * Revision 1.4  1994/07/28  17:17:16  newhall
 * removed deleted flag to struct HistogramPart
 *
 * Revision 1.3  1994/07/24  21:54:12  newhall
 * added deleted flag to struct HistogramPart to force graph to be
 * redrawn when a curve is deleted
 *
 * Revision 1.2  1994/05/27  20:50:27  newhall
 * added HistHideCurve, HistShowCurve
 * changed data values from doubles to floats
 * added HistHideCurve, HistShowCurve
 * changed data values from doubles to floats
 *
 * Revision 1.1  1994/03/31  18:11:24  krisna
 * implementation files for the Histogram widget
 *
 * Revision 1.3  1993/01/26  21:10:30  hollings
 * Changed to use IPS style selection of ymax and increments.
 *
 * Revision 1.2  1993/01/20  23:00:38  hollings
 * upgrade to use new xdr based transfer approach.
 *
 * Revision 1.1  1993/01/08  21:20:23  hollings
 * Initial revision
 *
 *
 * HistogramP.h	- Private header file for the Histogram Widget.
 *
 */

#ifndef _HistogramP_h
#define _HistogramP_h

#include <X11/Xaw/FormP.h>
#include "hist/h/Histogram.h"

#define FINAL 0

#define MAXSELECTED	3

typedef struct {
    double unitWidth;	/* size in pixels of one unit on the screen */
    double tickSpace;	/* size in pixels of one tichs on the screen */
    double interval;	/* increment between ticks (units not pixels) */
    char format[8];
    double max;		/* max value of axis (not the same as max data point */
    int	count;		/* number of marks */
} Label;


/* metrics in the same group have same y title,metricID,vertical, max */
typedef struct _group {
    int groupID;   /* metric identifier, we don't interpret this ID */
    int numMember;  /* number of members in this group */
    int numShown;
    char *title;      /* title of y-axis */
    int titleWidth; /* title width plus mark space */
    int needWidth; /* total width of this title with its label */
    Label y;      /* y-axis property */
    float yInc;  
} Group;

typedef struct {
	int	curveID;	
	int 	groupIndex;	/* used to find its group */
	Widget wID;
	Widget  labelW;	/*  widget for label in the bottom box */
	Widget  lineW;  /* widget for line */
	Pixmap  linemap;  /* line shown in the bottom label */
	int     select; /* if select for later deletion */
	char	*name;	/* curve name shown at the bottom */
	Dimension nameWidth;
	int maxBuckets;
	double bucketWidth;
	GC	 gc;    /* either color_gc or bw_gc */
	GC	 color_gc;
	GC	 bw_gc;
	int	 color_inPattern;  /* index of patternGC in hw->hist */
	int	 bw_inPattern;     /* index of patternGC in hw->hist */
	float	 *data;
	XPoint   *points;
	float    *smoothData;
	Boolean  smooth;
	int 	 npoints;
	Boolean visible;
	Boolean show_label;
} InternalCurveRec;


typedef struct {int empty;} HistogramClassPart;

typedef struct _HistogramClassRec {
    CoreClassPart	core_class;
    CompositeClassPart	composite_class;
    ConstraintClassPart	constraint_class;
    FormClassPart	form_class;
    HistogramClassPart	histogram_class;
} HistogramClassRec;

extern HistogramClassRec histogramClassRec;




typedef struct _HistogramPart {
    /* resources */
 Pixel 		foreground; 
 Pixel 		gForeground;
 Pixel 		gBackground;
 Dimension 	gWidth;
 Dimension 	gHeight;
 int		maxCurve;
 int		maxGroup;
 char		*font;  /* font for displaying text */
 int 		numXMarks; /* # markers in time-axis */
 float 		yInc;	/* increament in Y-axis when new y max too large */
 float 		yMin; /* min y value of high end of each y axis */
 int 		smoothNpts; /* #pts forward to compute smooth data */
 Boolean        defaultGC;   /* if true use color GC, otherwise use bw */ 

    /* private state */
 /**!!!!!!!!!!!!!!!delete this widget later */
 Widget emptyW;
 /************/
 Widget graphW;		/* main graph widget */
 Widget yAxesW;		/* Widget to hold data */	
 Widget bottomLabelW;
 Widget timeAxisW;
 Widget scrollW;	/* Horizontal scroll bar */
 Widget zoomBarW;	/* zoom scroll bar */

 Widget PANW;
 Widget ZOOMW;

 Boolean layoutPending;

 /* hold old width & height of Histogram Widget before being resized */
 int oldWidth;
 int oldHeight;

 Boolean clearGraph;
 Boolean clearY;
 Boolean clearX;

 int visibility; 	/* is graph visible */

 GC histFgGC;		/* bg & fg GC for space outside graph */
 GC histBgGC;

 int	timeUnit;

 XFontStruct *fontInfo;
 XCharStruct charStruct;
 XCharStruct numberStruct;
 int	numberHeight;
 int    charHeight;
 int 	numberWidth;
 int 	charWidth;
 Font	fontID;
 Dimension fontHeight;

 double	startTime; 	  /* value of origin of x-axis */
 double 	timeWidth;
 int	maxSample;
 int 	timeMarginWidth;

 InternalCurveRec **curves;
 int	currCurveID;
 int numCurve;
 int numShownCurve;
 int numGroup;
 Group **groups;
 int yAxesWidth;

 char *colorLineStyles;
 char *bwLineStyles;

 int currLineStyle;
 int numPatterns;
 int *patternUsed;
 GC *patternGC;

 int  *bw_patternUsed; /* for black and white patterns */
 GC   *bw_patternGC;   /* for black and white GC's */
 int   bw_numPatterns; /* for black and white number of patterns */


 Label time;

 /* min space between labels in y-axis and time axis */
 double minYLabelSpace, minXLabelSpace;
 
 double zoom; /* fraction (of full histogram) that is displayed */
 double start; /* fraction (of full histogram) from left to start */
 double focus; /*fraction (of full histogram) from left to center */
 int currXLeft; /* current left x pixel position in the pseudo timeW */
} HistogramPart;

typedef struct _HistogramRec {
    CorePart		core;
    CompositePart	composite;
    ConstraintPart	constraint;
    FormPart		form;
    HistogramPart	hist;
} HistogramRec;

typedef struct {
    int ignore;
} HistogramConstraintsPart;

typedef struct _HistogramConstraintsRec {
    FormConstraintsPart		form;
    HistogramConstraintsPart	hist;
} *HistogramConstraints, HistogramConstraintsRec;

#endif /* _HistogramP_h*/
