/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */


/*
 * req.c
 * Request input creates a nice pop up window which accepts
 * input from the user.  The window contains a title,
 * reqc areas of input, buttons to select areas of input,
 * two exit buttons, labeled with accept and cancel.
 * The areas of input are labeled with the requests in 
 * reqv and given default values from reqd.
 *
 * The result is returned through the answer function, which
 * accepts an boolean to indicate acceptance or cancellation,
 * An int to indicate the number of returned values.
 * An array of strings, to indicate the returned values.
 * An array of ints, (boolean) to indicate wether the string has
 * changed or not.
 *
 * The strings are free'ed after the call, so the strings should
 * be saved in the answer routines.
 *
 * $Log: req.c,v $
 * Revision 1.4  1999/08/09 05:38:19  csserra
 * - added support for (mips-sgi-irix6.4) native compiler build
 * - eliminated misc. compiler warnings
 *
 * Revision 1.3  1996/08/16 19:03:40  tamches
 * updated copyright notice for release 1.1
 *
 * Revision 1.2  1994/05/04 00:24:06  hollings
 * Fixed pand and zoom to work. The conversion of the data type to double
 * broke many other things since all floats can become doubles!
 *
 * Make it pass through gcc -Wall with fewer errors.
 *
 * Revision 1.1  1994/03/31  18:11:27  krisna
 * implementation files for the Histogram widget
 *
 * Revision 2.24  1993/08/03  22:18:48  krisna
 * solaris-2 (openwindows update)
 *
 * Revision 2.23  1993/01/26  22:48:17  krisna
 * Solaris-2 update: define XawChain interms of XtChain (for OpenWindows)
 *
 * Revision 2.22  1993/01/20  22:56:21  hollings
 * removed ips icon references.
 *
 * Revision 2.21  1992/10/08  15:30:59  jcargill
 * Corrected types in XGetGeometry call
 *
 * Revision 2.20  1992/07/31  18:20:44  rbi
 * Informix Merge
 *
 * Revision 2.19  1992/03/14  00:42:35  joerg
 * Additional changes for common usage
 *
 * Revision 2.18  1992/03/06  19:59:08  joerg
 * Changes for common usage, removed redundant check
 *
 * Revision 2.17  1992/01/13  16:42:56  hollings
 * x11r5 changes.
 *
 * Revision 2.16  1991/06/28  16:28:09  rbi
 * onthefly derived metrics.
 *
 * Revision 2.15  1990/11/06  14:10:56  hollings
 * Fixed bug in height of viewport caused by missing {}.
 *
 *
 */

#include <assert.h>
#include <stdio.h>
#include <math.h>
#include <malloc.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Viewport.h>
#include <X11/Shell.h>
#include <X11/Xaw/Toggle.h>

/* local include files */
#include "FormatBox.h"
#include "hist/h/req.h"

#define	FONTWIDTH	6	/* just for approximation */
#define EXPANSION       (5 * FONTWIDTH)
#define MIN_TEXT_WIDTH	(25 * FONTWIDTH)
#define DEF_DIST	15
#define MAX_COLS	3

void free();

static Display *display;	/*					*/
static Widget rootWidget;	/* internal copy of rootWidget		*/
static Pixmap icon_pixmap;	/* compiled ips icon from icon.h	*/
static int width;		/* of the Prop. Sheet			*/
static XtTranslations trans;	/* compiled translations		*/
int    start_x, start_y;	/* left upper edgde of Prop. Sheet	*/
				/* used also in onethefly.c		*/
static int MaxItemsPerCol ;	/* # of items in a prop. sheet col.	*/

static String translations = "#override \n\
    <Key>Return: no-op(RingBell)        \n\
    Ctrl<Key>M: no-op(RingBell)         \n\
    Ctrl<Key>N: no-op(RingBell)";

struct request {
    Widget shell;
    Widget form;
    Widget title;
    Widget vport;		/* view port for list of options */
    Widget data;		/* form holding options */
    Widget *labelW;
    Widget *inputW;
    Widget buttonBox;		/* box to hold buttons */
    Widget buttonW[MAX_OPTIONS];
    PropertySheet	*sheet;
};

/* ARGSUSED */
static void
PressFunc(w, top, cdata)
Widget w;
struct request *top;
int cdata;
{
    int curr;
    PropertySheet *sheet;

    sheet = top->sheet;
    for (curr=0; curr < sheet->itemCount; curr++) {
	if (top->labelW[curr] == w) break;
    }
    sheet->items[curr].value[0] ^= 1;  	/* flip the bit */
}


int
BuildPropertySheet(sheet)
PropertySheet *sheet;
{
    int col;
    int i, j;
    int count;
    Widget left;
    int leftIndex;
    static void Feep();
    static void XtSetValue();
    Arg args[200];
    PropItem *item;
    Widget rg, last;
    int itemsPerCol;
    XSizeHints hints;
    struct request *new;
    static void EndReqFunc(Widget, struct request *, int);
    int columns;
    static GetWidgetWidth();
    static TextWidth();

    if (sheet->itemCount < 0) return(-1);

    new = (struct request *) XtMalloc(sizeof(struct request));
    new->sheet = sheet;
    new->inputW = (Widget *) 
      XtMalloc(sizeof(Widget) * sheet->itemCount);
    new->labelW = (Widget *) 
      XtMalloc(sizeof(Widget) * sheet->itemCount);

    count = 0;
    XtSetArg(args[count], XtNx, start_x); count++;
    XtSetArg(args[count], XtNy, start_y); count++;
    XtSetArg(args[count], XtNiconPixmap, icon_pixmap); count++;
    XtSetArg(args[count], XtNmappedWhenManaged, False); count++;
    new->shell = XtCreatePopupShell("Prop. Sheet", transientShellWidgetClass,
		rootWidget, args, count);

    /* box to hold items */
    count = 0;
    XtSetArg(args[count], XtNdefaultDistance, DEF_DIST); count++;
    new->form = XtCreateManagedWidget("form", formWidgetClass,
	new->shell, args, count);
    XtAddEventHandler(new->form, ButtonPressMask, 0, Feep, NULL);

    /* title */
    count = 0;
    XtSetArg(args[count], XtNresizable, True); count++;
    XtSetArg(args[count], XtNborderWidth, 0); count++;
    XtSetArg(args[count], XtNjustify, XtJustifyCenter); count++;
    XtSetArg(args[count], XtNtop, XawChainTop); count++;
    XtSetArg(args[count], XtNbottom, XawChainTop); count++;
    XtSetArg(args[count], XtNleft, XawChainLeft); count++;
    XtSetArg(args[count], XtNright, XawChainRight); count++;
    XtSetArg(args[count], XtNlabel, sheet->title); count++;
    new->title = XtCreateManagedWidget("title", labelWidgetClass, 
	new->form, args, count);


    itemsPerCol = MaxItemsPerCol;
    columns = ceil((double) sheet->itemCount / MaxItemsPerCol);
    if (columns > MAX_COLS) {
	columns = MAX_COLS;
	itemsPerCol = sheet->itemCount / columns + 1;
    }

    /* View port */
    count = 0;
    XtSetArg(args[count], XtNresizable, True); count++;
    XtSetArg(args[count], XtNfromVert, new->title); count++;
    XtSetArg(args[count], XtNtop, XawChainTop); count++;
    XtSetArg(args[count], XtNbottom, XawChainBottom); count++;
    if ((itemsPerCol > MaxItemsPerCol) || 
	((sheet->type == REQ_STRING) && 
	 ((sheet->itemCount * 2) > MaxItemsPerCol))) {
	XtSetArg(args[count], XtNheight, 375); count++;
	XtSetArg(args[count], XtNallowVert, True); count++;
	XtSetArg(args[count], XtNallowHoriz, True); count++;
	XtSetArg(args[count], XtNuseBottom, True); count++;
	/* need borders on these */
    } else {
	XtSetArg(args[count], XtNborderWidth, 0); count++;
    }

    /* force vport to have some height so it won't core dump */
    if (!sheet->itemCount) {
	XtSetArg(args[count], XtNheight, 10); 
	count++;
    }
    new->vport = XtCreateManagedWidget("vport", viewportWidgetClass, new->form,
	args, count);

    /* build the holding form */
    count = 0;
    XtSetArg(args[count], XtNresizable, True); count++; 
    new->data = XtCreateManagedWidget("data", formWidgetClass, new->vport,
	args, count);

    rg = NULL;
    last = NULL;
    left = NULL;
    col = 0;
    for (i = 0; i < sheet->itemCount; i++) {
      item = &(sheet->items[i]);
      count = 0;
      if (i && (i % itemsPerCol) == 0) {
	/* Start new column - left is first item in last column */
	leftIndex = col * itemsPerCol;
	left = new->labelW[leftIndex];
	last = NULL;
	col++;
      } else if (col) {
	left = new->labelW[leftIndex++];
      }
      XtSetArg(args[count], XtNfromVert, last); count++;
      XtSetArg(args[count], XtNfromHoriz, left); count++;
      XtSetArg(args[count], XtNjustify, XtJustifyLeft); count++;
      XtSetArg(args[count], XtNborderWidth, 0); count++;
      XtSetArg(args[count], XtNresizable, True); count++;
      if ((item->type == REQ_STRING) || (item->type == REQ_ACTION)
	  || (item->type == REQ_LABEL)) {
	/* create label */
	XtSetArg(args[count], XtNinternalWidth, 0); count++;
	XtSetArg(args[count], XtNleft, XtChainLeft); count++;
	XtSetArg(args[count], XtNright, XtChainLeft); count++;
	if (i != 0) {
	  if (sheet->items[i-1].type == REQ_LABEL) {
	    XtSetArg(args[count], XtNvertDistance, 10); count++;
	  }
	  else {
	    XtSetArg(args[count], XtNvertDistance, 20); count++;
	  }
	}
	new->labelW[i] = XtCreateManagedWidget(item->label,
			labelWidgetClass, new->data, args, count);
	
        last = new->labelW[i];

	if (item->type == REQ_STRING) {
	  /* config edit widget */
	  count = 0;
	  XtSetArg(args[count], XtNresizable, True); count++;
	  XtSetArg(args[count], XtNtranslations, trans); count++;
	  XtSetArg(args[count], XtNeditType, XawtextEdit); count++;
	  XtSetArg(args[count], XtNresize, XawtextResizeWidth); count++;
	  XtSetArg(args[count], XtNfromVert, new->labelW[i]); count++;
	  XtSetArg(args[count], XtNvertDistance, 4); count++;
	  XtSetArg(args[count], XtNleft, XtChainLeft); count++;
	  XtSetArg(args[count], XtNright, XtChainRight); count++;
	  XtSetArg(args[count], XtNuseStringInPlace, True); count++;
	  XtSetArg(args[count], XtNstring, item->value); count++;
	  XtSetArg(args[count], XtNlength, RESULTLENGTH); count++;
	  XtSetArg(args[count], XtNinsertPosition, strlen(item->value));count++;

	  new->inputW[i] = XtCreateManagedWidget("dataBox", 
			 asciiTextWidgetClass, new->data, args, count);

	  if ((item->HLeft >= 0) && (item->HLeft <= item->HRight)) {
	    XawTextSetSelection(new->inputW[i],item->HLeft,item->HRight);
	    XawTextSetInsertionPoint(new->inputW[i],item->HLeft);
	  }
	} 
	if (item->type == REQ_ACTION) {
	  /* list of buttons to press */
	  count = 0;
	  XtSetArg(args[count], XtNresizable, True); count++;
	  XtSetArg(args[count], XtNformat, XtNcenter); count++;
	  XtSetArg(args[count], XtNspanning, False); count++;
	  XtSetArg(args[count], XtNorientation, XtorientHorizontal); 
	  count++;
	  XtSetArg(args[count], XtNfromVert, new->labelW[i]); count++;
	  XtSetArg(args[count], XtNvertDistance, 4); count++;
	  XtSetArg(args[count], XtNborderWidth, 0); count++;
	  XtSetArg(args[count], XtNleft, XtChainLeft); count++;
	  XtSetArg(args[count], XtNright, XtChainRight); count++;
	  new->inputW[i] = XtCreateManagedWidget("dataRegion",
		     formatBoxWidgetClass, new->data, args, count);
	  
	  /* now build buttons */
	  rg = NULL;
	  for (j = 0; j < item->numOptions; j++) {
	    count = 0;
	    XtSetArg(args[count], XtNradioData, j+1); count++;
	    XtSetArg(args[count], XtNradioGroup, rg); count++;
	    if (item->initValue == item->options[j]) {
	      /* make it selected */
	      XtSetArg(args[count], XtNstate, True); count++;
	    }
	    rg = item->buttonW[j] =  
	      XtCreateManagedWidget(item->options[j], toggleWidgetClass, 
				    new->inputW[i], args, count);
	  }
	}
	if (item->type != REQ_LABEL)
	  last = new->inputW[i];
      } else {
	/* multi-or single selection */
	XtSetArg(args[count], XtNresize, True); count++;
	XtSetArg(args[count], XtNresizable, True); count++;
	XtSetArg(args[count], XtNvertDistance, 5); count++;
	XtSetArg(args[count], XtNstate, item->value[0]); count++;
	if (item->type == REQ_EXCLUSIVE) {
	    XtSetArg(args[count], XtNradioData, i+1); count++;
	    XtSetArg(args[count], XtNradioGroup, rg); count++;
	} 
	new->labelW[i] = XtCreateManagedWidget(item->label,
	    toggleWidgetClass, new->data, args, count);
	
	/* hack to work around limitation of toggle widget that it doesn't
	 *   infer that one item radio groups should be a member of their
	 *   own group */
	if (!i && (item->type == REQ_EXCLUSIVE)) {
	  XawToggleChangeRadioGroup(new->labelW[i], new->labelW[i]);
	}
	
	if (item->type ==  REQ_BOOLEAN) {
	  XtAddCallback(new->labelW[i], XtNcallback,
		(XtCallbackProc) PressFunc, new);
	}
	rg = last = new->labelW[i];
      }
    }

    count = 0;
    XtSetArg(args[count], XtNresizable, True); count++;
    XtSetArg(args[count], XtNfromVert, new->vport); count++;
    XtSetArg(args[count], XtNtop, XawChainBottom); count++;
    XtSetArg(args[count], XtNbottom, XawChainBottom); count++;
    XtSetArg(args[count], XtNleft, XawChainLeft); count++;
    XtSetArg(args[count], XtNright, XawChainRight); count++;
    XtSetArg(args[count], XtNborderWidth, 0); count++;
    XtSetArg(args[count], XtNhSpace, DEF_DIST); count++; 
    XtSetArg(args[count], XtNformat, XtNcenter); count++;
    XtSetArg(args[count], XtNorientation, XtorientHorizontal); count++;
    new->buttonBox = XtCreateManagedWidget("buttonBox", formatBoxWidgetClass, 
	new->form, args, count);

    /* build buttons */
    for (i = 0; i < sheet->buttonCount; i++) {
	count = 0;	/*#######*/
	new->buttonW[i] = XtCreateManagedWidget(sheet->buttons[i], 
	    commandWidgetClass, new->buttonBox, args, count);
	XtAddCallback(new->buttonW[i], XtNcallback, 
	    (XtCallbackProc) EndReqFunc, new);
    }

    /* make the widget */
    XtRealizeWidget(new->shell);

    /* find max width item */
    width = 0;
    for (i=0; i < sheet->itemCount; i++) {
	int	itemwidth ;

	itemwidth = GetWidgetWidth(new->labelW[i]);
	width = (itemwidth > width) ? itemwidth : width;
	if ((sheet->items[i].type == REQ_STRING) ||
	    (sheet->items[i].type == REQ_ACTION)) {
	    itemwidth = GetWidgetWidth(new->inputW[i]);
	    width = (itemwidth > width) ? itemwidth : width;
	}
    }

    XawFormDoLayout(new->data, False);
    for (i=0; i < sheet->itemCount; i++) {
	if (sheet->items[i].type == REQ_STRING) {
	    XtSetValue(new->inputW[i], XtNwidth, (char *) width);
	    for (j=0; j < sheet->buttonCount; j++) {
		XtInstallAccelerators(new->inputW[i], new->buttonW[j]);
	    }
	} else if (sheet->items[i].type == REQ_ACTION) {
	    XtSetValue(new->inputW[i], XtNwidth, (char *) width);
	} else {
	    XtSetValue(new->labelW[i], XtNwidth, (char *) width);
	    for (j=0; j < sheet->buttonCount; j++) {
		XtInstallAccelerators(new->labelW[i], new->buttonW[j]);
	    }
	}
    }
    XawFormDoLayout(new->data, True);
    /* Ask viewport to get bigger */
    XtSetValue(new->vport, XtNwidth, 
	(char *) (columns * (width + DEF_DIST) + 2*DEF_DIST));
    XtSetValue(new->title, XtNwidth, 
	(char *) (columns * (width + DEF_DIST) + 2*DEF_DIST));
    XtSetValue(new->buttonBox, XtNwidth, 
	(char *) (columns * (width + DEF_DIST) + 2*DEF_DIST));

    hints.y = start_y;
    hints.x = start_x;
    hints.flags = USPosition;
    XSetNormalHints(XtDisplay(new->shell), XtWindow(new->shell), &hints);

    XtMapWidget(new->shell);	/* map if after we move it */
    return(0);
}


/*
 * Initialize Request
 */
void InitReqWindow(w, x, y, maxItemsPerCol, icon_bits, icon_width, icon_height)
Widget w;
int x, y;
int maxItemsPerCol ;
char *icon_bits;
int icon_width;
int icon_height;
{
    rootWidget = w ;
    start_x = x;
    start_y = y;
    MaxItemsPerCol = maxItemsPerCol ;

    trans = XtParseTranslationTable(translations);

    display = XtDisplay(w);
    if (icon_bits) {
	icon_pixmap = XCreateBitmapFromData(display, DefaultRootWindow(display),
	    icon_bits, icon_width, icon_height);
    }
}


/*
 * Helper function for simple interactive property sheets.
 */
int
RequestInput(answer, title, accept, cancel, reqc, reqv, original, type, cdata)
void (*answer)();	/* call function on accept */ 
char *title;    	/* Title the window */
char *accept;   	/* lable for the accept button */
char *cancel;   	/* label for the cancel button */
int reqc;       	/* request count */
char **reqv;    	/* request value */
char **original;     	/* input default */
int type;    		/* request type: boolean, string, or exclusive */
char *cdata;		/* data to pass on callback */
{

    int i;
    PropItem *items;
    PropertySheet *sheet;

    sheet = (PropertySheet *) calloc(sizeof(PropertySheet), 1);
    items = (PropItem *) calloc(sizeof(PropItem), reqc);

    for (i=0; i < reqc; i++) {
	items[i].type = type;
	items[i].label = reqv[i];
	if (original)
	    items[i].initValue = original[i];
	if(items[i].initValue)
	    strcpy(items[i].value, items[i].initValue);
	items[i].HLeft = -1;	/* make sure the insert pointer in the */
				/* PropSheet is moved to the end       */
    }

    sheet->callBack = answer;
    sheet->title = title;
    sheet->buttonCount = 2;
    sheet->buttons[0] = accept;
    sheet->buttons[1] = cancel;
    sheet->itemCount = reqc;
    sheet->type = type;
    sheet->cdata = (int) cdata;
    sheet->items = items;

    BuildPropertySheet(sheet);

    return(0);
}


/* ARGSUSED */
static void EndReqFunc(Widget w, struct request *top, int cdata)
{
    int i;
    int sel;
    PropItem *item;
    PropertySheet *sheet;

    sheet = top->sheet;
    if (sheet->type == REQ_EXCLUSIVE) {
	sel = (int) XawToggleGetCurrent(top->labelW[0]);
	/* sel is N-1 to make it different than non-selected - which is 0 */
	for (i=0; i < sheet->itemCount; i++) {
	    sheet->items[i].value[0] = (char) 0;
	}
	if (sel) sheet->items[sel-1].value[0] = (char) 1;
    } else {
	for (i=0; i < sheet->itemCount; i++) {
	    item = &(sheet->items[i]);
	    if (item->type == REQ_ACTION) {
		sel = (int) XawToggleGetCurrent(item->buttonW[0]);
		if (sel) strcpy(item->value, item->options[sel - 1]);
	    }
	}
    }
    XtUnmapWidget(top->shell);
    XtDestroyWidget(top->shell);

    if (sheet->callBack) {
      (sheet->callBack)(w==top->buttonW[0], sheet->items, sheet->itemCount,
			sheet->cdata);
    }

    /* now clean up */
    if (top->sheet->items)
      free(top->sheet->items);
    free(top->sheet);
    if (top->inputW)
      free(top->inputW); 
    if (top->labelW)
      free(top->labelW); 
    free(top);
}


/* 
 * Function to create an empty property sheet 
 *
 */
PropertySheet *
InitPropertySheet(callBack, type, title, numItems)
void (*callBack)();	/* call function on accept */ 
int type;
char *title;
int numItems;
{
    PropertySheet *new;

    new = (PropertySheet *) calloc(sizeof(PropertySheet), 1);
    new->callBack = callBack;
    new->type = type;
    new->itemCount = numItems;
    new->items = (PropItem *) calloc(sizeof(PropItem), numItems);
    new->title = title;
    return(new);
}

/*
 * Function returns width of a given widget
 */
static int GetWidgetWidth(Widget w)
{
    unsigned int width, dummy_uint_return;
    int dummy_int_return;
    Window root_return;

    XGetGeometry(XtDisplay(w), XtWindow(w), &root_return, &dummy_int_return,
		 &dummy_int_return, &width, &dummy_uint_return, 
		 &dummy_uint_return, &dummy_uint_return);
    return width;
}

/*
 * Feeps.
 */
static void
Feep()
{
    XBell(display, 0);
}

/*
 * The X tool kit should have had this function defined
 */
static void
XtSetValue(widget, field, result)
Widget widget;
char *field;
caddr_t result;
{
    Arg AList[1];

    XtSetArg(AList[0], field, result);
    XtSetValues(widget, AList, 1);
}
