/*
 * Copyright (c) 1996 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

// $Id: rthist.C,v 1.58 2000/08/21 00:35:00 paradyn Exp $

/************************************************************************
 * rthist.c: a realtime histogram front-end based on the visi library.
************************************************************************/

#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Paned.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/Sme.h>
#include <X11/Xaw/SmeBSB.h>
#include <X11/Xaw/SmeLine.h>
#include <X11/Xaw/Toggle.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "pdutil/h/pdsocket.h"
#include "hist/h/Histogram.h"
#include "hist/h/req.h"
#include "visi/h/visualization.h"
#include "paradyn/xbm/logo.xbm"
#include "common/h/String.h"
#include "common/h/Types.h"

#include "common/h/Ident.h"
extern "C" const char V_rthist[];
Ident V_id(V_rthist,"Paradyn");
extern "C" const char V_libpdutil[];
Ident V_Uid(V_libpdutil,"Paradyn");
extern "C" const char V_libhist[];
Ident V_Hid(V_libhist,"Paradyn");
extern "C" const char V_libvisi[];
Ident V_Vid(V_libvisi,"Paradyn");

#define NEWCURVE      0
#define DELETECURVE   1
#define MAKEPERSISTENT 2
#define NUM_MENU_ITEM 3

#define SMOOTH        0
#define UNSMOOTH      1
#define HIDECURVE     2
#define UNHIDECURVE   3
#define BW_COLOR      4
#define YMAX          5
#define NUM_DISP_MENU_ITEM 6

#define CLOSE         0
#define NUM_FILE_MENU_ITEM 1

#define GENERAL         0
#define CONTEXT         1

#define MAXCURVES	16

static const char *menu_item[] = {
    "New Curve(s)",
    "Delete Curve(s)",
    "Keep on Exit",
};

static const char *disp_menu_item[] = {
    "Smooth Curve(s)",
    "Unsmooth Curve(s)",
    "Hide Curve",
    "Show Curve",
    "BW/Color Toggle",
    "Refresh Display",
};

static const char *file_menu_items[] = {
    "Close Histogram",
};

struct HistData {
    int      curve_id;
    unsigned prev_last;
    char*    title;
    unsigned metric;
    unsigned resource;
    int      groupId;
};

extern int start_x, start_y;

static Widget hw;
static Widget toplevel;
static Widget titleBar, outBox, menu, menuBar, fileentry[NUM_FILE_MENU_ITEM], 
	      entry[NUM_MENU_ITEM], form, fileBar, filemenu, box, outerform,
	      innerform, dispmenuBar, dispmenu, dispentry[NUM_DISP_MENU_ITEM],
	      phaseLabel;

static int  numGroups;
static int rtHist_numCurves;
static int phase_displayed;
static Boolean color;

static const char defaultTranslations[] =
    "None<Btn1Down>:    XawPositionSimpleMenu(menu) MenuPopup(menu)";
static const char dispdefaultTranslations[] =
    "None<Btn1Down>:    XawPositionSimpleMenu(dispmenu) MenuPopup(dispmenu)";
static const char filedefaultTranslations[] =
    "None<Btn1Down>:    XawPositionSimpleMenu(filemenu) MenuPopup(filemenu)";

/***********************************************************************/

static int get_groupId(const char *axis_label){

  int id = -1;
  HistData* hdp = 0;
  const char *temp;

  for (int m = 0; m < visi_NumMetrics(); m++) {
     temp = visi_MetricLabel(m);
     for (int r = 0; r < visi_NumResources(); r++) {
	if(visi_Enabled(m,r)){
	   if(P_strcmp(temp,axis_label) == 0){
	      hdp = (HistData *)visi_GetUserData(m,r);
	      if(hdp){
                id =  hdp->groupId;  
		break;
              }
           }
	}
  }}
  
  if(id == -1){  // create a new group number
    id = numGroups++;
  }

  return(id);

}


/***********************************************************************/

static int add_new_curve(unsigned m,unsigned r)
{
    HistData* hdp = 0;

    if (visi_Enabled(m,r) &&
       (hdp = (HistData *) visi_GetUserData(m,r)) == 0) {

	hdp = (HistData *) malloc(sizeof(HistData));
	assert(hdp);

	hdp->curve_id  = -1;
	hdp->groupId  = -1;
	hdp->prev_last = 0;
	hdp->metric    = m;
	hdp->resource  = r;

	const char* m_name = visi_MetricName(m);
	const char* r_name = visi_ResourceName(r);

	m_name = (m_name) ? (m_name) : "";
	r_name = (r_name) ? (r_name) : "";

	hdp->title   = (char *) malloc(strlen(m_name)+strlen(r_name)+4);
	assert(hdp->title);
	sprintf(hdp->title, "%s <%s>", m_name, r_name);

	CurveRec curve_rec;
	curve_rec.name = hdp->title;
	// hdp->title = 0;
	curve_rec.groupID = 0;
	curve_rec.curveInfo.yTitle = visi_MetricLabel(m);

	// check for group membership based on yTitle
	curve_rec.groupID = get_groupId(curve_rec.curveInfo.yTitle);
	hdp->groupId = curve_rec.groupID;

	// TODO: set start time to real start time rather that 0.0
	// HistSampleInfo(hw, visi_BucketWidth(), visi_NumBuckets(), 0.0, 0);
	HistSampleInfo(hw, visi_BucketWidth(), visi_NumBuckets(), 
		       visi_GetStartTime(), 0);
	HistAddNewCurve(hw, YTITLE, &curve_rec, 1);
	hdp->curve_id = curve_rec.curveID;

	visi_SetUserData(m,r,(void *) hdp);

        const float* data = visi_DataValues(m,r);

        if (data != 0) {
           HistSetCurveData(hw, hdp->curve_id, hdp->prev_last,
                     visi_LastBucketFilled(m,r)-hdp->prev_last,
                    &(data[hdp->prev_last]));
              hdp->prev_last = visi_LastBucketFilled(m,r);
        } else {
                    float zero = 0.0;
                    HistSetCurveData(hw, hdp->curve_id, 0, 1, &zero);
        }
	rtHist_numCurves++;
    }
    return 0;
}

/************************************************************************/
void ChangeLabel(){

    if(!phase_displayed && XtIsRealized(phaseLabel)) {
        Arg       args[20];
	int handle = visi_GetMyPhaseHandle();
	int i=0;
	if (handle == -1) { // global phase
            XtSetArg(args[i], XtNshowGrip, False);  i++;
  	    XtSetArg(args[i],XtNlabel,"Phase: Global"); i++;
	    XtSetValues(phaseLabel,args,i);
            phase_displayed = 1;
	}
	else if (handle >= 0){
            const char *name = visi_GetMyPhaseName(); 
	    if(name) {
		string new_name;
		new_name = new char[100];
		new_name = string("Phase: ") + string(name);
                XtSetArg(args[i], XtNshowGrip, False);  i++;
  	        XtSetArg(args[i],XtNlabel,P_strdup(new_name.string_of())); i++;
	        XtSetValues(phaseLabel,args,i);
                phase_displayed = 1;
	    }
	    else {
                XtSetArg(args[i], XtNshowGrip, False);  i++;
  	        XtSetArg(args[i],XtNlabel,"Phase: Current"); i++;
	        XtSetValues(phaseLabel,args,i);
	    }
	}
    }
}

/************************************************************************/
static void enable_data(){
   
    HistData* hdp;
    string msg;
    string metricList;
    int numCurves=MAXCURVES;

    if(!phase_displayed)  ChangeLabel(); 

    for (int m = 0; m < visi_NumMetrics(); m++) {
	for (int r = 0; r < visi_NumResources(); r++) {
	    // newly enabled curve
	    if (visi_Enabled(m,r) &&   
		((hdp = (HistData *) visi_GetUserData(m,r)) == 0)) {
		// check if MAXCURVES been added if so call stop_data 
		if(rtHist_numCurves >= MAXCURVES){
		    visi_SetUserData(m,r,0);
	            visi_StopMetRes(m,r);
                    numCurves++;
                    metricList += string(visi_MetricName(m));
                    metricList += string("<");
                    metricList += string(visi_ResourceName(r));
                    metricList += string("> ");
//fprintf(stderr,"too many curves to add %s %s num curves %d\n",
//MetricName(m),ResourceName(r),rtHist_numCurves);
                }
		else {
		    add_new_curve(m,r);
		    hdp = (HistData *) visi_GetUserData(m,r);
		    assert(hdp);
// fprintf(stderr,"added new curves to add %s %s num curves %d\n",
// 	visi_MetricName(m),visi_ResourceName(r),rtHist_numCurves);
                }
	    }
     }}
     hdp = 0;
     if (numCurves > MAXCURVES) {
       msg = string("Maximum number of curves has been exceeded. Only ") + 
             string(MAXCURVES);
       msg += string(" out of ");
       msg += string(numCurves);
       msg += string(" can be displayed at the same time.");
       msg += string(" The following selected metrics are not being displayed: ");
       msg += metricList;
       visi_showErrorVisiCallback(msg.string_of());
     }
}


/************************************************************************/

static void draw_data(unsigned is_fold)
{
    HistData* hdp;

    if(!phase_displayed) ChangeLabel();
    for (int m = 0; m < visi_NumMetrics(); m++) {
	for (int r = 0; r < visi_NumResources(); r++) {
	    if (!visi_Valid(m,r)) {  // invalid histo-cell 
		continue;
	    }
	    // this curve was previously deleted and now has new data 
	    // need to add new curve to histogram widget 
	    if((hdp = (HistData *) visi_GetUserData(m,r)) == 0){  
		add_new_curve(m,r);
		hdp = (HistData *) visi_GetUserData(m,r);
		fprintf(stderr,"$$$$$$$$this should never execute\n");
		assert(hdp);
	    }

	    const float* data = visi_DataValues(m,r);

	    //
	    // why is this being set everytime not just outside the loop ???
	    // jkh 5/2/94

	    // TODO: set start time to real start time rather that 0.0
	    // HistSampleInfo(hw, BucketWidth(), NumBuckets(),
	    // 0.0, FALSE);
	    HistSampleInfo(hw,visi_BucketWidth(),visi_NumBuckets(),
			   visi_GetStartTime(),0);

	    if (is_fold) {
		HistSetCurveData(hw,hdp->curve_id,0,
				 visi_LastBucketFilled(m,r),data);
	    } else {
		HistSetCurveData(hw, 
		    	hdp->curve_id, 
		    	hdp->prev_last,
		    	visi_LastBucketFilled(m,r) - hdp->prev_last,
		    	&(data[hdp->prev_last]));
	    }
	    hdp->prev_last = visi_LastBucketFilled(m,r);
	}
    }  

    return;
}

/***********************************************************************/

static int
process_datavalues(int)
{
    draw_data(0);
    return 0;
}

static int
process_newmr(int)
{
    enable_data();
    return 0;
}

static int
process_fold(int)
{
    draw_data(1);
    return 0;
}

static int
process_phasestart(int)
{
    return 0;
}

static int
process_phaseend(int)
{
    return 0;
}

static int
process_phasedata(int)
{
    return 0;
}

static int
process_paradynexited(int)
{
    return 0;
}

/***********************************************************************/

static HistData*
curve_to_hdp(int cid)
{
    HistData* hdp;
    for (int m = 0; m < visi_NumMetrics(); m++) {
    for (int r = 0; r < visi_NumResources(); r++) {
        if ((hdp = (HistData *) visi_GetUserData(m,r)) == 0) {
            continue;
        }

        if (hdp->curve_id == cid) {
            return hdp;
        }
    }}

    return 0;
}


/***********************************************************************/

void 
Stopdata(int cid){

   HistData* hdp = 0;
   
   for (int m = 0; m < visi_NumMetrics(); m++) {
     for (int r = 0; r < visi_NumResources(); r++) {
       if ((hdp = (HistData *) visi_GetUserData(m,r)) != 0) {
         if(hdp->curve_id == cid){
	    free(hdp->title);
	    free(hdp);
	    visi_SetUserData(m,r,0);
	    visi_StopMetRes(m,r);
            rtHist_numCurves--;
   } } } }
}

/***********************************************************************/

static void
rthist_cleanup(void)
{
    HistData* hdp = 0;

    HistClose(hw);
    for (unsigned i=0; i < NUM_MENU_ITEM; i++) {
        XtDestroyWidget(entry[i]);
    }
    XtDestroyWidget(menu);
    XtDestroyWidget(dispmenu);
    XtDestroyWidget(titleBar);
    XtDestroyWidget(outBox);
    XtDestroyWidget(toplevel);

    for(int i1=0; i1 < visi_NumMetrics(); i1++) {
	for(int j = 0; j < visi_NumResources(); j++){
	    if((hdp = (HistData *)(visi_GetUserData(i1,j))) != 0){
	        free(hdp->title);
		free(hdp);
    	    }
	}
    }
    visi_QuitVisi();
    exit(0);

    /* NOTREACHED */
    return;
}

/***********************************************************************/

struct ReqInfo {
    char**    names;
    char**    display;
    unsigned* curves;
};


/***********************************************************************/

static void
my_visi_callback(void* , int* , long unsigned int* )
{
    int ret;

    ret = visi_callback();
    if (ret == -1) exit(0);
}

/***********************************************************************/

static void file_menu_select(Widget, void* selectp, caddr_t)
{

    int select = 0; 
    select = (int)(Address)selectp;
    switch (select) {

    case CLOSE:
        rthist_cleanup();
        break;

    default:
        fprintf(stderr, "file_menu_select: unknown request from menu\n");
        abort();

        /* NOTREACHED */
        break;
    }
}
/***********************************************************************/

static void menu_select(Widget , void* selectp, caddr_t )
{
    int select = 0; 
    select = (int)(Address)selectp;

    int* curves   = 0;
    int  n_curves = 0;

    switch (select) {
    case NEWCURVE:
        visi_GetMetsRes(0,0);
        break;

    case DELETECURVE:
        HistDeleteCurve(hw, &curves, &n_curves);
	{ // THIS WILL GO AWAY WHEN -fno-for-scope is removed
        for (int i = 0; i < n_curves; i++) {
            HistData* hdp = curve_to_hdp(curves[i]);
	    /* stop data flow for this curve */
	    if(hdp != 0){
               Stopdata(hdp->curve_id);
	    }
        }
	}
        break;

    case MAKEPERSISTENT:
        (void) visi_RegistrationCallback(PARADYNEXITED,process_paradynexited);
	break;
    default:
        fprintf(stderr, "menu_select: unknown request from menu\n");
        abort();

        /* NOTREACHED */
        break;
    }

/* possible memory leak. I need to free `curves' */

    return;
}


/***************************************************************/
static void disp_menu_select(Widget , void* selectp, caddr_t )
{
    int select = 0; 
    select = (int)(Address)selectp;

    int* curves   = 0;
    int  n_curves = 0;

    switch (select) {

    case SMOOTH:
        HistSmoothCurve(hw, &curves, &n_curves);
        break;

    case UNSMOOTH:
        HistUnsmoothCurve(hw, &curves, &n_curves);
        break;

    case YMAX:
        HistResetYmax(hw);
        break;

    case HIDECURVE:
	n_curves = 0;
        HistHideCurve(hw);
	break;
    
    case BW_COLOR:
	if(color){ 
	    color = False;
        }
	else { 
	    color = True;
	}
	HistColorChange(hw,color);
	break;

    case UNHIDECURVE:
	n_curves = 0;
        HistShowCurve(hw);
	break;

    default:
        fprintf(stderr, "menu_select: unknown request from menu\n");
        abort();

        /* NOTREACHED */
        break;
    }

/* possible memory leak. I need to free `curves' */

    return;
}

/***************************************************/

/*
 * default resources
 */

String fallback_resources[] = { 
 "*background: Grey",
 "*foreground: Black",
 "*font: *-Helvetica-*-r-*-12-*",
 "*title*font: *-New*Century*Schoolbook-Bold-R-*-18-*",
 "*title*background: DarkSlateBlue",
 "*title*foreground: White",
 "*menu*font: *-New*Century*Schoolbook-Bold-R-*-14-*",
 "*filemenu*font: *-New*Century*Schoolbook-Bold-R-*-14-*",
 "*dispmenu*font: *-New*Century*Schoolbook-Bold-R-*-14-*",
 "*filemenu*borderColor: Grey",
 "*dispmenu*borderColor: Grey",
 "*box*borderColor: Grey",
 "*box*foreground: HotPink4",
 "*innerform*borderColor: Grey",
 "*menu*borderColor: Grey",
 NULL,
};

/***********************************************************************/

int
main(int argc, char** argv)
{
    int            i;
    Arg            args[20];
    XtAppContext   app_con;
    XtTranslations trans_table;
    XtTranslations disp_trans_table;
    XtTranslations file_trans_table;
    Display	   *dpy;
    Pixmap	   *logobitmap;

    color = True;
    /* check to see if first argument is -bw, if so set start color to bw */
    if((argc > 1) && (strcmp(argv[1],"-bw") == 0)){
        argc--;
        color = False;
	argv[1] = argv[0];
	argv = &(argv[1]);
    }

    phase_displayed = 0;
    numGroups = 0;
    rtHist_numCurves = 0;
    toplevel = XtAppInitialize(&app_con, "Rthist", 0, 0,
        &argc, argv, fallback_resources, NULL, ZERO);

    dpy = XtDisplay(toplevel);

    XtVaSetValues(toplevel, XtNminHeight, 380, NULL);
    XtVaSetValues(toplevel, XtNminWidth, 570, NULL);

    i = 0;
    XtSetArg(args[i], XtNorientation, XtorientVertical); i++;
    XtSetArg(args[i], XtNshowGrip, False);  i++;
    outBox = XtCreateManagedWidget("hist", panedWidgetClass, toplevel,
        args, i);

    logobitmap = (Pixmap *)XCreateBitmapFromData( dpy,
				RootWindowOfScreen(XtScreen(toplevel)),
			  	(const char *)logo_bits,
			 	logo_width,
				logo_height);

    i = 0;
    XtSetArg(args[i], XtNshowGrip, False);  i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNheight, logo_height);  i++;
    outerform = XtCreateManagedWidget("outerform", formWidgetClass, outBox,
        args, i);

    i = 0;
    XtSetArg(args[i], XtNshowGrip, False);  i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNvertDistance, 0);  i++;
    XtSetArg(args[i], XtNhorizDistance, 0);  i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    XtSetArg(args[i], XtNwidth, 590);  i++;
    XtSetArg(args[i], XtNheight, logo_height);  i++;
    innerform = XtCreateManagedWidget("innerform", panedWidgetClass, outerform,
        args, i);



    i = 0;
    XtSetArg(args[i], XtNbitmap, logobitmap); i++;
    XtSetArg(args[i], XtNfromHoriz, innerform); i++;
    XtSetArg(args[i], XtNshowGrip, False);  i++;
    XtSetArg(args[i], XtNjustify, XtJustifyRight);  i++;
    XtSetArg(args[i], XtNwidth, (logo_width + 5));  i++;
    XtSetArg(args[i], XtNheight, logo_height);  i++;
    XtSetArg(args[i], XtNvertDistance, 0);  i++;
    XtSetArg(args[i], XtNhorizDistance, 0);  i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNright, XtChainRight); i++;
    XtSetArg(args[i], XtNleft, XtChainRight); i++;
    box = XtCreateManagedWidget( "box", labelWidgetClass, outerform, args, i); 

    i = 0;
    XtSetArg(args[i], XtNshowGrip, False);            i++;
    XtSetArg(args[i], XtNlabel, "Time Histogram Display"); i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNvertDistance, 0);  i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    XtSetArg(args[i], XtNhorizDistance, 0);  i++;
    titleBar = XtCreateManagedWidget("title", labelWidgetClass,
        innerform, args, i);

    i=0;
    XtSetArg(args[i],XtNshowGrip, False);            i++;
    XtSetArg(args[i], XtNfromVert, titleBar); i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    form = XtCreateManagedWidget( "form", formWidgetClass, innerform,
	   args, i);


    i = 0;
    XtSetArg(args[i], XtNshowGrip, False);            i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNright, XtChainLeft); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNlabel,  "File"); i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    fileBar = XtCreateManagedWidget("filemenu", menuButtonWidgetClass,
        form, args, i);

    i = 0;
    XtSetArg(args[i], XtNfromHoriz, fileBar); i++;
    XtSetArg(args[i], XtNshowGrip, False);            i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNright, XtChainLeft); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    XtSetArg(args[i], XtNlabel,  "Actions"); i++;
    menuBar = XtCreateManagedWidget("menu", menuButtonWidgetClass,
        form, args, i);

    i = 0;
    XtSetArg(args[i], XtNfromHoriz, menuBar); i++;
    XtSetArg(args[i], XtNshowGrip, False);            i++;
    XtSetArg(args[i], XtNresizable, True);  i++;
    XtSetArg(args[i], XtNright, XtChainLeft); i++;
    XtSetArg(args[i], XtNleft, XtChainLeft); i++;
    XtSetArg(args[i], XtNborderWidth, 0);  i++;
    XtSetArg(args[i], XtNlabel,  "View"); i++;
    dispmenuBar = XtCreateManagedWidget("dispmenu", menuButtonWidgetClass,
        form, args, i);

    file_trans_table = XtParseTranslationTable(filedefaultTranslations);
    XtOverrideTranslations(fileBar, file_trans_table);

    i = 0;
    XtSetArg(args[i], XtNborderColor, "black"); i++;
    filemenu = XtCreatePopupShell("filemenu", simpleMenuWidgetClass,
        fileBar, args,i);

    for (i = 0; i < NUM_FILE_MENU_ITEM; i++) {
	fileentry[i] = XtCreateManagedWidget(file_menu_items[i],
	   smeBSBObjectClass, filemenu, NULL, 0);
        XtAddCallback(fileentry[i], XtNcallback,
		      (XtCallbackProc)file_menu_select,
		      (XtPointer) i);
    }

    trans_table = XtParseTranslationTable(defaultTranslations);
    XtOverrideTranslations(menuBar, trans_table);

    i = 0;
    XtSetArg(args[i], XtNborderColor, "black"); i++;
    menu = XtCreatePopupShell("menu", simpleMenuWidgetClass,
        menuBar, args, i);

    XtSetArg(args[0], XtNborderColor, "black");
    for (i = 0; i < NUM_MENU_ITEM; i++) {
        entry[i] = XtCreateManagedWidget(menu_item[i],
            smeBSBObjectClass, menu, args, 1);
        XtAddCallback(entry[i], XtNcallback,
		      (XtCallbackProc)menu_select,
		      (XtPointer) i);
    }

    disp_trans_table = XtParseTranslationTable(dispdefaultTranslations);
    XtOverrideTranslations(dispmenuBar, disp_trans_table);

    i = 0;
    XtSetArg(args[i], XtNborderColor, "black"); i++;
    dispmenu = XtCreatePopupShell("dispmenu", simpleMenuWidgetClass,
        dispmenuBar, args, i);

    for (i = 0; i < NUM_DISP_MENU_ITEM; i++) {
        dispentry[i] = XtCreateManagedWidget(disp_menu_item[i],
            smeBSBObjectClass, dispmenu, args, 1);
        XtAddCallback(dispentry[i], XtNcallback,
		      (XtCallbackProc)disp_menu_select,
		      (XtPointer)i);
    }

    i = 0;
    XtSetArg(args[i], XtNshowGrip, False);  i++;
    XtSetArg(args[i],XtNlabel,"Phase: "); i++;
    phaseLabel = XtCreateManagedWidget("phase", labelWidgetClass,
        outBox, args, i);

    i = 0;
    XtSetArg(args[i], XtNdefaultGC, color); i++;
    XtSetArg(args[i], XtNsmoothNpts, 10); i++;
    XtSetArg(args[i], XtNfromVert, outerform); i++;
    XtSetArg(args[i],XtNyInc,0.1); i++;
    hw = XtCreateManagedWidget("graph", histogramWidgetClass,
        outBox, args, i);



    // InitReqWindow(toplevel, 0, 0, 13, NULL, 0, 0);

    //
    // code to handle the initialization of the Visi library
    //

    int vfd = -1;
    if ((vfd = visi_Init()) == -1) {
        fprintf(stderr, "error initializing Visi library\n");
        exit(-1);
    }

    (void) visi_RegistrationCallback(DATAVALUES,          process_datavalues);
    (void) visi_RegistrationCallback(FOLD,                process_fold);
    (void) visi_RegistrationCallback(ADDMETRICSRESOURCES, process_newmr);
    (void) visi_RegistrationCallback(PHASESTART,     process_phasestart);
    (void) visi_RegistrationCallback(PHASEEND, 	     process_phaseend);
    (void) visi_RegistrationCallback(PHASEDATA,      process_phasedata);
    XtAppAddInput(app_con, vfd, (XtPointer) XtInputReadMask, my_visi_callback,
       (XtPointer) hw);

    for (int l = 0; l < visi_NumMetrics(); l++) {
	for (int r = 0; r < visi_NumResources(); r++) {
	    visi_SetUserData(l,r,0);
	}   
    }

    //
    // Visi library should now be initialized
    //
    XtRealizeWidget(toplevel);

    XtAppMainLoop(app_con);

    /* NOTREACHED */
    return 0;
}
