/*
 * Copyright (c) 1996-1999 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */
#ifndef WAITSET_H
#define WAITSET_H


class WaitSet
{
private:
#if defined(i386_unknown_nt4_0)
#define	WAITSET_MAX_OBJECTS	(MAXIMUM_WAIT_OBJECTS - 1)

	struct ItemInfo
	{
		thread_t	tid;
		bool		isSocket;
        HANDLE      obj;

		void	Set( thread_t _tid, HANDLE h );
		void	Reset( void )		{ tid = THR_TID_UNSPEC;
                                      isSocket = false;
                                      obj = INVALID_HANDLE_VALUE; }
	};
	ItemInfo	waitInfo[WAITSET_MAX_OBJECTS];
	unsigned	nObjects;
	thread_t	wmsgTid;

	unsigned	Find( thread_t tid ) const;
	bool		DoAdd( thread_t tid );
	unsigned	DoRemove( thread_t tid );
	unsigned	FindNextValid( unsigned index ) const;
	unsigned	FindNextInvalid( unsigned index ) const;
	void		Compact( void );
	void		Shift( unsigned idx_start, unsigned idx_end, unsigned idx_newstart );
#else // defined(i386_unknown_nt4_0)
	fd_set	fds;
#endif // defined(i386_unknown_nt4_0)
	unsigned	nWaitees;


public:
	enum WaitReturn
	{
		WaitInput,
		WaitTimeout,
		WaitError
	};

	WaitSet( void )		{ Clear(); }
	~WaitSet( void )	{ Clear(); }

	bool		Add( thread_t tid );
	bool		Add( const WaitSet& ws );
	void		Remove( thread_t tid );
	void		Remove( const WaitSet& ws );
	bool		Contains( thread_t tid ) const;
	void		Clear( void );
	unsigned	Count( void ) const		{ return nWaitees; }
	bool		IsEmpty( void ) const	{ return (nWaitees == 0); }
	WaitReturn	Wait( bool block,
						thread_t last_tid_hint,
						thread_t* tidp,
						tag_t* tagp ) const;

#if defined(i386_unknown_nt4_0)
    void    UpdateAvailableDataState( SOCKET sock,
                                        char* buf,
                                        unsigned int bufLen,
                                        int recvRet );
#endif // defined(i386_unknown_nt4_0)

	friend class WaitSetIterator;
};


class WaitSetIterator
{
private:
	const WaitSet& ws;
	unsigned index;

public:
	WaitSetIterator( const WaitSet& wset )	: ws( wset )	{ Reset(); }

	bool	Next( thread_t* ptid );
	void	Reset( void )									{ index = 0; }
};

#if defined(THREADLIB_DEBUG)
ostream& operator<<( ostream& s, const WaitSet& ws );
#endif // defined(THREADLIB_DEBUG)

#endif // WAITSET_H
