/**************** f_meas.c ***************************************/
/* MIMD version 6 */
/* DT 12/97 */
/* Kogut-Susskind fermions  -- this version for "fat plus Naik"
   or general "even plus odd" quark actions.  Assumes "dslash" has
   been defined to be the appropriate "dslash_fn" or "dslash_eo"
*/

/* Measure fermionic observables:
    psi-bar-psi (separately on even and odd sites)
    fermion action

    This routine uses g_rand, phi, xxx, and other vectors used by
    the matrix inversion.
*/

#include "generic_ks_includes.h"	/* definitions files and prototypes */

void f_meas_imp( field_offset phi_off, field_offset xxx_off, float mass ){
    float r_psi_bar_psi_even, r_psi_bar_psi_odd;
    float r_ferm_action;
    /* local variables for accumulators */
    register int i;
    register site *st;
    float rpbp_e,rpbp_o,rfaction;
    complex cc;
#ifdef FN
    if(!(valid_longlinks==1)) load_longlinks();
    if(!(valid_fatlinks==1)) load_fatlinks();
#endif

    rpbp_e = rpbp_o = rfaction = 0.0;

    /* Make random source, and do inversion */
    grsource_imp( phi_off, mass, EVENANDODD );
    mat_invert_uml( F_OFFSET(g_rand), xxx_off, phi_off, mass );

    /* fermion action = phi.xxx */
    /* psi-bar-psi on even sites = g_rand.xxx */
    FOREVENSITES(i,st){
        cc = su3_dot( (su3_vector *)F_PT(st,phi_off),
	    (su3_vector *)F_PT(st,xxx_off) );
	rfaction += cc.real;
        cc = su3_dot( &(st->g_rand), (su3_vector *)F_PT(st,xxx_off) );
	rpbp_e += cc.real;
    }
    /* psi-bar-psi on odd sites */
    FORODDSITES(i,st){
	cc = su3_dot( &(st->g_rand), (su3_vector *)F_PT(st,xxx_off) );
	rpbp_o += cc.real;
    }

    g_floatsum( &rpbp_o );
    g_floatsum( &rpbp_e );
    g_floatsum( &rfaction );

    r_psi_bar_psi_odd =  rpbp_o*(2.0/(double)volume) ;
    r_psi_bar_psi_even =  rpbp_e*(2.0/(double)volume) ;
    r_ferm_action =  rfaction*(1.0/(double)volume) ;
    node0_printf("PBP: mass %e     %e  %e\n",mass,
	r_psi_bar_psi_even,r_psi_bar_psi_odd);
    node0_printf("FACTION: mass = %e,  %e\n",mass,r_ferm_action);
}
