#ifndef _COMDEFS_H
#define _COMDEFS_H
/************************ comdefs.h *************************************
*									*
*  Macros and declarations for communications routines com_*.c          *
*  MIMD version 6 							*
*									*
*/

/* Version that allows 32 sublattices for extended actions */

/* 
   
   Communications routines will assume that the lattice is stored as an
   array of structures of type "site".

   Modifications:
   1/27/00 Included option of 32 sublattices (for improved actions) -- UMH
   2/26/98 Changed type32 to signed
   8/5/97 ANSI prototyping for com_*.c routines C.D.
   8/10/96  Added defs for broadcast_bytes and wrappers for parallel file IO C.D.
*/

/* For portability we require a 32-bit integer type for writing
   integer data to binary files.  Machine-dependent. Add more as
   needed. Compile with -DSHORT32 for 64-bit integer machines.
   like the T3E */

#include "../include/type32.h"

#include "../include/su3.h"
#include "../include/complex.h"
#include "../include/macros.h"
#include <lattice.h>     /* Needed ONLY for "site" in this header */

/* message types  (not all are used on all machines) */
#define PARAM_TYPE 11	/* type of parameter message to nodes */
#define FIELD_TYPE 12	/* type of field sent from one node to another */
#define BROADCAST_FLOAT_TYPE 13	/* broadcast of floating point number */
#define BROADCAST_DOUBLE_TYPE 14 /* broadcast of double */
#define BROADCAST_COMPLEX_TYPE 15 /* broadcast of single precision complex */
#define BROADCAST_DCOMPLEX_TYPE 16 /* broadcast of double precision complex */
#define SEND_INTEGER_TYPE 17	/* send an integer to one other node */
#define CM_GATHER_TYPE 18	/* type for CM5 (cooperative) messages */
#define BROADCAST_BYTES_TYPE 19 /* send n raw bytes from node 0 to all nodes */
#define SYNC_TYPE 50		/* Synchronize all nodes */
#define SUM_FLOAT_TYPE 51	/* Sum float over all nodes */
#define SUM_DOUBLE_TYPE 52	/* Sum double over all nodes */
#define SUM_COMPLEX_TYPE 53	/* Sum complex over all nodes */
#define SUM_DCOMPLEX_TYPE 54	/* Sum double_complex over all nodes */
#define MAX_FLOAT_TYPE 55	/* Maximum float over all nodes */
#define MAX_DOUBLE_TYPE 56	/* Maximum double over all nodes */
#define SUM_VECDOUBLE_TYPE 61   /* Sum a vector of doubles over all nodes */
#define SUM_VECDCOMPLEX_TYPE 62 /* Sum a vector of double complex over nodes */

/* For MPI, need to include mpi.h to define MPI_Request */
/* For MPI on paragon, mynode() and numnodes() can't replace nx routines */
#ifdef MPI
#include <mpi.h>
#define mynode MILC_mynode
#define numnodes MILC_numnodes
#define dclock MILC_dclock
#endif

/* Added for pvm */
#ifdef PVM
#define ANY_MSG -1              /* Any message */

/* Message structures for communication between host and nodes */
/* The first two fields must always be the same for each type */
/* The basic structure must be the shortest */

/* For most messages */
struct hcs_basic {
  int msg_type;
  int node;	/* integer identifies caller's instance */
  int arg1,arg2,arg3;  /* Use depends on which routine */
} ;
#define HCS_BASIC_SIZE (sizeof(struct hcs_basic))
int put_hcs_basic( struct hcs_basic * hcs );
int get_hcs_basic( struct hcs_basic * hcs );

/* For printf and scanf calls */
#define STRING_LENGTH 256
struct hcs_stdio {
  int msg_type;
  int node;	/* integer identifies caller's instance */
  int length;
  char s[STRING_LENGTH];
} ;
#define HCS_STDIO_SIZE (sizeof(struct hcs_stdio))
int put_hcs_stdio( struct hcs_stdio * hcs );
int get_hcs_stdio( struct hcs_stdio * hcs );

/* For initialization call */
#define MAX_NUMBER_NODES 512
#define HOST_NAME_LENGTH 128
struct hcs_ident {
  int msg_type;
  int node;               /* integer identifies caller's instance */
  int your_node;          /* logical node number for this node */
  int number_nodes;       /* how many nodes for this partition */
  int node_instance[MAX_NUMBER_NODES];   /* instance number for logical node */
  char host_name[HOST_NAME_LENGTH];     /* name of host */
} ;
#define HCS_IDENT_SIZE (sizeof(struct hcs_ident))
int put_hcs_ident( struct hcs_ident * hcs );
int get_hcs_ident( struct hcs_ident * hcs );

union {
  struct hcs_basic basic;
  struct hcs_stdio stdio;
  struct hcs_ident ident;
} hcs;

#define HOST_CALL 77	/* pvm message type for call to host for service */
#define HOST_REPLY 87

/* Message subtypes internal to host-node service calls */
/* (Not used by pvm to identify messages) */
#define PRINTF_HOST_CALL 11
#define SCANF_HOST_CALL 12
#define FPRINTF_HOST_CALL 13 /* Not used */ 
#define FSCANF_HOST_CALL 14 /* Not used */
#define FFLUSH_HOST_CALL 20
#define FOPEN_HOST_CALL 30 /* Not used */
#define FCLOSE_HOST_CALL 31 /* Not used */
#define OPEN_HOST_CALL 32 /* Not used */
#define CLOSE_HOST_CALL 33 /* Not used */
#define CREAT_HOST_CALL 34 /* Not used */
#define READ_HOST_CALL 40 /* Not used */
#define WRITE_HOST_CALL 41 /* Not used */
#define NODE_IDENT_CALL 78	
#define NODES_DONE_HOST_CALL 99

/* end of pvm additions */
#endif	/* end ifdef PVM */

/* Added for pvm */
#ifdef PVM24
#define ANY_MSG -1              /* Any message */

/* Message structures for communication between host and nodes */
/* The first two fields must always be the same for each type */
/* The basic structure must be the shortest */

/* For initialization call */
#define MAX_NUMBER_NODES 512
#define HOST_NAME_LENGTH 128
struct hcs_ident_struct {
  int msg_type;
  int node;               /* integer identifies caller's instance */
  int your_node;          /* logical node number for this node */
  int number_nodes;       /* how many nodes for this partition */
  int node_instance[MAX_NUMBER_NODES];   /* instance number for logical node */
  char host_name[HOST_NAME_LENGTH];     /* name of host */
} hcs_ident ;
#define HCS_IDENT_SIZE (sizeof(struct hcs_ident_struct))
int put_hcs_ident( struct hcs_ident_struct * hcs );
int get_hcs_ident( struct hcs_ident_struct * hcs );

#define NODE_IDENT_CALL 77	
#define terminate g_terminate   /* Because of name conflict */

/* end of pvm version 2.4 additions */
#endif	/* end ifdef PVM24 */


/* Added for pvm */
#ifdef PVM3
#define ANY_MSG -1              /* Any message */
#define ANY_NODE -1              /* Any node */

/* Message structures for communication between host and nodes */

/* For initialization call */
#define MAX_NUMBER_NODES 512
#define HOST_NAME_LENGTH 128
struct hcs_ident_struct {
  int msg_type;
  int node;               /* integer identifies caller's instance */
  int your_node;          /* logical node number for this node */
  int number_nodes;       /* how many nodes for this partition */
  int node_tid[MAX_NUMBER_NODES];   /* instance number for logical node */
  char host_name[HOST_NAME_LENGTH];     /* name of host */
} hcs_ident ;
#define HCS_IDENT_SIZE (sizeof(struct hcs_ident_struct))
int put_hcs_ident( struct hcs_ident_struct * hcs );
int get_hcs_ident( struct hcs_ident_struct * hcs );

#define NODE_IDENT_CALL 77	

/* end of pvm version 3 additions */
#endif	/* end ifdef PVM3 */

/* Added for MPL */
#ifdef MPL
#define MPL_NOT_COMPLETED -1          /* For mpc_status */
#define MPL_INACTIVE -2         /* For mpc_status */
#define DONTCARE -1             /* Any message or any node */
int mperrno;             /* Used for error reporting */
/* The following are from /usr/lpp/poe/include/mpproto.h */
extern int mpc_environ(int *howmany,int *whoami);
extern int mpc_stopall(int errcode);
extern int mpc_group(int gsize,int glist[],int label,int *gid);
extern int mpc_send(char *sarr,int len,int dest,int type,int *msgid);
extern int mpc_recv(char *darr,int len,int *src,int *type,int *msgid);
extern int mpc_bsend(char *sarr,int len,int dest,int type);
extern int mpc_brecv(char *darr,int len,int *src,int *type,int *nbytes);
extern int mpc_status(int msgid);
extern int mpc_wait(int *msgid,int *nbytes);
extern int mpc_sync(int gid);

extern void s_vadd(float in1[],float in2[],float out[],int *len);
extern void d_vadd(double in1[],double in2[],double out[],int *len);
extern void s_vmax(float in1[],float in2[],float out[],int *len);
extern void d_vmax(double in1[],double in2[],double out[],int *len);
#define MAX_NUMBER_NODES 512
#define HOST_NAME_LENGTH 128

/* end of MPL additions */
#endif	/* end ifdef MPL */

#define FIELD_REQUEST 100	/* used by field_pointer...() */
#define FIELD_REPLY 101		/* used by field_pointer...() */
#define GENERAL_GATHER_BASE 102	/* types from this to this+number_of_nodes
				are used by the general_gather routines */
#define GATHER_BASE 614	/* types greater than or equal to this are used
			   by the gather routines */
/* pid */
#define NODE_PID 0
#define HOST_PID 0
#define ALL_NODES -1	/* works for Ncube, Intel */ /* Don't use with pvm */

/* Directions, and a macro to give the opposite direction */
/*  These must go from 0 to 7 because they will be used to index an
    array. */
/* Also define NDIRS = number of directions */
#define XUP 0
#define YUP 1
#define ZUP 2
#define TUP 3
#define TDOWN 4
#define ZDOWN 5
#define YDOWN 6
#define XDOWN 7

#define OPP_DIR(dir)	(7-(dir))	/* Opposite direction */
#define NDIRS 8				/* number of directions */

#define MAX_GATHERS 32	/* Maximum number of gather tables */

/* arguments to the make_gather() routine */
#define FORWARDS 1
#define BACKWARDS (-1)	/* BACKWARDS = -FORWARDS */
#define OWN_INVERSE 0
#define WANT_INVERSE 1
#define NO_INVERSE 2
#define ALLOW_EVEN_ODD 0
#define NO_EVEN_ODD 1
#define SAME_PARITY 0
#define SWITCH_PARITY 1
#define SCRAMBLE_PARITY 2

/* "comlink" is the basic structure used in gathering neighboring sites.
   Each node will maintain one such structure for each direction for each
   (other) node that contains sites that are neighbors of the sites on
   this node.  For example, if the XUP neighbors of sites on this node
   are found on two other nodes, then this node will maintain a linked
   list of two comlink structures for gathering from the XUP direction.
*/
#ifdef N_SUBL32
struct comlink {
	/* pointer to next in list, NULL if this is last */
    struct comlink *nextcomlink;
	/* number of the node to which we connect */
    int othernode;
	/* number of sites in a sublattice on this node that have
	   neighbors on other node connected by this "comlink"
	   as well as the total number. */
    int n_subl_connected[N_SUBL32], n_connected;
	/* Address of list of indices of sites in one sublattice
	   (on this node) whose neighbors are found through this comlink.
	   The lists follow in order of the sublattice number,
	   so to get all sites you start at sitelist[0] and take
	   n_connected addresses. */
	/* When the sublattices are treated separately, the sublattice
	   index always corresponds to the the sublattice of the site
	   on this node. Different comlink structures may point to the
	   same list. For example, the receive list for one gather may
	   be a send list for the opposite gather. */
    int send_subl[N_SUBL32];
	/* The sublattice on the other node for ALLOW_EVEN_ODD.
	   Is not used otherwise. */
    int *sitelist[N_SUBL32];
};
#else
struct comlink {
	/* pointer to next in list, NULL if this is last */
    struct comlink *nextcomlink;
	/* number of the node to which we connect */
    int othernode;
	/* number of even sites on this node that have neighbors on
	   other node connected by this "comlink", and same for odd
	   sites on this node. */
    int n_even_connected, n_odd_connected;
	/* Address of list of indices of even sites (on this node)
	   whose neighbors are found through this comlink, same for odd.
	   The odd list follows the even list, so to get all sites you
	   start at esitelist and take n_even_connected+n_odd_connected
	   addresses. */
	/* When the comlink is describing sites to be sent, the "odd"
	   list lists sites whose neighbors are even.  This convention
	   is natural for the nearest neighbor gathers.  For gathers
	   which don't allow even and odd site gathers, the even list
	   is used for list of sites to be received and the odd
	   list for sites to be sent.  Different comlink structures
	   may point to the same list.  For example, the receive list
	   for one gather may be a send list for the opposite gather. */
    int *esitelist, *ositelist;
};
#endif
typedef struct comlink comlink;


/* Structure to keep track of outstanding sends and receives */
typedef struct {
	/* node sending or receiving message */
    int msg_node;
	/* size of message in bytes */
    int msg_size;
	/* address of buffer malloc'd for message */
    char *msg_buf;
	/* message id returned by system call */
#ifdef MPI
    MPI_Request msg_id;
#else
    int msg_id;
#endif
#if defined(PVM) || defined(PVM24) || defined(PVM3) || defined(MPL)
    int msg_OK;
        /* flag to track the asynchronous arrival of messages */
#endif
#ifdef MPL
    int mpl_msgid;
        /* MPL assigned message id for checking status with mpc_status */
#endif
} msg_tag;

/* Structure for requesting a field from another node */
typedef struct {
    int field;	/* offset of field in site */
    int size;	/* size of field */
    int index;	/* index of field on other node */
} msg_request;



/**********************************************************************/
/* Declarations for all routines called in the com_*.c files */

void start_handlers();
void initialize_machine(int argc, char **argv);
void make_nn_gathers();
void sort_eight_special(void **pt);

void neighbor_coords_special(
 int x,int y,int z,int t, /* coordinates of site */
 int *dirpt,              /* direction (eg XUP) */
 int fb,                  /* "forwards/backwards"  */
 int *x2p,int *y2p,int *z2p,int *t2p);
                          /* pointers to coordinates of neighbor */
int make_gather(
 void (*func)(int, int, int, int, int *, int, int *, int *, int *, int *),
         		/* function which defines sites to gather from */
 int *args,		/* list of arguments, to be passed to function */
 int inverse,		/* OWN_INVERSE, WANT_INVERSE, or NO_INVERSE */
 int want_even_odd,	/* ALLOW_EVEN_ODD or NO_EVEN_ODD */
 int parity_conserve);	/* {SAME,SWITCH,SCRAMBLE}_PARITY */

void neighbor_coords(
 int x, int y, int z, int t,  /* coordinates of site */
 int dir,	              /* direction (eg XUP) */
 int *x2p, int *y2p, int *z2p, int *t2p);
                             /* pointers to coordinates of neighbor */
msg_tag * start_gather(
/* arguments */
 field_offset field,	/* which field? Some member of structure "site" */
 int size,		/* size in bytes of the field (eg sizeof(su3_vector))*/
 int index,		/* direction to gather from. eg XUP - index into
			   neighbor tables */
 int parity,		/* parity of sites whose neighbors we gather.
			   one of EVEN, ODD or EVENANDODD. */
 char ** dest);		/* one of the vectors of pointers */

void restart_gather(
/* arguments */
 field_offset field,	/* which field? Some member of structure "site" */
 int size,		/* size in bytes of the field (eg sizeof(su3_vector))*/
 int index,		/* direction to gather from. eg XUP - index into
			   neighbor tables */
 int parity,		/* parity of sites whose neighbors we gather.
			   one of EVEN, ODD or EVENANDODD. */
 char ** dest,		/* one of the vectors of pointers */
 msg_tag *mbuf);        /* previously returned by start_gather */

msg_tag * start_gather_from_temp(
/* arguments */
 void * field,		/* which field? pointer returned by malloc() */
 int size,		/* size in bytes of the field (eg sizeof(su3_vector))*/
 int index,		/* direction to gather from. eg XUP - index into
			   neighbor tables */
 int parity,		/* parity of sites whose neighbors we gather.
			   one of EVEN, ODD or EVENANDODD. */
 char ** dest);		/* one of the vectors of pointers */

void restart_gather_from_temp(
/* arguments */
 void * field,		/* which field? pointer returned by malloc() */
 int size,		/* size in bytes of the field (eg sizeof(su3_vector))*/
 int index,		/* direction to gather from. eg XUP - index into
			   neighbor tables */
 int parity,		/* parity of sites whose neighbors we gather.
			   one of EVEN, ODD or EVENANDODD. */
 char ** dest,		/* one of the vectors of pointers */
 msg_tag *mbuf);        /* previously returned by start_gather */

void wait_gather(msg_tag *mbuf);
void cleanup_gather(msg_tag *mbuf);

msg_tag * start_general_gather(
/* arguments */
 field_offset field,	/* which field? Some member of structure "site" */
 int size,		/* size in bytes of the field (eg sizeof(su3_vector))*/
 int *displacement,	/* displacement to gather from. four components */
 int parity,		/* parity of sites to which we gather.
			   one of EVEN, ODD or EVENANDODD. */
 char ** dest);		/* one of the vectors of pointers */

void wait_general_gather(msg_tag *mbuf);
void cleanup_general_gather(msg_tag *mbuf);

char * field_pointer_at_coordinates(
/* arguments */
 int field,	/* offset of one of the fields in lattice[] */
 int size,	/* size of the field in bytes */
 int x,int y,int z,int t);	/* coordinates of point to get field from */

char * field_pointer_at_direction(
/* arguments */
 field_offset field,	/* offset of one of the fields in lattice[] */
 int size,	/* size of the field in bytes */
 site *s,	/* pointer to a site on this node */
 int direction);	/* direction of site's neighbor to get data from.
		   one of XUP,XDOWN,YUP... */

void cleanup_field_pointer(char * buf);
void send_field(char *buf, int size, int tonode);
void get_field(char *buf, int size);
char * machine_type();
/* On the Paragon mynode() and numnodes() are library routines with the
   same functionality as ours. */
#ifndef PARAGON
int mynode();
int numnodes();
#endif
void g_sync();
void g_floatsum( float *fpt);
void g_doublesum( double *dpt);
void g_vecdoublesum( double *dpt, int ndoubles);
void g_complexsum( complex *cpt);
void g_dcomplexsum( double_complex *cpt);
void g_veccomplexsum( complex *cpt, int ncomplex);
void g_wvectorsum( wilson_vector *wvpt);
void g_xor32( u_type32 *pt );
void g_floatmax( float *fpt);
void g_doublemax( double *dpt);
void broadcast_float(float *fpt);
void broadcast_double(double *dpt);
void broadcast_complex(complex *cpt);
void broadcast_dcomplex(double_complex *cpt);
void broadcast_bytes(char *buf,int size);
void send_integer(int tonode, int *address);
void receive_integer(int *address);

/* On the Paragon dclock is a library routine with the
   same functionality as ours */
/* Either way, it needs to be declared double */
double dclock();
void time_stamp(char *msg);

void terminate(int status);

/**********************************************************************/
/* Each node maintains a list of headers to lists of comlinks */
/**EXTERN comlink * neighborlist[NDIRS];**/
/* addresses of neighboring sites, NULL if off-node */
/**EXTERN site ** neighbor[NDIRS];**/

/**********************************************************************/
#ifdef NONSENSE
#ifdef PE_CODE
#include <cm/cmmd.h>
/* On CM5, redefine system calls */
#define main CMPE_control
#define printf host_printf
#define scanf host_scanf
#define fprintf host_fprintf
#define fscanf host_fscanf
#define fflush host_fflush
#define fopen host_fopen
#define fclose host_fclose
#define open host_open
#define close host_close
#define creat host_creat
#define read host_read
#define write host_write
#endif
#endif

#ifdef PVM
#ifndef HOST_CODE
#define terminate g_terminate   /* Because of name conflict */
#define scanf host_scanf
#define printf host_printf
#define fflush host_fflush
#endif
#endif

#endif	/* _COMDEFS_H */
